(* (C) 2009-2014 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE ARRAYOFCHAR (*. [AOC] FOR CHAR .*);

  (* temporay kludge until ARRAY OF CHAR is ready *)
IMPORT Perror;
IMPORT ISO646;
FROM Transition IMPORT CAPACITYch;


PROCEDURE (*. [..] .*) CopySlice
  ( VAR   target : ARRAY OF CHAR; tgtStartIndex, tgtEndIndex : INTEGER;
    (*. CONST .*) source : ARRAY OF CHAR; srcStartIndex, srcEndIndex : INTEGER );

VAR
  tgtStart, tgtEnd, srcStart, srcEnd, tLen, sLen, tCount, sCount : CARDINAL;
  hiTarget, hiSource : CARDINAL;
  tgtRev, srcRev : BOOLEAN;

BEGIN (* CopySlice *)
  tLen := LENGTH (target);
  sLen := LENGTH (source);
  hiTarget := tLen - 1;
  hiSource := sLen - 1;
        
    (* since some of the indices may be negative, convert all to absolute indices *)
    (* while we are at it check on the sanity of the supplied indices *)

  tgtStart := ABS (tgtStartIndex);
  IF tgtStartIndex < 0 THEN
    tgtStart := (tLen - tgtStart)
  END; (* if tgtStartIndex *)
  IF tgtStart > hiTarget THEN Perror.prerr ( "tgtStartIndex Out of Bounds "); RETURN END;

  tgtEnd := ABS (tgtEndIndex);
  IF tgtEndIndex < 0 THEN
    tgtEnd := (tLen - tgtEnd)
  END; (* IF tgtEndIndex < 0 *)
  IF tgtEnd > hiTarget THEN Perror.prerr ( "tgtEndIndex Out of Bounds "); RETURN END;

  srcStart := ABS (srcStartIndex);
  IF srcStartIndex < 0 THEN
    srcStart := (sLen - srcStart)
  END; (* IF srcStartIndex < 0 *)
  IF srcStart > hiTarget THEN Perror.prerr ( "srcStartIndex Out of Bounds "); RETURN END;

  srcEnd := ABS (srcEndIndex);
   IF srcEndIndex < 0 THEN
    srcEnd := (sLen - srcEnd)
  END; (*  IF srcEndIndex < 0 *)
  IF srcEnd > hiTarget THEN Perror.prerr ( "srcEndIndex Out of Bounds "); RETURN END;

  tgtRev := tgtEnd < tgtStart;
  srcRev := srcEnd < srcStart;

  (* we have verified all parameters and that there is something to do, so... *)
  tCount := tgtStart;
  sCount := srcStart;
  (* four cases, depending on whether things are reversed or not *)
  IF tgtRev AND NOT srcRev THEN 
    WHILE  (tCount >= tgtEnd) AND (sCount <= srcEnd) DO
      target [tCount] := source [sCount];
      tCount := tCount + 1;
      sCount := sCount - 1;
    END; (* while *)
  ELSIF NOT tgtRev AND srcRev THEN 
    WHILE  (tCount <= tgtEnd) AND (sCount >= srcEnd) DO
      target [tCount] := source [sCount];
      tCount := tCount + 1;
      sCount := sCount - 1;
    END; (* while *)
  ELSE (* no reversals or both reversed, which is the same thing*)
    WHILE (tCount <= tgtEnd) AND (sCount <= srcEnd) DO
      target [tCount] := source [sCount];
      tCount := tCount + 1;
      sCount := sCount + 1;
    END; (* while *)
  
  END; (*if tgtRev*)

  (* finally check to see if we need to terminate the target *)
  IF ( ( tgtEnd = hiTarget) AND (tgtStart = 0) )
         OR ( ( tgtEnd = 0) AND (tgtStart = hiTarget) ) THEN 
   target [tCount] := ISO646.NUL
  END (* if *)
END CopySlice;


PROCEDURE (*. [INSERT] .*) InsertChars
  ( VAR target : ARRAY OF CHAR; startIndex : CARDINAL; (*. CONST .*) source : ARRAY OF CHAR );
(* Inserts source into destination at startIndex. If the resulting string length exceeds
   the capacity of destination, runtime fault StringCapacityExceeded is raised. *)

VAR
  lenSrc, lenTgt, maxTgt, countS, uCountT, lCountT : CARDINAL; 

BEGIN
  lenSrc :=  LENGTH (source);
  lenTgt := LENGTH (target);
  IF lenSrc = 0 THEN (* empty string, so insert nothing *)
    RETURN
  END; (* if *)
  maxTgt :=  CAPACITYch (target); 

  IF lenSrc + lenTgt + 1 > maxTgt THEN (* room for null *)
    Perror.prerr ( "StringCapacityExceeded error: ");
    RETURN
  ELSIF startIndex > lenTgt  THEN
    Perror.prerr ( "String Index Out Of Bounds error: ");
    RETURN
  END; (* IF lenSrc *)

(* now, move the chars in the destination over far enough to make room for the source chars *)
  IF lenTgt > 0 THEN
    lCountT :=  lenTgt;
    uCountT := lenTgt + lenSrc; (* room for both strings and a null per above check *)
    WHILE lCountT >= startIndex DO
      target[uCountT] := target[lCountT];
      uCountT := uCountT - 1;
      lCountT := lCountT - 1
    END; (* while *)
  END; (* if *)
  lCountT := startIndex;
 
  countS := 0;
  WHILE (countS < lenSrc) DO
    target [lCountT] := source [countS];
    countS := countS + 1;
    lCountT := lCountT + 1
  END; (* while *)
  target [lCountT] :=  ISO646.NUL
END InsertChars;

PROCEDURE (*. [REMOVE] .*) DeleteChars 
  ( VAR target : ARRAY OF CHAR; startIndex, numToDelete : LONGCARD);
(* removes numToDelete characters from target starting at startIndex, moving subsequent chars 
  back to fill up the gap and re-terminating the string. If the sum of startIndex and numToDelete incorporates more
  characters than there are following startIndex, all are deleted and there is no error. *)
VAR
  uCount, lCount, last : LONGCARD;
  
BEGIN
  last := LENGTH (target) - 1;
  lCount := startIndex;
  uCount := startIndex + numToDelete;
  IF uCount > last THEN
    uCount := last + 1;
    lCount := lCount + 1
  END; (* and then the following loop is skipped *)
  WHILE (uCount <= last) DO
    target [lCount] := target [uCount];
    lCount := lCount + 1;
    uCount := uCount + 1
  END; (* while *)
  target [lCount] := ISO646.NUL
END DeleteChars;


PROCEDURE Assign ( VAR destination: ARRAY OF CHAR; source: ARRAY OF CHAR);
  (* Copies source to destination *)

VAR
  count : CARDINAL; (* Counting Variable *)
  maxDest, hiSource : CARDINAL; (* hold high of source and dest for comparisons *)

BEGIN
  IF source [0] =  ISO646.NUL THEN
    destination [0] :=  ISO646.NUL;
    RETURN
  END;(* IF source[0] *)
  hiSource :=  LENGTH (source) - 1;
  maxDest :=  CAPACITYch (destination) - 1; (* room for null *)
  IF hiSource + 1 > maxDest THEN
    Perror.prerr ( "StringCapacityExceeded error: ");
    RETURN
  END; (* IF hiSource *)

  count := 0;
  WHILE (count <= hiSource) AND (source[count] # ISO646.NUL ) DO
    destination[count] := source[count];
    INC(count)
  END; (* WHILE *)
    destination[count] := ISO646.NUL
 END Assign;


PROCEDURE (*. [CONCAT] .*) Concat
  ( VAR result : ARRAY OF CHAR; (*. ARGLIST OF .*) firstItem, secondItem : ARRAY OF CHAR );

VAR
 countD : CARDINAL; 

BEGIN
  countD := LENGTH (firstItem);
  Assign (result, firstItem);
  InsertChars (result, countD, secondItem);
END Concat;


END ARRAYOFCHAR.