(* (C) 2009, 2010 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE Date;

IMPORT TimeUnits;

PROCEDURE isLeapYear (year: Years): BOOLEAN;
(* 45 BC, 42 BC, 39 BC, 36 BC, 33 BC, 30 BC, 27 BC, 24 BC, 21 BC, 18 BC, 
   15 BC, 12 BC, 9 BC, 8 AD, 12 AD, and every fourth year thereafter 
   but in our date system we use 0 for 1BC so these years are as shown below *)
BEGIN
  IF year >= 1582 THEN
    RETURN (year MOD 400 = 0) OR ((year MOD 4 = 0) AND (year MOD 100 # 0))
  ELSIF ( year >= 8 ) AND ( year < 1582 ) THEN
    RETURN (year MOD 4 = 0)
  ELSIF (year < -7) AND (year > -45 ) (* i.e. between 45BC and 9 BC *) THEN
    RETURN ( ABS (year) + 1 ) MOD 3 = 0 (* -44 becomes 45 *)
  ELSE
   RETURN FALSE (* None before 45BC or between 9BC and 8 AD *)
  END (* if *)
END isLeapYear;


PROCEDURE DaysInYear (year: Years): DayNumRange;
BEGIN
  IF isLeapYear (year) THEN
    RETURN 366
  ELSE 
    RETURN 365
  END (* if *)
END DaysInYear;

  
PROCEDURE isValidDate ( calDate : Date ) : BOOLEAN;
(* If the date is valid for days in month etc, returns TRUE, else returnd FALSE 
  Does proleptic Gregorian analysis only *)

BEGIN
  CASE calDate.month OF
      TimeUnits.Jan, TimeUnits.Mar, TimeUnits.May, TimeUnits.Jul, TimeUnits.Aug, TimeUnits.Oct, TimeUnits.Dec:
        RETURN ( calDate.date <= 31 )
    | TimeUnits.Feb:
        IF isLeapYear (calDate.year) THEN
          RETURN ( calDate.date <= 29 )
        ELSE
          RETURN ( calDate.date <= 28 )
        END; (* if *)
    | TimeUnits.Apr, TimeUnits.Jun, TimeUnits.Sep, TimeUnits.Nov:
        RETURN( calDate.date <= 30 )
  END (* case *)
END isValidDate;


END Date.