(* (C) 2009-2018 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE DateStrs;

IMPORT ISO646, AOC, ARRAYOFCHAR, CharArray, DateTime, DateTimeFormats;
IMPORT WholeStrs;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT Date; FROM Date IMPORT DateT;
IMPORT Time; FROM Time IMPORT TimeT;


PROCEDURE DayStrToDayNum (daStr : ARRAY OF CHAR; VAR res : IOStatusT ) : CARDINAL;
(* Takes a string representing a day name--either three letters or the full name, and
  returns the day number from Sun  to Sat. If the string is not a valid day name
  res.code is set to OutOfRange.*)

VAR
  lStr, dNum : CARDINAL;
BEGIN
  dNum := 7;
  res := IOStatusT{FALSE, IOStatus.Success};
  lStr := LENGTH (daStr);
  IF (lStr < 3) OR (lStr > 10) THEN RETURN 0 END;
  daStr[0] := ISO646.toUpper (daStr[0]); (* forgive first letter lowercase *) 
  IF lStr = 3 THEN
    WHILE dNum > 0 DO
      IF CharArray.matches ( daStr, shortDayStrs [dNum] ) THEN
        RETURN  dNum;
      ELSE
        dNum := dNum -1
      END; (* if *)
    END; (* while *)
    RETURN 0 (* if nothing matched *)
  ELSIF  lStr > 3 THEN
    WHILE dNum > 0 DO
      IF  CharArray.matches ( daStr, longDayStrs [dNum] ) THEN
        RETURN dNum;
      ELSE
        dNum := dNum -1
      END; (* if *)
    END; (* while *)
    res := IOStatusT{TRUE, IOStatus.OutOfRange};
    RETURN 0 (* if nothing matched *)
  ELSE
    RETURN 0;
    res := IOStatusT{TRUE, IOStatus.OutOfRange}
  END (* if *)

END DayStrToDayNum;

PROCEDURE DayToDayStr (VAR daStr : ARRAY OF CHAR; day : Date.Days; wStyle : DateTimeFormats.WDayStyle ); 
(* Takes a day from Sun  to Sat . and produces a string. For dNum = Sun; wStyle w ==> "Sun", wStyle W ==> "Sunday". *)

BEGIN
  IF wStyle = DateTimeFormats.w THEN
    ARRAYOFCHAR.Assign (daStr, shortDayStrs [ORD (day)] )
  ELSIF wStyle = DateTimeFormats.W THEN
    ARRAYOFCHAR.Assign (daStr, longDayStrs [ORD (day)] )
  END (* if *)

END DayToDayStr;


PROCEDURE DateStrToDateNum (dStr : ARRAY OF CHAR; VAR res : IOStatusT ) : CARDINAL;
VAR
  dat : CARDINAL;
BEGIN
  WholeStrs.DecStrToCard (dStr, dat, res);
  IF res.code <> IOStatus.Success THEN
    RETURN 0
  ELSIF dat > 31 THEN
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
    RETURN 0
  ELSE
    RETURN dat
  END (* if *)
END DateStrToDateNum;


PROCEDURE DateNumToDateStr (VAR dStr : ARRAY OF CHAR; date : Date.Dates; dStyle : DateTimeFormats.DateStyle ); 
(* Takes a day number from 1 to 31. and produces a string. For dNum = 1; dStyle d ==> 1, dStyle D ==> 01.*)
VAR
  temp : AOC.Len3;
BEGIN
  IF ( date > 10 ) OR ( dStyle = DateTimeFormats.d ) THEN
    WholeStrs.CardToDecStr (date, dStr)
  ELSE
    WholeStrs.CardToDecStr (date, temp);
    dStr [0] := "0";
    dStr [1] := temp [0];
    dStr [3] := ISO646.NUL
  END (* if *)
END DateNumToDateStr;


PROCEDURE MonthStrToMonthNum (mStr : ARRAY OF CHAR; VAR res : IOStatusT ) : CARDINAL;
(* Valid returns are 1..12. A return of 0 is invalid *)

VAR
  lStr, mNum, mCount, sCount : CARDINAL;
  mStr1 : AOC.Len10;

BEGIN
  mCount := 0;
  sCount := 0;
  WHILE DateTimeFormats.isSeparatorChar (mStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)
  IF ISO646.isLetter ( mStr [sCount ] ) THEN
    WHILE ISO646.isLetter ( mStr [sCount ] ) DO
      mStr1[mCount] := mStr [sCount];
      mCount := mCount + 1;
      sCount := sCount + 1
    END; (* while *)
    mStr1[mCount] := ISO646.NUL; 
    mNum := 12;
    lStr := LENGTH (mStr1);
    IF (lStr < 3) OR (lStr > 10) THEN
      res := IOStatusT{TRUE, IOStatus.OutOfRange};
      RETURN 0
    END; (* IF (lStr < 3) *)
    mStr1[0] := ISO646.toUpper (mStr1[0]); (* forgive first letter lowercase *) 
    IF lStr = 3 THEN
      WHILE mNum > 0 DO
        IF CharArray.matches ( mStr1, shortMonthStrs [mNum] ) THEN
          res := IOStatusT{FALSE, IOStatus.Success};
          RETURN  mNum;
        ELSE
          mNum := mNum -1
        END; (* if *)
      END; (* while *)
      res := IOStatusT{TRUE, IOStatus.OutOfRange};
      RETURN 0 (* if nothing matched *)
    ELSIF  lStr > 3 THEN
      WHILE mNum > 0 DO
        IF CharArray.matches ( mStr1, longMonthStrs [mNum] ) THEN
          res := IOStatusT{FALSE, IOStatus.Success};
          RETURN mNum;
        ELSE
          mNum := mNum -1;
          res := IOStatusT{TRUE, IOStatus.OutOfRange}
        END; (* if CharArray *)
      END; (* while mNum *)
      res := IOStatusT{TRUE, IOStatus.WrongFormat};
      RETURN 0 (* if nothing matched *)
    ELSE
      res := IOStatusT{TRUE, IOStatus.WrongFormat};
      RETURN 0
    END (* if lStr = 3*)

  ELSE (* it's a string of digits *)
    WHILE ISO646.isDigit (mStr [sCount]) DO
      mStr1[mCount] := mStr [sCount];
      mCount := mCount + 1;
      sCount := sCount + 1
    END; (* while *)
    mStr1[mCount] := ISO646.NUL;
    WholeStrs.StrToCard (mStr1, mNum, res);
    IF mNum > 0 THEN  
      res := IOStatusT{TRUE, IOStatus.OutOfRange};
      RETURN mNum
    ELSE
      res := IOStatusT{TRUE, IOStatus.WrongFormat};
      RETURN 0
    END; (* IF mon > 0 *)
  END (* main if *)
END MonthStrToMonthNum;


PROCEDURE MonthNumToMonthStr (VAR mStr : ARRAY OF CHAR; month : Date.Months; mStyle : DateTimeFormats. MonthStyle ); 
(* Takes a month number in the range Jan  to Dec. and produces a string.
  For mNum = Jan; mStyle m ==> "Jan", M  ==> "January", n ==> "1", N) ==> "01" *)
VAR
  mNum : CARDINAL;
BEGIN
  mNum := ORD (month) + 1;
  IF mStyle = DateTimeFormats.m THEN
     ARRAYOFCHAR.Assign (mStr, shortMonthStrs [ORD (mNum)] )
  ELSIF mStyle = DateTimeFormats.M THEN
     ARRAYOFCHAR.Assign(mStr, longMonthStrs [ORD (mNum)] )
  ELSIF (mStyle = DateTimeFormats.n) AND (mNum < 10) THEN
    mStr [0] := CHR (ORD("0") +  mNum );
    mStr [1] := ISO646.NUL
  ELSE
    mStr [0] := "1";
    mStr [1] := CHR (ORD("0") + mNum MOD 10 );
    mStr [2] := ISO646.NUL
  END (* if *)
END MonthNumToMonthStr;


PROCEDURE YearStrToYearNum (yStr : ARRAY OF CHAR; VAR res : IOStatusT ) : Date.Years;
(* Takes a string representing a year, and returns the year number.
 If the string is not a valid year number it returns 0. A two letter digit string is
 not interpreted in the range 1950-2049 but as a two digit year. *)
VAR
  int : INTEGER;

BEGIN
  WholeStrs.StrToInt (yStr, int, res);
  IF (res.failed) OR ( int > MAX (Date.Years) ) OR ( int < MIN (Date.Years) ) THEN
    int := 0;
    res := IOStatusT{TRUE, IOStatus.OutOfRange} 
  ELSE
    res := IOStatusT{FALSE, IOStatus.Success};
    RETURN VAL (Date.Years, int)
  END; (* if *)
END YearStrToYearNum;


PROCEDURE YearNumToYearStr (VAR yStr : ARRAY OF CHAR; year : Date.Years; yStyle : DateTimeFormats. YearStyle ); 
(* Takes a year number in the range DateTime.Years. and produces a string.
  For yStyle y ==> 1-2 digits Y  ==> all digits *)

BEGIN
  IF yStyle = DateTimeFormats.y THEN
    IF year > 0 THEN
      year := year MOD 100
    ELSIF year < 0 THEN
      year := - (ABS (year) MOD 100)
    END (* if *)
  END; (* if *)
  WholeStrs.IntToDecStr (year, yStr)
END YearNumToYearStr;


END DateStrs.