(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE DateTimeArrayIO (*[ProtoArrayIO] FOR DateTime*);

  (*
    Implementation and Revisions:
    ============================

    Author  Rick Sutcliffe        Description of change
    ------  ----        ---------------------
    RS    2018 03 15  First implementation *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYao, CAPACITYn;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenIO, FormatIO;
IMPORT  WholeStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ISO646, ARRAYOFCHAR, CharArray;

(* R10 Stuff with mods *)
IMPORT ChanIO;

(* Type specific *)
IMPORT DateTimeBaseIO, DateTimeFormats;
IMPORT DateTime; FROM DateTime IMPORT DateTimeT;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  CustomType = DateTimeT;

CONST
  DateToStr = DateTimeFormats.DateToString;
  TimeToStr = DateTimeFormats.TimeToString;
  SetCustomScanSet = FormatIO.SetCustomScanSet;
  SetValidCustomFmtChars = FormatIO.SetValidCustomFmtChars;
  Ts = DateTimeFormats.TSeparator;
  ParseFmtStrToCdfr = DateTimeFormats.ParseFmtStrToCdfr;
  ParseFmtStrToTfr = DateTimeFormats.ParseFmtStrToTfr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : ChanIO.Channel; VAR array : ARRAY OF CustomType );

VAR
  count, last : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  last := CAPACITYn (array, SIZE (CustomType) ) - 1;
  WHILE count <= last DO
    DateTimeBaseIO.Read (infile, array [count] );
    status := ChanIO.statusOf (infile);
    count := count + 1;
    IF NOT status.failed THEN RETURN END
  END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : ChanIO.Channel; (*CONST*) array : ARRAY OF CustomType );
VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYn (array, SIZE (CustomType) ) - 1;
  (* open brace for the array *)
  CharStrIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <= last DO
    DateTimeBaseIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF CustomType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)

VAR
  temp, temp2 : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : CustomType;
  dfr : DateTimeFormats.CalDateFmtRec;
  tfr : DateTimeFormats.TimeFmtRec;

BEGIN
  numItems := CAPACITYn (items, SIZE (CustomType) ) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := items[count];
    temp[0] := ISO646.NUL;

    (* convert to the appropriate string *)
    ParseFmtStrToCdfr ( dfr, fsr.customFormatStr, res);
    DateTimeFormats.DateToString (temp, nc.calDate, dfr);
    ParseFmtStrToTfr ( tfr, fsr.customFormatStr, res);
    DateTimeFormats.TimeToString (temp2, nc.time, tfr);
    ARRAYOFCHAR.Concat (temp, temp, " T");
    ARRAYOFCHAR.Concat (temp, temp, temp2);

    (* format that converted string using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

       (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharBaseIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;

PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF CustomType );
VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   numItems := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  SetCustomScanSet (FormatIO.DefaultCustomScanSet);
  SetValidCustomFmtChars (FormatIO.DefaultValidCustomFmtChars);
  ScanStrForFormat (fmtStr, FormatIO.GetCustomScanSet(), FormatIO.GetValidCustomFmtChars(), fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= numItems DO
    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND ((count # numItems) OR NOT lastArray) THEN
      CharStrIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END; (* if *)
    count := count + 1
  END; (* while *)
  (* and if an escaped char is in the fsr, write it *)
  IF lastArray THEN CharStrIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *) END
END WriteFArray;


TYPE
  Ary1 = ARRAY [0..0] OF CustomType;
  Ary2 = ARRAY [0..1] OF CustomType;
  Ary3 = ARRAY [0..2] OF CustomType;


VAR
  lastArray : BOOLEAN; (* this is a kludge anyway so use a global variable to control the following *)


PROCEDURE Write1FArray
  ( outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; value1 : (*. ARGLIST .*) ARRAY OF CustomType );
BEGIN
    WriteFArray (outfile, fmtStr,value1 )
END Write1FArray;


PROCEDURE Write2FArray
  ( outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; value1, value2 : (*. ARGLIST .*) ARRAY OF CustomType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value2 )
END Write2FArray;


PROCEDURE Write3FArray
  ( outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; value1, value2, value3 : (*. ARGLIST .*) ARRAY OF CustomType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  WriteFArray (outfile, fmtStr, value2 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value3 )
END Write3FArray;

BEGIN (* module *)
  lastArray := TRUE;

END DateTimeArrayIO.