(* (C) 2009-2013 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE DateTimeBaseIO ;
FROM IOSupport IMPORT*;

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYn;
IMPORT ChanIO;
IMPORT DateTimeFormats, DateStrs, TimeUnits, DateBaseIO, TimeBaseIO;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT DateTime; FROM DateTime IMPORT DateTimeT;
IMPORT Date; FROM Date IMPORT DateT;
IMPORT CharStrIO, ARRAYOFCHAR, CharArray, FormatIO, WholeStrs;
IMPORT AOC, ISO646;

TYPE
  CalDateFmtRec = (* ALIAS *) DateTimeFormats.CalDateFmtRec;
  Order = (* ALIAS *) DateTimeFormats.Order;
  FormatScanRec = (* ALIAS *) FormatIO.FormatScanRec;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatTextStr = FormatIO.FormatTextStr;
  SetCustomScanSet = FormatIO.SetCustomScanSet;
  SetValidCustomFmtChars = FormatIO.SetValidCustomFmtChars;
  Ts = DateTimeFormats.TSeparator;
  ParseFmtStrToCdfr = DateTimeFormats.ParseFmtStrToCdfr;
  ParseFmtStrToTfr = DateTimeFormats.ParseFmtStrToTfr;
  
  
VAR
  gdtr : CalDateFmtRec;

PROCEDURE (*[READ]*) Read( inChan : ChanIO.Channel; VAR dt : DateTimeT );

VAR
  inputStr, dtStr, tStr : AOC.Len512;
  iCount, dCount, tCount, len : CARDINAL;
  res : IOStatusT;
  validStatus : BOOLEAN;

BEGIN  (* main proc *)
  CharIO.ReadStr (inChan, inputStr);
  iCount:= 0;
  dCount := 0;
  tCount := 0;
  len := LENGTH (inputStr);
  
 (* split the input string in two *)
  WHILE (inputStr [iCount] <> Ts) AND (iCount < len ) DO
    dtStr [dCount] := inputStr [iCount];
    iCount := iCount + 1; 
    dCount := dCount + 1;
  END; (* while *)
  dtStr [dCount] := ISO646.NUL;
  
    (* Skip over the parts separator *)
  IF inputStr [iCount] = Ts THEN
    iCount := iCount + 1
  END; (* if *)
  WHILE iCount < len DO
    tStr [tCount] := inputStr [iCount];
    iCount := iCount + 1; 
    tCount := tCount + 1;
  END; (* while *)  sCharIO.WriteStr (dtStr);sCharIO.WriteStr ("   ");sCharIO.WriteStr (tStr);

(* now convert the two strings into the subrecords *)
  DateTimeFormats.StringToDate ( dt.calDate, dtStr, res );
  IF NOT res.failed THEN
    DateTimeFormats.StringToTime ( dt.time, tStr, res );
  END;
  IF res.failed OR (NOT Date.isValidDate (dt.calDate)) THEN
      ChanIO.SetStatus (inChan, IOStatusT { TRUE, IOStatus.WrongFormat }, validStatus )
  END (* if *)
END Read;


PROCEDURE (*[WRITE]*) Write( outChan : ChanIO.Channel; dt : DateTimeT );
(* Writes the textual representation of value dt to stream outChan. This
   procedure is substituted for invocations of WRITE with a DateTime argument.*)
BEGIN
  DateBaseIO.Write (outChan, dt.calDate);
  CharIO.WriteStr (outChan, " T");
  TimeBaseIO.Write  (outChan, dt.time);
END Write;

TYPE
  CustomType = DateTimeT;

PROCEDURE fetchStrFromFmtStr (str : ARRAY OF CHAR; VAR index : FormatIO.StrRange) : FormatIO.Str;

VAR
  fCount, sLen : FormatIO.StrRange;
  theStr : FormatIO.Str;

BEGIN (* fetchStrFromFmtStr *)
  fCount := 0;
  sLen := LENGTH (str);
  WHILE  (index <  sLen) AND NOT ( (str [index]) = ISO646.COMMA ) DO (* copy the chars *)
    theStr [fCount] := str [index];
    fCount := fCount + 1;
    index := index + 1
  END (* while *);
  IF fCount # 0 THEN
    theStr [fCount] := CHR (0);
  END; (* if *)
  RETURN theStr
END fetchStrFromFmtStr;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items : ARRAY OF CustomType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp, temp2 : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : CustomType;
  dfr : DateTimeFormats.CalDateFmtRec;
  tfr : DateTimeFormats.TimeFmtRec;

BEGIN
  numItems := CAPACITYn (items, SYSTEM.TSIZE (CustomType)) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := items[count];
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    ParseFmtStrToCdfr ( dfr, fsr.customFormatStr, res);
    DateTimeFormats.DateToString (temp, nc.calDate, dfr);
    ParseFmtStrToTfr ( tfr, fsr.customFormatStr, res);
    DateTimeFormats.TimeToString (temp2, nc.time, tfr);
    ARRAYOFCHAR.Concat (temp, temp, " T");
    ARRAYOFCHAR.Concat (temp, temp, temp2);

    (* format that converted string using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

       (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END (* if *);

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END (* if *);

    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST*) ARRAY OF DateTimeT );
(* Writes a formatted textual representation of one or more TZ values to
   output stream outChan. The output format is determined by fmtStr. This
   procedure is substituted for invocations of WRITEF with one or more
   DateTime arguments. *)

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  SetCustomScanSet (FormatIO.DefaultCustomScanSet);
  SetValidCustomFmtChars (FormatIO.DefaultValidCustomFmtChars);
  ScanStrForFormat (fmtStr, FormatIO.GetCustomScanSet(), FormatIO.GetValidCustomFmtChars(), fsr);
  WriteFWithfsr ( outfile, fsr , items  )
END WriteF;


BEGIN  (* init *)
  gdtr := DateTimeFormats.GetStdCalDateFmtRec (); (* initial value; subject to change *)
END DateTimeBaseIO.