(* (C) 2009, 2010 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE DateTimeFormats (*[DateType]*);
FROM IOSupport IMPORT*;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT DateTime; FROM DateTime IMPORT DateTimeT;
IMPORT Date; FROM Date IMPORT DateT;
IMPORT Time; FROM Time IMPORT TimeT;
IMPORT DateTimeUtilities, TZ, FormatIO;
IMPORT AOC, ISO646, CharArray, WholeStrs, DateStrs;
IMPORT ARRAYOFCHAR; FROM ARRAYOFCHAR IMPORT CopySlice, Assign, Concat;
FROM Transition IMPORT CAPACITYch;

CONST
  MonthStrToMonthNum = DateStrs.MonthStrToMonthNum;
  YearStrToYearNum = DateStrs.YearStrToYearNum;
  DateStrToDateNum = DateStrs.DateStrToDateNum;
  Justify = FormatIO.JustifyA;
  Right = FormatIO.Right;
  Zero = ISO646.DIGIT_ZERO;

VAR
  gCalDateFmtRec : CalDateFmtRec;
  gTimeFmtRec : TimeFmtRec;
  AlwaysAD, StrictISOTime : BOOLEAN;
  

PROCEDURE isSeparatorChar (ch : CHAR) : BOOLEAN;
BEGIN
  RETURN ch IN Separators
END isSeparatorChar;

PROCEDURE ParseFmtStrToCdfr (VAR dfr : CalDateFmtRec; fmtStr : ARRAY OF CHAR; res : IOStatusT);
(* takes a string consisting of letters in the styles above and creates from them a Date Format
Record. The styles may be in any order. A second style trigger for the same field
overrides the first. If the format string is valid, <res> is set to {FALSE, Success
Otherwise, <res> is set to {TRUE, InvalidWriteFString} *)
VAR
  count, len, last : CARDINAL;
BEGIN
  len := LENGTH (fmtStr);
  IF len = 0 THEN RETURN END;
  count := 0;
  last := len - 1;
  res := IOStatusT{TRUE, IOStatus.InvalidWriteFString};
  dfr := DefaultCalDateFmtRec;
  WHILE count <= last DO
    CASE  fmtStr [count] OF
      (* OrderEndianStyle *)
        "L" :
          dfr.order := dmy
      | "X" :
          dfr.order := mdy
      | "B" :
          dfr.order := ymd

      (* YearStyle *)  
      | "y" :
          dfr.yStyle := y
      | "Y" :
          dfr.yStyle := Y

      (* MonthStyle *)  
      | "m" :
          dfr.mStyle := m
      | "M" :
          dfr.mStyle := M
      | "n" :
          dfr.mStyle := n
      | "N" :
          dfr.mStyle := N

      (* DateStyle *)  
      | "d" :
          dfr.dStyle := d
      | "D" :
          dfr.dStyle := D

      (* WeekDayStyle *)  
      | "w" :
          dfr.wStyle := w
      | "W" :
          dfr.wStyle := W
      | "k" :
          dfr.wStyle := k
      | "K" :
          dfr.wStyle := K
      | "O" :
          dfr.wStyle := O        
    ELSE
(*
      Ignore all else. May be valid for times
*)
    END; (* case *)
    count := count + 1;
  END (* while *)
END ParseFmtStrToCdfr;


PROCEDURE ParseFmtStrToTfr (VAR tfr : TimeFmtRec; fmtStr : ARRAY OF CHAR; res : IOStatusT);
(* takes a string consisting of letters in the styles above and creates from them a Date Format
Record. The styles may be in any order. A second style trigger for the same field
overrides the first. If the format string is valid, <res> is set to {FALSE, Success
Otherwise, <res> is set to {TRUE, InvalidWriteFString} *)
VAR
  count, last : CARDINAL;
BEGIN
  count := 0;
  last := LENGTH (fmtStr) - 1;
  res := IOStatusT{FALSE, IOStatus.Success};
  tfr := DefaultTimeFmtRec;
  WHILE count <= last DO
    CASE  fmtStr [count] OF
      (* Hour Style *)  
      | "h" :
          tfr.hStyle := h
      | "H" :
          tfr.hStyle := H
      (* Second Style *)  
      | "s" :
          tfr.sStyle := s
      | "S" :
          tfr.sStyle := S
      (* TZStyle *)  
      | "a", "A" :
          tfr.tzStyle := a
      | "u", "U" :
          tfr.tzStyle := u
    ELSE
(*
      We ignore everything else. If the format string is used as well for dates, fine
*)
    END; (* case *)
    count := count + 1;
  END (* while *)
END ParseFmtStrToTfr;

PROCEDURE DateToString (VAR str: ARRAY OF CHAR; dt : DateT; dfr : CalDateFmtRec);
(* Given a DateTime and a CalDateFmtRec, formats the date and time as a string for output. 
   PRE: years Y  before 1AD are expressed in -(Y-1) format as Julian day years. *)

VAR
  sCount, max, Wn: CARDINAL;
  yr : Date.Years;
  wn : Date.WeekRange;
  wd : Date.ISODayRange;
  yStr : AOC.Len15;
  dStr, dStr1, oStr : AOC.Len5;
  mStr, mStr1, wStr : AOC.Len10;
  sep, wSep : CHAR;
  BC : BOOLEAN;
  
BEGIN
  str[0] := ISO646.NUL;
  yStr[0] := ISO646.NUL;
  dStr1[0] := ISO646.NUL;

(* year *)
  BC := dt.year < 1;
  IF BC THEN dt.year := - (dt.year - 1) END;
  IF dfr.yStyle = _y THEN
    yStr[0] := ISO646.NDASH;
    yStr[1] := ISO646.NDASH;
    yStr[2] := ISO646.NUL;
  ELSE
    IF (dfr.yStyle = y) THEN
      yr := dt.year MOD 100
    ELSE 
      yr := dt.year
    END;
    WholeStrs.CardToDecStr (yr, yStr);
    IF (dfr.yStyle = y) THEN
      Justify (yStr, yStr, Zero, 2, Right);
    ELSE
      Justify (yStr, yStr, Zero, 4, Right);
    END; (* if *)
    IF BC THEN
      Concat (yStr, yStr, BCStr)
    ELSIF AlwaysAD THEN
      Concat (yStr, yStr, ADStr)
    END
  END;

IF ( dfr.wStyle = k ) OR ( dfr.wStyle = K ) THEN
  DateTimeUtilities.ISOWeekNumInYear (dt, yr, wn, wd)
END;  (*if*)
(*  weekday *)  
  wStr [0] := ISO646.NUL;
  CASE dfr.wStyle OF
     _w :
    | w :
        Assign (wStr, DateStrs.shortDayStrs [ORD ( DateTimeUtilities.WeekDayValue (dt) ) + 1] )
    | W :
        Assign (wStr, DateStrs.longDayStrs [ORD ( DateTimeUtilities.WeekDayValue (dt) ) + 1] )
    | k,K :
        WholeStrs.CardToDecStr (yr, yStr);
        Justify (yStr, yStr, Zero, 4, Right);
        Concat (str, yStr, "-W");
        WholeStrs.CardToDecStr (wn, wStr);
        Justify (wStr, wStr, Zero, 2, Right);
        Concat (str, str, wStr);
        IF dfr.wStyle = K THEN
          WholeStrs.CardToDecStr (wd, dStr1);
          Concat (str, str, "-");
          Concat (str, str, dStr1);
        END; (* if *)
        RETURN
    | O :
        WholeStrs.CardToDecStr (DateTimeUtilities.DayNumInYear (dt), oStr);
        WholeStrs.CardToDecStr (dt.year, str);
        Justify (yStr, yStr, Zero, 4, Right);
        sCount := LENGTH (str);
        str[sCount] := "-";
        sCount := sCount + 1;
        Justify (yStr, yStr, Zero, 3, Right);
        str[sCount] := ISO646.NUL;
        Concat (str, str, oStr);
        RETURN 
    ELSE
  END; (* case *)

  (* month *)  
  CASE dfr.mStyle OF
      m :
        Assign (mStr, DateStrs.shortMonthStrs [ORD (dt.month) + 1] )
    | M :
        Assign (mStr, DateStrs.longMonthStrs [ORD (dt.month) + 1] )
    | n :
        WholeStrs.CardToDecStr ( ORD (dt.month) + 1, mStr )
    | N :
        WholeStrs.CardToDecStr ( ORD (dt.month) + 1, mStr1);
        IF dt.month < Date.Oct THEN
          mStr[0] := "0";
          mStr[1] := mStr1[0]
        ELSE
          mStr[0] := mStr1[0];
          mStr[1] := mStr1[1]
        END; (* if *)
        mStr[2] := ISO646.NUL
    ELSE
  END; (* case *)

(* date *)
  IF (dfr.dStyle = d) THEN
    WholeStrs.CardToDecStr ( ORD (dt.date), dStr )
  ELSE
    WholeStrs.CardToDecStr ( ORD (dt.date), dStr1 );
    IF dt.date < 10 THEN
      dStr[0] := "0";
      dStr[1] := dStr1[0]
    ELSE
      dStr[0] := dStr1[0];
      dStr[1] := dStr1[1]
    END; (* if *)
    dStr[2] := ISO646.NUL
  END; (* if *)

(* now put it together *)
  sCount := 0;
  max := CAPACITYch (str);
  sep := GetCalDateSeparator();
  wSep :=  GetWeekdaySeparator();

  Assign(str, wStr);
  sCount := LENGTH (str);
  IF sCount > 0 THEN 
    IF wSep <> sep THEN
      str [ sCount ]:= wSep;
      sCount := sCount + 1
    END; (* if *) 
    str [ sCount ] := ISO646.SPACE;
    sCount := sCount + 1
  END; (* if *)

  CASE dfr.order OF
    dmy:
      Concat (str, str, dStr);
      sCount := sCount + LENGTH (dStr);
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, mStr);
      sCount := sCount + LENGTH (mStr);
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, yStr);
      sCount := sCount + LENGTH (yStr);
      str[sCount] := ISO646.NUL
  | mdy:
      Concat (str, str, mStr);;
      sCount := sCount + LENGTH (mStr);
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, dStr);
      sCount := sCount + LENGTH (dStr);
      IF (sep = ISO646.SPACE) AND ( (dfr.mStyle = M) OR (dfr.mStyle = m) ) THEN
        str[sCount] := ISO646.COMMA;
        sCount := sCount + 1;
      END; (* if *)
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, yStr);
      sCount := sCount + LENGTH (yStr);
      str[sCount] := ISO646.NUL

  | ymd:
      Concat (str, str, yStr);
      sCount := sCount + LENGTH (yStr);
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, mStr);
      sCount := sCount + LENGTH (mStr);
      str[sCount] := sep;
      sCount := sCount + 1;
      str[sCount] := ISO646.NUL;
      Concat (str, str, dStr);
      sCount := sCount + LENGTH (dStr);
      str[sCount] := ISO646.NUL
  END; (* case *)

END DateToString;

PROCEDURE StringToDate (VAR dt : DateT; inputStr: ARRAY OF CHAR; res : IOStatusT);
(* Given a Sting and a CalDateFmtRec, sets up the CalDate Record *)

VAR
  sCount : CARDINAL;

(* internal procedures *)
PROCEDURE ParseYear;
VAR
  yStr : AOC.Len6;
  yCount : CARDINAL;
  int : INTEGER;
  
BEGIN
  yCount := 0;
  IF (inputStr [sCount] = "-") AND (inputStr [sCount+ 1] = "-") THEN 
    dt.year:= DateTimeUtilities.ThisYear();
    RETURN
  END;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)  
  WHILE (inputStr [sCount]) = "0" DO
    sCount := sCount + 1
  END;

  WHILE ISO646.isDigit (inputStr [sCount]) DO
    yStr[yCount] := inputStr [sCount];
    yCount := yCount + 1;
    sCount := sCount + 1
  END; (* while *)
  yStr[yCount] := ISO646.NUL;
  dt.year := YearStrToYearNum (yStr, res);
END ParseYear;

PROCEDURE ParseMonth;
VAR
  mStr : AOC.Len10;
  mCount, mon : CARDINAL;
BEGIN
  mCount := 0;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1;
  END; (* while *)
  
  (* now either parse off letters or numbers *)
  IF ISO646.isLetter ( inputStr [sCount] ) THEN
    WHILE ISO646.isLetter ( inputStr [sCount] ) DO
      mStr[mCount] := inputStr [sCount];
      mCount := mCount + 1;
      sCount := sCount + 1
    END; (* while *) 
  ELSIF ISO646.isDigit ( inputStr [sCount] )THEN
    WHILE ISO646.isDigit ( inputStr [sCount] ) DO
      mStr[mCount] := inputStr [sCount];
      mCount := mCount + 1;
      sCount := sCount + 1
    END; (* while *)
  END; (* if *) 
  mStr[mCount] := ISO646.NUL;
  mon := MonthStrToMonthNum (mStr, res);
  IF (mon > 0) AND (mon < 14) THEN 
    dt.month := VAL (Date.Months, mon - 1)
  ELSE
    dt.month := VAL (Date.Months, 1);
    res := IOStatusT { TRUE, IOStatus.OutOfRange };
  END (* IF mon > 0 *)
END ParseMonth;

PROCEDURE ParseDate;
VAR
  dStr : AOC.Len10;
  dCount, dat : CARDINAL;
BEGIN
  dCount := 0;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)
  WHILE (inputStr [sCount]) = "0" DO
    sCount := sCount + 1
  END; (* while *)

  WHILE ISO646.isDigit (inputStr [sCount]) DO
    dStr[dCount] := inputStr [sCount];
    dCount := dCount + 1;
    sCount := sCount + 1
  END; (* while *)
  dStr[dCount] := ISO646.NUL;
  dat := DateStrToDateNum (dStr, res);
  IF (dat < 32) AND (dat > 0) AND (res.code = IOStatus.Success) THEN
    dt.date := VAL (Date.Dates, dat);
  ELSE
    dt.date := VAL (Date.Dates, 1);
    res := IOStatusT { TRUE, IOStatus.OutOfRange };
  END (* if *) 
END ParseDate;

PROCEDURE ParseDayNum;
VAR
  dStr : AOC.Len10;
  dCount, DNum : CARDINAL;
BEGIN
  dCount := 0;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)
  WHILE (inputStr [sCount]) = "0" DO
    sCount := sCount + 1
  END; (* while *)

  WHILE ISO646.isDigit (inputStr [sCount]) DO
    dStr[dCount] := inputStr [sCount];
    dCount := dCount + 1;
    sCount := sCount + 1
  END; (* while *)
  dStr[dCount] := ISO646.NUL;
  WholeStrs.DecStrToCard (dStr, DNum, res);
  IF (DNum > 0) AND (DNum < 367) THEN
     DaNumInYear := VAL (Date.DayNumRange, DNum)
  ELSE 
    res := IOStatusT { TRUE, IOStatus.OutOfRange };
    DaNumInYear := 1
  END (* if *)
END ParseDayNum;

PROCEDURE ParseISOWeek;
VAR
  wStr : AOC.Len10;
  wCount, WNum : CARDINAL;
BEGIN
  wCount := 0;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)
  WHILE (inputStr [sCount]) = "0" DO
    sCount := sCount + 1
  END; (* while *)

  WHILE ISO646.isDigit (inputStr [sCount]) DO
    wStr[wCount] := inputStr [sCount];
    wCount := wCount + 1;
    sCount := sCount + 1
  END; (* while *)
  wStr[wCount] := ISO646.NUL;
  WholeStrs.DecStrToCard (wStr, WNum, res);
  IF (WNum > 0) AND (WNum < 54) THEN
     WNumInYear := VAL (Date.WeekRange, WNum)
  ELSE 
    res := IOStatusT { TRUE, IOStatus.OutOfRange };
    WNumInYear := 1
  END (* if *)
END ParseISOWeek;

PROCEDURE ParseISOWeekDayNum;
VAR
  dStr : AOC.Len10;
  dCount, DNum : CARDINAL;
BEGIN
  dCount := 0;
  WHILE isSeparatorChar (inputStr [sCount]) DO
    sCount := sCount + 1
  END; (* while *)
  WHILE (inputStr [sCount]) = "0" DO
    sCount := sCount + 1
  END; (* while *)

  WHILE ISO646.isDigit (inputStr [sCount]) DO
    dStr[dCount] := inputStr [sCount];
    dCount := dCount + 1;
    sCount := sCount + 1
  END; (* while *)
  dStr[dCount] := ISO646.NUL;
  WholeStrs.DecStrToCard (dStr, DNum, res);
  IF (DNum > 0) AND (DNum < 8) THEN
     DaNumInWeek := VAL (Date.ISODayRange, DNum)
  ELSE 
    res := IOStatusT { TRUE, IOStatus.OutOfRange };
    DaNumInWeek := 1
  END (* if *)
END ParseISOWeekDayNum;


(* set out global variables to the main procedure *)
VAR
  dateOrder : Order;
  validStatus : BOOLEAN;
  DaNumInYear : Date.DayNumRange;
  WNumInYear  : Date.WeekRange;
  DaNumInWeek, Jan1ISODay : Date.ISODayRange;
  Jan1DayNum, WeekAdj : CARDINAL;
  Jdn, DayAdj : LONGREAL;

BEGIN  (* main proc *)
  (* try to do some intelligence about what we get *)
  IF ISO646.isLetter ( inputStr [0] ) THEN
    dateOrder := mdy
  ELSIF ISO646.isDigit ( inputStr [0] ) AND (ISO646.isLetter (inputStr [1] ) ) OR ( ISO646.isLetter (inputStr [2] ) ) THEN
    dateOrder := dmy
  ELSIF ( inputStr [0] > "3")
      OR ( ( inputStr [0] = "3") AND ( inputStr [1] > "1") )
      OR ( ( ISO646.isDigit (inputStr [0]) ) AND ( ISO646.isDigit (inputStr [1]) ) AND ( ISO646.isDigit (inputStr [2]) ) )THEN 
    dateOrder := ymd
  ELSE
    dateOrder := GetOrder();
  END;

  (* start off assuming success. One of the subprocedures may change this *)
  res := IOStatusT { FALSE, IOStatus.Success };

  sCount := 0;
  CASE dateOrder OF
      dmy:
        ParseDate; ParseMonth; ParseYear
    | mdy:
        ParseMonth; ParseDate; ParseYear
    | ymd:  (* actually anything starting with year including ISO yyyy-### and yyy-W##-# *)
        ParseYear;
        (* first check to see if this is in the ISO YYYY-DDD format *)
        IF ( isSeparatorChar (inputStr [sCount]) )
          AND ( CharArray.nextIndexOfChar ( inputStr, sCount + 1, inputStr [sCount] )  = LENGTH (inputStr) ) (* no more of this separator *) THEN
             ParseDayNum;
          Jdn := DateTimeUtilities.JDNFromYearAndDayNum (dt.year, DaNumInYear);
          dt := DateTimeUtilities.JDNToCalendar ( Jdn );
        (* if not is it in the ISO YYYY-Wxx-x format *)
        ELSIF (inputStr [sCount + 1] = "W") OR (inputStr [sCount + 1] = "w") THEN (* being a bit permissive on reads here *)
          sCount := sCount + 2; (* skip separator and "W" *)
          
          (* get the two values after the "W" *)
          ParseISOWeek; ParseISOWeekDayNum; 
          
          (* find the ISO day of the week for January 1 *)
          Jan1DayNum := DateTimeUtilities.WeekDayNum( DateT{dt.year, Date.Jan, 1});
          IF Jan1DayNum = 0 THEN
            Jan1ISODay := 7
          ELSE
            Jan1ISODay := Jan1DayNum
          END; 

          (* start with the JDN of Jan1  and subtract to get the JDN of the Monday before--could be last year *)
          Jdn := DateTimeUtilities.JDNFromYearAndDayNum (dt.year, 1);
          Jdn := Jdn - VAL (LONGREAL, Jan1ISODay - 1);
          
          (* now, to this add the number of days in the weeks and weekday going forward *)
          WeekAdj := WNumInYear;
          IF NOT (DateTimeUtilities.isShortFirstWeek (dt.year) ) THEN
            WeekAdj := WeekAdj - 1;
          END; (* if *)
          Jdn := Jdn + VAL (LONGREAL, 7 * ( WeekAdj ) + DaNumInWeek - 1 );
          dt := DateTimeUtilities.JDNToCalendar ( Jdn )
        ELSE
          (* only possibility left is the yyyy-mm-dd format *)
          ParseMonth; ParseDate;
        END; (* if *)
  END; (* case *)
   (* but it could be messed up so *)
  IF NOT Date.isValidDate (dt) THEN
    res := IOStatusT { TRUE, IOStatus.WrongFormat }
  END (* if *)

END StringToDate;


(* ================================= *)

(* The following procedures operate on the stored standard date/time records 
  that are used to parse strings being READ, by WRITE, and by WRITEF if an
  empty custom format string is passed. *)

PROCEDURE SetOrder ( order :Order );

BEGIN
  gCalDateFmtRec.order := order; 
END SetOrder;

PROCEDURE GetOrder () : Order;

BEGIN
  RETURN gCalDateFmtRec.order;
END GetOrder;

PROCEDURE SetCalDateSeparator ( ch : CHAR );

BEGIN
  gCalDateFmtRec.separator := ch 
END SetCalDateSeparator;


PROCEDURE GetCalDateSeparator () : CHAR;

BEGIN
  RETURN gCalDateFmtRec.separator;
END GetCalDateSeparator;


PROCEDURE SetWeekdaySeparator ( ch : CHAR );
(* changes the separator between weekday and date for Write *)

BEGIN
  gCalDateFmtRec.wSeparator := ch 
END SetWeekdaySeparator;


PROCEDURE GetWeekdaySeparator () : CHAR;
(* Returns the saved separator between weekday and date for Write *)  

BEGIN
  RETURN gCalDateFmtRec.wSeparator;
END GetWeekdaySeparator;


PROCEDURE SetYearStyle ( ys : YearStyle );
(* changes the year style for Write  *)
BEGIN
  gCalDateFmtRec.yStyle := ys
END SetYearStyle;


PROCEDURE GetYearStyle () : YearStyle;

BEGIN
  RETURN gCalDateFmtRec.yStyle;
END GetYearStyle;


PROCEDURE SetMonthStyle ( ms : MonthStyle );
(* changes the month style for Write  *)
BEGIN
  gCalDateFmtRec.mStyle := ms
END SetMonthStyle;


PROCEDURE GetMonthStyle () : MonthStyle;

BEGIN
  RETURN gCalDateFmtRec.mStyle;
END GetMonthStyle;


PROCEDURE SetDateStyle ( ds : DateStyle );
(* changes the date style for Write  *)
BEGIN
  gCalDateFmtRec.dStyle := ds
END SetDateStyle;


PROCEDURE GetDateStyle () : DateStyle;

BEGIN
  RETURN gCalDateFmtRec.dStyle;
END GetDateStyle;


PROCEDURE SetStdCalDateFmtRec (dfr : CalDateFmtRec);
BEGIN
  gCalDateFmtRec := dfr;
END SetStdCalDateFmtRec;


PROCEDURE GetStdCalDateFmtRec () : CalDateFmtRec;
BEGIN
  RETURN gCalDateFmtRec
END GetStdCalDateFmtRec;



(* ============== Procedures for time and strings *) 


PROCEDURE StringToTime (VAR time : TimeT; tStr : ARRAY OF CHAR; res : IOStatusT);
(* Parses a string to a time record *)

VAR
  strCount, hCount, mCount, sCount, pCount, hr, min, sec, psec, zHr, zMin  : CARDINAL;
  hStr, mStr, sStr, xStr : AOC.Len5;
  fStr : AOC.Len15;
  sign : CHAR;

  PROCEDURE ParseTimeUnitNum (VAR unit : CARDINAL; max, maxDigits : CARDINAL; res : IOStatusT);
  
  VAR
    uStr : AOC.Len15;
    uCount : CARDINAL;
  BEGIN
    SkipSeparators ();
    uCount := 0;
    unit := 0;
    WHILE ISO646.isDigit (tStr [strCount]) AND (uCount < maxDigits) DO
      uStr[uCount] := tStr [strCount];
      uCount := uCount + 1;
      strCount := strCount + 1
    END; (* while *)
    uStr[uCount] := ISO646.NUL;
    WholeStrs.StrToCard (uStr, unit, res);
    IF unit > max THEN
      res := IOStatusT{TRUE, IOStatus.OutOfRange}
    END 
  END ParseTimeUnitNum;

  PROCEDURE SkipSeparators ();
  BEGIN
    WHILE isSeparatorChar (tStr [strCount]) DO
      strCount := strCount + 1
    END; (* while *)
  END SkipSeparators;

  PROCEDURE ParseText (VAR xStr : ARRAY OF CHAR);
  VAR
    uCount : CARDINAL;

  BEGIN
    SkipSeparators ();
    uCount := 0;
    WHILE ISO646.isLetter (tStr [strCount]) DO
      xStr [uCount] := tStr [strCount];
      uCount := uCount + 1;
      strCount := strCount + 1
    END;
    xStr[uCount] := ISO646.NUL;
  END ParseText;


BEGIN (* main *)
  strCount := 0;
  time := Time.ZeroTime;
  res := IOStatusT{FALSE, IOStatus.Success};
  SkipSeparators ();
  IF (tStr [strCount] = "T") THEN
    strCount := strCount + 1
  END; (* if *)

  (* Parse hour *)
  ParseTimeUnitNum (hr,  MAX (Time.Hours), 2, res);
  IF (res.failed) THEN
    time := Time.ZeroTime;
    RETURN 
  ELSE
    time.hour :=  VAL (Time.Hours, hr)
  END; (* if *)

  (* Parse minute *)
  ParseTimeUnitNum (min,  MAX (Time.Minutes), 2, res);
  IF (res.failed) THEN
    time := Time.ZeroTime;
    RETURN 
  ELSE
    time.mins :=  VAL (Time.Minutes, min)
  END; (* if *)

  (* Parse seconds *)
  ParseTimeUnitNum (sec,  MAX (Time.Seconds), 2, res);
  IF (res.failed) THEN
    time := Time.ZeroTime;
    RETURN 
  ELSE
    time.secs :=  VAL (Time.Seconds, sec)
  END; (* if *)

  (* Parse parts of seconds *)
  ParseTimeUnitNum (psec,  MAX (Time.FracSeconds), 20, res);
  IF (res.failed) THEN
    time := Time.ZeroTime;
    RETURN 
  ELSE
    time.fracs :=  VAL (Time.FracSeconds, psec)
  END; (* if *)

  (* parse for AM/PM and*)
  ParseText (xStr);
  IF ( LENGTH (xStr) = 2 ) AND (xStr[1] = "M") THEN
    IF (xStr[0] = "P") AND (time.hour < 12) THEN
      time.hour := time.hour + 12 
    END; (* if *)
       (* and see if there is another string *)
    ParseText (xStr);
  END;
  IF CharArray.matches (xStr, UTCStr) AND (ISO646.isSign (tStr [strCount] ) ) THEN
    sign := tStr [strCount];
    strCount := strCount + 1;
    ParseTimeUnitNum (zHr, 12, 2, res);
    ParseTimeUnitNum (zMin, 59, 2, res);
    zMin := zMin + zHr * 60;
    time.tz.value := zMin;
    IF sign = "-" THEN
      time.tz.value := - time.tz.value
    END (* if *)
  ELSIF LENGTH (xStr) > 0 THEN
    res := IOStatusT{TRUE, IOStatus.WrongFormat}
  ELSE
    DateTimeUtilities.GetZone(time.tz)
  END (* if *)
END StringToTime;


PROCEDURE TimeToString (VAR  tStr : ARRAY OF CHAR; time : TimeT; tfr : TimeFmtRec);
(* Creates a string from a time record according to the time format record *)

VAR
  sCount, tzCount, max, absTime : CARDINAL;
  tzTime : INTEGER;
  len : CARDINAL;
  hStr, mStr, sStr : AOC.Len3;
  pStr : AOC.Len12;
  tzStr : FormatIO.Str;
  tSep : CHAR;
  pm, isNegTime : BOOLEAN;
  
BEGIN
  tStr [0] := ISO646.NUL;
    (* deal with hours *)
  pm := time.hour >= 12;
  IF (time.hour > 12) AND ( tfr.hStyle = h ) THEN time.hour := time.hour -12 END;
  WholeStrs.CardToDecStr ( time.hour, hStr);
  IF ( tfr.hStyle = H ) AND ( time.hour < 10 )THEN 
    tStr [0] := "0";
    tStr [1] := ISO646.NUL;
  END;  
  ARRAYOFCHAR.Concat (tStr, tStr, hStr);
  sCount := LENGTH (tStr);
  IF ( tfr.hStyle = h ) OR StrictISOTime THEN
    tStr [sCount] := ":";
    tStr [sCount + 1] := ISO646.NUL;
    sCount := sCount + 1
  END; (* if *)
    (* deal with minutes *)
  WholeStrs.CardToDecStr ( time.mins, mStr) ;
  IF time.mins < 10 THEN
    tStr [sCount] := "0";
    tStr [sCount + 1] := ISO646.NUL;
    sCount := sCount + 1
  END; (* if *)
  ARRAYOFCHAR.Concat (tStr, tStr, mStr);
  sCount := LENGTH (tStr);
    (* deal with seconds *)
  IF tfr.sStyle <> _s  THEN
    WholeStrs.CardToDecStr ( time.secs, sStr);
    tStr [sCount] := ":";
    tStr [sCount+1] := ISO646.NUL;
    ARRAYOFCHAR.Concat (tStr, tStr, sStr);
    sCount := LENGTH (tStr)
  END; (* if *)
    (* deal with parts of seconds *)
  IF tfr.sStyle = S THEN
    WholeStrs.CardToDecStr ( time.fracs, pStr);
    tStr [sCount] := ":";
    tStr [sCount+1] := ISO646.NUL;
    ARRAYOFCHAR.Concat (tStr, tStr, pStr);
    sCount := LENGTH (tStr)
  END; (* if *)
  IF ( tfr.hStyle = h  ) THEN
    IF pm THEN 
      ARRAYOFCHAR.Concat (tStr, tStr, " PM");
    ELSE
      ARRAYOFCHAR.Concat (tStr, tStr, " AM")
    END (* inner if *)
  END; (* outer if *)
  sCount := LENGTH (tStr);
  CASE tfr.tzStyle OF
       a : 
        TZ.toSXF (time.tz, tzStr);
        IF time.summerTimeFlag THEN
          IF tzStr [1] = "S" THEN
            tzStr [1] := "D"
          ELSE
            len := LENGTH (tzStr);
            tzStr [len] := "D";
            tzStr [len+1] := ISO646.NUL
          END (* inner if *)
        END; (*outer if*)
        tStr [sCount] := ISO646.SPACE;
        tStr [sCount + 1] := ISO646.NUL;
        ARRAYOFCHAR.Concat (tStr, tStr, tzStr)
    |  u :
        tStr [sCount] := ISO646.SPACE;
        tStr [sCount + 1] := ISO646.NUL;
        ARRAYOFCHAR.Concat (tStr, tStr, "UTC");
        isNegTime := ( time.tz.value < 0 );
        absTime := ABS (time.tz.value);
        tzTime := 100 * ( absTime DIV 60 ) + ( absTime MOD 60 );
        WholeStrs.CardToDecStr ( tzTime, tzStr );
        IF isNegTime THEN
          ARRAYOFCHAR.Concat (tStr, tStr, "-");
        ELSE
          ARRAYOFCHAR.Concat (tStr, tStr, "+");
        END;
        tzStr := FormatIO.JustifyC (tzStr, "0", 4, FormatIO.Right);
        ARRAYOFCHAR.Concat (tStr, tStr, tzStr)
    | _z :  
         RETURN;
    ELSE
  END; (* case *)
END TimeToString;


PROCEDURE SetTimeSeparator ( ch : CHAR );
(* changes the separator between parts of the time for Write *)

BEGIN
  gTimeFmtRec.tSeparator := ch 
END SetTimeSeparator;


PROCEDURE GetTimeSeparator () : CHAR;
(* Returns the saved separator for expected reads and for writes *)  

BEGIN
  RETURN gTimeFmtRec.tSeparator;
END GetTimeSeparator;


PROCEDURE SetHourStyle ( ds : HourStyle );
(* changes the hour style for writes  *)
BEGIN
  gTimeFmtRec.hStyle := ds
END SetHourStyle;


PROCEDURE GetHourStyle () : HourStyle;
(* Returns the hour style for writes *)  
BEGIN
  RETURN gTimeFmtRec.hStyle
END GetHourStyle;


PROCEDURE SetSecStyle ( ss : SecondStyle );
(* changes whether seconds (and parts thereof) are used by default for writes  *)
BEGIN
  gTimeFmtRec.sStyle := ss
END SetSecStyle;


PROCEDURE GetSecStyle () : SecondStyle;
(* Returns the stored boolean for how seconds are used for writes *)  
BEGIN
  RETURN gTimeFmtRec.sStyle
END GetSecStyle;


PROCEDURE SetStdTimeFmtRec (tfr :TimeFmtRec);
BEGIN
  gTimeFmtRec := tfr
END SetStdTimeFmtRec;


PROCEDURE GetStdTimeFmtRec () : TimeFmtRec;
BEGIN
  RETURN gTimeFmtRec
END GetStdTimeFmtRec;


PROCEDURE SetAlwaysAD (ADStyle : BOOLEAN);
  (* require the AD string for writing positive years. The BC String is
   always used for negative ones *)
BEGIN
  AlwaysAD := ADStyle
END SetAlwaysAD;   

   
PROCEDURE GetAlwaysAD () : BOOLEAN;
  (* return the stored value for the AlwaysAD  *)
BEGIN
  RETURN AlwaysAD
END GetAlwaysAD;

(* combined Procs *)

PROCEDURE SetStrictISOTime (ISOTimeStyle : BOOLEAN);
  (* require the colon when writing 24-hour times if TRUE, omits if false and uses railroad style *)
BEGIN
  StrictISOTime := ISOTimeStyle
END SetStrictISOTime;   
   
   
PROCEDURE GetStrictISOTime () : BOOLEAN;
  (* return the stored value for the StrictISOTime  *)
BEGIN
  RETURN StrictISOTime
END GetStrictISOTime;


PROCEDURE DateTimeToString (VAR  dtStr : ARRAY OF CHAR; dt : DateTimeT; dtfr : DateTimeFmtRec);

VAR
  tStr : AOC.Len20;
  len : CARDINAL;
BEGIN
  dtStr[0] := ISO646.NUL;
  DateToString (dtStr, dt.calDate, dtfr.cdfr );
  len := LENGTH (dtStr);
  TimeToString (tStr, dt.time, dtfr.tfr );
  tStr[len] := "T";
  tStr[len + 1] := ISO646.NUL;;
  ARRAYOFCHAR.Concat (dtStr,dtStr,tStr) 
END DateTimeToString;
(* =============*)

BEGIN  (* init *)
  SetStdCalDateFmtRec (DefaultCalDateFmtRec);
  SetStdTimeFmtRec (DefaultTimeFmtRec);
  AlwaysAD := DefaultAlwaysAD;
  StrictISOTime := DefaultISOTime;
  
END DateTimeFormats. (* DateTime *)