(* (C) 2009-2018 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE DateTimeUtilities;

IMPORT SYSTEM;
IMPORT DateTime; FROM DateTime IMPORT DateTimeT;
IMPORT Date; FROM Date IMPORT DateT;
IMPORT Time; FROM Time IMPORT TimeT;
FROM TZ IMPORT TZ;
FROM IOSupport IMPORT*;

IMPORT CTime, SysClock;

PROCEDURE GetDateToday (VAR userData: DateT);
VAR
  DaysOnClock, JSD, JD : LONGREAL;
  ElapsedSecs, CorrSecs, TimeZoneOffset : INTEGER;
  ElapsedParts : Time.FracSeconds;
  DST : BOOLEAN;
  status : SysClock.Status;

BEGIN
  SysClock.GetClock ( ElapsedSecs, ElapsedParts, TimeZoneOffset, DST, status );
    IF status.failed THEN sCharIO.WriteStr ("get clock failed\n\n") END;
    
  IF DST THEN TimeZoneOffset := TimeZoneOffset + 60 END;  (* Correct for DST *)
  CorrSecs := ElapsedSecs + ( TimeZoneOffset * 60 );
  DaysOnClock := VAL (LONGREAL, (CorrSecs  DIV secPerDay) );
  JSD := CalendarToJDN (clockStartDay.year, ORD (clockStartDay.month) + 1, clockStartDay.date, AD);
    sLongRealIO.Write (JSD);sCharIO.WriteStr ("\n\n");
    sLongRealIO.Write (DaysOnClock);sCharIO.WriteStr ("\n\n");

  JD := DaysOnClock + JSD;
    sLongRealIO.Write (JD);sCharIO.WriteStr ("\n\n");
  userData := JDNToCalendar (JD);
END GetDateToday;

PROCEDURE ThisYear () : Date.Years;
(* From the system clock obtain the current year. *)
VAR
  Temp : DateT;
BEGIN
  GetDateToday (Temp);
  RETURN Temp.year
END ThisYear;

PROCEDURE GetTimeNow (VAR userData: TimeT);
VAR
  RemSecs, CorrSecs, hour, min, sec : INTEGER;
  ElapsedSecs, TimeZoneOffset : INTEGER;
  ElapsedParts : Time.FracSeconds;
  DST : BOOLEAN;
  status : SysClock.Status;

BEGIN
  SysClock.GetClock ( ElapsedSecs, ElapsedParts, TimeZoneOffset, DST, status );
  IF DST THEN TimeZoneOffset := TimeZoneOffset + 60 END;  (* Correct for DST *)
  CorrSecs := ElapsedSecs + ( TimeZoneOffset * 60 );
  RemSecs :=  CorrSecs MOD secPerDay;
  hour :=  RemSecs DIV 3600;
  RemSecs := RemSecs MOD 3600;
  min := RemSecs DIV 60;
  sec := RemSecs MOD 60;
(* fill data structure for return *)
  userData.hour := hour ;
  userData.mins := min;
  userData.secs := sec;
  userData.fracs := ElapsedParts ;
  userData.tz.value := TimeZoneOffset;
  userData.summerTimeFlag := DST;
END GetTimeNow;

PROCEDURE GetZone ( VAR userData : TZ );
(* From the system clock obtain the current time zone corrected for DST. *)

VAR  ElapsedSecs, TimeZoneOffset : INTEGER;
  ElapsedParts : Time.FracSeconds;
  DST : BOOLEAN;
  status : SysClock.Status;

BEGIN
  SysClock.GetClock ( ElapsedSecs, ElapsedParts, TimeZoneOffset, DST, status );
  IF DST THEN TimeZoneOffset := TimeZoneOffset + 60 END;  (* Correct for DST *)
  userData.value := TimeZoneOffset;
END GetZone;

PROCEDURE GetDateTime(VAR userData: DateTimeT);

BEGIN
  GetDateToday (userData.calDate);
  GetTimeNow (userData.time);
END GetDateTime;


PROCEDURE CalendarToJDN (Year, Month, Day : INTEGER; AD : BOOLEAN) : LONGREAL;
VAR
   Y, M, D, LeapAdjust, DaysByYears, DaysByMonths : INTEGER;
BEGIN
  M := Month;
  D := Day;
  Y :=  Year;
  IF NOT AD THEN
     Y := -(Y - 1) (* year 1BC becomes year 0 in this system *)
  END;
  IF M < 3 THEN (* switch to previous year months 13, 14 *)
    Y := Y -1;
    M := M + 12
  END;
  IF AD THEN
    DaysByYears := VAL (INTEGER, (365.25 * VAL (LONGREAL, Y ) ) );
    IF ( (Year < 1582) OR
         ( (Year = 1582) AND (Month < 10) ) OR
         ( (Year = 1582) AND (Month = 10) AND (Day < 15) ) )
         THEN
      (* before start of Gregorian calendar *)
      LeapAdjust := 0
    ELSE
        (* after start of Gregorian calendar *)
      LeapAdjust := 2 - INT (Y DIV 100) + INT (Y DIV 400)
    END; (* if AD *) 
  ELSE (* NOT AD *)
    LeapAdjust := 0;
    DaysByYears := VAL (INTEGER, (365.25 * VAL (LONGREAL, Y ) ) -0.75 );
  END; (* main if *)
  DaysByMonths :=  VAL (INTEGER, 30.6001 * LFLOAT (M + 1) );
  RETURN VAL (LONGREAL, D + LeapAdjust + DaysByYears + DaysByMonths ) + 1720994.5 
END CalendarToJDN;


PROCEDURE JDNToCalendar (Jdn : LONGREAL) : DateT;

(* Algorithm parameters for Gregorian/Julian calendar*)
CONST
  v = 3;
  j = 1401;
  u = 5;
  MonthAdjust = 2;
  s = 153;
  MonthsInYear = 12;
  w = 2;
  r = 4;
  B = 274277;
  C = -38;

VAR
  Dt : DateT;
  Ijdn, e,f,g,h,D,M,Y : INTEGER;

BEGIN
  Jdn := Jdn + 0.5; (* go to next full day *)
  Ijdn := VAL (INTEGER, Jdn);
  (* For Julian calendar *)
  f := Ijdn + j;
  (* For Gregorian calendar, correct this: *)
  IF Ijdn >= startDayGregorian THEN
    f := f + (((4 * Ijdn + B) DIV 146097) * 3) DIV 4 + C
  END;
  (* For Julian or Gregorian, continue: *)
  e := r * f + v;
  g := (e MOD daysPerQuadYear) DIV r;
  h := u * g + w;

  D := (h MOD s) DIV u + 1;
  M := (h DIV s + MonthAdjust) MOD MonthsInYear + 1;
  Y := (e DIV daysPerQuadYear) - julianYearAdjust + (MonthsInYear + MonthAdjust - M) DIV MonthsInYear;
     (* D, M, and Y are the numbers of the day, month, and year respectively
     for the afternoon at the beginning of the given Julian day.*)
  Dt.year := Y;
  Dt.month := VAL (Date.Months, M - 1);
  Dt.date := D;
  RETURN Dt
END JDNToCalendar; 

PROCEDURE DayNumInYear (dt : DateT) : Date.DayNumRange;
(* returns a number in the range 0 .. 366 *)
VAR
  leap : BOOLEAN;
  month, adjMonth, daysSinceMarch, maxDays, result : CARDINAL;
  
BEGIN
  (* we use a method developed by RS for his Modua-2 textbook *)
  leap:= Date.isLeapYear (dt.year);
  month := ORD (dt.month) + 1;
  adjMonth := month + 12 * (( 12 - month) DIV 10);
  daysSinceMarch := TRUNC (30.6 * FLOAT (adjMonth - 3) + 0.5);
  result := 59 + daysSinceMarch + dt.date;
  IF result > 365 THEN result := result - 365 END;
  IF leap AND (month > 2) THEN result := result + 1 END;
  RETURN result
END DayNumInYear;


PROCEDURE JDNFromYearAndDayNum (yr : Date.Years; dn : Date.DayNumRange): LONGREAL;

BEGIN
  RETURN CalendarToJDN ( yr, 1, 1 , AD) + VAL (LONGREAL, dn) - 1.0
  (* the -1.0 is because we want to add the difference from Jan 1, and
  Jan 1 is day number 1 not day number 0 *)
END JDNFromYearAndDayNum;


PROCEDURE WeekDayNum (dt : DateT) : Date.WeekRange;
(* returns a number in the range 0 .. 6 *)
VAR
  Month : INTEGER;
  AD : BOOLEAN;
  Jdn : LONGREAL; 
BEGIN
  AD := ( dt.year > 0 );
  IF NOT AD THEN dt.year := -(dt.year - 1) END;
  Month := ORD (dt.month) + 1;
  Jdn := CalendarToJDN ( dt.year, Month, dt.date, AD );
  RETURN (VAL (INTEGER, Jdn + 2.0)  MOD 7); 
END WeekDayNum;

PROCEDURE DayToISODayNum (Day : Date.Days) : Date.ISODayRange;

BEGIN
  IF Day = Date.Sun THEN
    RETURN 7;
  ELSE 
    RETURN ORD (Day)
  END (* if *)  
END DayToISODayNum;

PROCEDURE WeekDayValue (dt : DateT) : Date.Days;
(* returns a day in the range Sun .. Sat *)

VAR
  wn : CARDINAL;
BEGIN
  RETURN VAL ( Date.Days, WeekDayNum (dt) );
END WeekDayValue;


PROCEDURE isShortFirstWeek (yr: Date.Years) : BOOLEAN;
VAR
  DnJan1 : CARDINAL;
  
BEGIN 
   DnJan1 := WeekDayNum ( DateT { yr, Date.Jan, 1 } ); 
   RETURN ( DnJan1 = 5) OR (DnJan1 = 6) OR (DnJan1 = 0)
END isShortFirstWeek;

 
PROCEDURE ISOWeekNumInYear (dt : DateT; VAR yr: Date.Years; VAR wk : Date.WeekRange; VAR da : Date.ISODayRange );
(* returns a number in the range 0..53  Days are 1..7=Sun *)

VAR
  dv, DnJan1 : CARDINAL;
  
BEGIN
  yr := dt. year;
  DnJan1 := WeekDayNum ( DateT { yr, Date.Jan, 1 } );
  IF DnJan1 = 0 THEN DnJan1 := 7 END;
  wk := ( DayNumInYear (dt) + DnJan1 + 5 ) DIV 7;

  IF isShortFirstWeek (yr) THEN (* up to three days go in last week of previous year *)
    wk := wk - 1;
    IF wk = 0 THEN 
      yr := yr - 1; (* go back a year *)
      IF isShortFirstWeek (yr) THEN (* two in a row *)
        wk := 52
      ELSE 
        wk := 53
      END (* if *)
    END  (* if *)
  ELSE (* not a short week so in any case only 52 weeks *)
    IF wk = 53 THEN
      wk := 1;
      yr := yr + 1;
    END;
  END; (* if *) 

  dv := WeekDayNum (dt);
  IF dv = 0 THEN (* and in this ISO format, days are numbered 1..7 rather than 0..6 *)
   da := 7
  ELSE
   da := dv
  END (* if *) 
END ISOWeekNumInYear;


END DateTimeUtilities.