(* (C) 2009-2018 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE SysClock;

(* Interface to the System's Clock *)

IMPORT SYSTEM;
IMPORT UNSAFE;
IMPORT Time; FROM Time IMPORT TimeT;
FROM INTEGER32 IMPORT INTEGER32;
IMPORT CTime, TimeUnits;
FROM TZ IMPORT TZ;

PROCEDURE GetClock ( VAR Secs  : INTEGER;
                     VAR Parts : Time.FracSeconds;
                     VAR tzOffset : INTEGER;
                     VAR DST : BOOLEAN;
                     VAR status   : Status );

VAR
  t : CTime.time_t;
  tv : CTime.timeval;
  tvp : CTime.timevalPtr;
  err : INTEGER;
  zone : CTime.timezone;
  tzp : CTime.tzPtr;

BEGIN
  tvp := SYSTEM.ADR (tv);
  tzp := SYSTEM.ADR (zone);
  tzOffset := 0;
  tv := CTime.timeval{0,0}; (* this shouldn't be necessary, but it is. *)
  zone := CTime.timezone{0,0};
  err := CTime.gettimeofday( tvp, tzp);
  IF err <> 0 THEN 
    status := Status{TRUE, AccessDenied};
    RETURN
  END;
  Secs := tvp^.tv_sec;
  Parts := tvp^.tv_usec;
  tzOffset := -tzp^.tz_minuteswest;
  DST := (tzp^.tz_dsttime > 0);
  status := Status{FALSE, Success};
  RETURN;
END GetClock;

PROCEDURE SetClock ( newSecValue   : INTEGER;
                     newPartsValue: Time.FracSeconds;
                     tzOffset   : INTEGER;
                     VAR status : Status );
BEGIN
  status := Status{TRUE, AccessDenied};
  RETURN;
END SetClock;

PROCEDURE ProcessMaySetClock (): BOOLEAN;
(* Returns TRUE if the current process is privileged to set the system clock,
   otherwise FALSE. *)

BEGIN
  RETURN FALSE; (* mac *)
END ProcessMaySetClock;


END SysClock.
