(* (C) 2009-2018 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE TZ (*[IntervalType] *);
IMPORT AOC, ISO646, ARRAYOFCHAR;
FROM INTEGER16 IMPORT INTEGER16;

CONST
  maxZones = 92;

TYPE
  TZRec =  RECORD
    tzStr : AOC.Len5;
    int : INTEGER16
  END; (* record *)
  TZArray = ARRAY [0..maxZones-1] OF TZRec;

CONST
  TZLookup = TZArray 
  {
    TZRec{"UTC",  0},      (* Universal Time Coordinated, UTC *)
    TZRec{"GMT",  0},      (* Greenwich Mean Time, UTC+0000 *)
    TZRec{"WET",  0},      (* Western European Time, UTC+0000 *)
    TZRec{"CET",  60},     (* Central European Time, UTC+0100 *)
    TZRec{"MET",  60},     (* Middle European Time, UTC+0100 *)
    TZRec{"WAT",  60},     (* West African Time, UTC+0100 *)
    TZRec{"EET",  120},    (* Eastern European Time, UTC+0200 *)
    TZRec{"USZ1", 120},    (* Russia Zone 1 Time, UTC+0200 *)
    TZRec{"CAT",  120},    (* Central African Time, UTC+0200 *)
    TZRec{"SAST", 120},    (* South African Standard Time, UTC+0200 *)
    TZRec{"TST",  180},    (* Turkish Standard Time, UTC+0300 *)
    TZRec{"AST",  180},    (* Arabian Standard Time, UTC+0300 *)
    TZRec{"EAT",  180},    (* East African Time, UTC+0300 *)
    TZRec{"MSK",  180},    (* Moscow Time, UTC+0300 *)
    TZRec{"USZ2", 180},    (* Russia Zone 2 Time, UTC+0300 *)
    TZRec{"IRT",  210},    (* Iran Time, UTC+0330 *)
    TZRec{"USZ3", 240},    (* Russia Zone 3 Time, UTC+0400 *)
    TZRec{"AZT",  240},    (* Azerbaijan Time, UTC+0400 *)
    TZRec{"RET",  240},    (* Reunion Time, UTC+0400 *)
    TZRec{"MUT",  240},    (* Mauritius Time, UTC+0400 *)
    TZRec{"AFT",  270},    (* Afghanistan Time, UTC+0430 *)
    TZRec{"USZ4", 300},    (* Russia Zone 4 Time, UTC+0500 *)
    TZRec{"KGT",  300},    (* Kyrgyzstan Time, UTC+0500 *)
    TZRec{"TMT",  300},    (* Turkmenistan Time, UTC+0500 *)
    TZRec{"TJT",  300},    (* Tajikistan Time, UTC+0500 *)
    TZRec{"UZT",  300},    (* Uzbekistan Time, UTC+0500 *)
    TZRec{"PKT",  300},    (* Pakistan Time, UTC+0500 *)
    TZRec{"MVT",  300},    (* Maldives Time, UTC+0500 *)
    TZRec{"IST",  330},    (* Indian Standard Time, UTC+0530 *)
    TZRec{"NPT",  345},    (* Nepal Time, UTC+0545 *)
    TZRec{"LKT",  360},    (* Lanka Time, UTC+0600 *)
    TZRec{"BTT",  360},    (* Bhutan Time, UTC+0600 *)
    TZRec{"BDT",  360},    (* Bangladesh Time, UTC+0600 *)
    TZRec{"USZ5", 360},    (* Russia Zone 5 Time, UTC+0600 *)
    TZRec{"NSUT", 390},    (* North Sumatra Time, UTC+0630 *)
    TZRec{"MMT",  390},    (* Myanmar/Burma Time, UTC+0630 *)
    TZRec{"USZ6", 420},    (* Russia Zone 6 Time, UTC+0700 *)
    TZRec{"THA",  420},    (* Thailand Standard Time, UTC+0700 *)
    TZRec{"ICT",  420},    (* Indochina Time, UTC+0700 *)
    TZRec{"JAVT", 420},    (* Java Time, UTC+0700 *)
    TZRec{"MYT",  480},    (* Malaysia Time, UTC+0800 *)
    TZRec{"SGT",  480},    (* Singapore Time, UTC+0800 *)
    TZRec{"BNT",  480},    (* Brunei Time, UTC+0800 *)
    TZRec{"CCT",  480},    (* China Coastal Time, UTC+0800 *)
    TZRec{"HKT",  480},    (* Hong Kong Time, UTC+0800 *)
    TZRec{"PHT",  480},    (* Philippine Time, UTC+0800 *)
    TZRec{"USZ7", 480},    (* Russia Zone 7 Time, UTC+0800 *)
    TZRec{"ULAT", 480},    (* Ulaan Bataar Time *)
    TZRec{"WST",  480},    (* Australian Western Standard Time, UTC+0800 *)
    TZRec{"USZ8", 540},    (* Russia Zone 8 Time, UTC+0900 *)
    TZRec{"KST",  540},    (* Korean Standard Time, UTC+0900 *)
    TZRec{"JST",  540},    (* Japanese Standard Time, UTC+0900 *)
    TZRec{"ACST", 570},    (* Australian Central Standard Time, UTC+0930 *)
    TZRec{"PGT",  600},    (* Papua New Guinea Time, UTC+1000 *)
    TZRec{"USZ9", 600},    (* Russia Zone 9 Time, UTC+1000 *)
    TZRec{"AEST", 600},    (* Australian Eastern Standard Time, UTC+1000 *)
    TZRec{"UZ10", 660},    (* Russia Zone 10 Time, UTC+1100 *)
    TZRec{"NCT",  660},    (* New Caledonia Time, UTC+1100 *)
    TZRec{"VUT",  660},    (* Vanuatu Time, UTC+1100 *)
    TZRec{"FJT",  720},    (* Fiji Time, UTC+1200 *)
    TZRec{"TVT",  720},    (* Tuvalu Time, UTC+1200 *)
    TZRec{"NZT",  720},    (* New Zealand Time, UTC+1200 *)
    TZRec{"NZST", 720},    (* New Zealand Standard Time, UTC+1200 *)
    TZRec{"IDLE", 720},    (* International Date Line East, UTC+1200 *)
    TZRec{"AZOT", -60},    (* Azores Time, UTC-0100 *)
    TZRec{"CVT",  -60},    (* Cape Verde Time, UTC-0100 *)
    TZRec{"VTZ",  -120},   (* Greenland Eastern Standard Time, UTC-0200 *)
    TZRec{"BRA",  -180},   (* Brazil Standard Time, UTC-0300 *)
    TZRec{"ART",  -180},   (* Argentina Time, UTC-0300 *)
    TZRec{"UYT",  -180},   (* Uruguay Time, UTC-0300 *)
    TZRec{"SRT",  -180},   (* Suriname Time, UTC-0300 *)
    TZRec{"GFT",  -180},   (* French Guyana Time, UTC-0300 *)
    TZRec{"UTZ",  -180},   (* Greenland Western Standard Time, UTC-0300 *)
    TZRec{"CLT",  -240},   (* Chile Time, UTC-0400 *)
    TZRec{"BOT",  -240},   (* Bolivia Time, UTC-0400 *)
    TZRec{"GYT",  -240},   (* Guyana Time, UTC-0400 *)
    TZRec{"PYT",  -240},   (* Paraguay Time, UTC-0400 *)
    TZRec{"FKT",  -240},   (* Falkland Island Time, UTC-0400 *)
    TZRec{"EST",  -300},   (* US Eastern Standard Time, UTC-0500 *)
    TZRec{"COT",  -300},   (* Colombia Time, UTC-0500 *)
    TZRec{"PET",  -300},   (* Peru Time, UTC-0500 *)
    TZRec{"CST",  -360},   (* US Central Standard Time, UTC-0600 *)
    TZRec{"MEX",  -360},   (* Mexico Time, UTC-0600 *)
    TZRec{"MST",  -420},   (* US Mountain Standard Time, UTC-0700 *)
    TZRec{"PST",  -480},   (* US Pacific Standard Time, UTC-0800 *)
    TZRec{"AKST", -540},   (* US Alaska Standard Time, UTC-0900 *)
    TZRec{"YST",  -540},   (* US Yukon Standard Time, UTC-0900 *)
    TZRec{"HST",  -600},   (* US Hawaiian Standard Time, UTC-1000 *)
    TZRec{"HAST", -600},   (* US Hawaii-Aleutian Standard Time, UTC-1000 *)
    TZRec{"THAT", -600},   (* Thahiti Time, UTC-1000 *)
    TZRec{"BET",  -660},   (* Bering Standard Time, UTC-1100 *)
    TZRec{"IDLW", -720}   (* International Date Line West, UTC-1200 *)
  };

PROCEDURE (*[SXF]*) toSXF ( n : TZ; VAR s : ARRAY OF CHAR );
(* Converts TZ value n to a string in scalar exchange format.
   This procedure is bound to pervasive primitive SXF for type TZ. *)
VAR
  count : CARDINAL;

BEGIN
  count := 0;
  s[0] := ISO646.NUL;
  WHILE count <= maxZones DO
    IF n.value = TZLookup[count].int THEN
      ARRAYOFCHAR.Assign (s, TZLookup[count].tzStr);
      RETURN
    ELSE
      count := count + 1
    END (* if*)
  END (* while *)
END toSXF;


PROCEDURE (*[VAL]*) fromSXF ( (*CONST*) s : ARRAY OF CHAR; VAR n : TZ );
(* Converts a string in scalar exchange format to a TZ value.
   This procedure is bound to pervasive primitive VAL for type TZ. *)
BEGIN

END fromSXF;

    
END TZ.