(* (C) 2009-2018 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE TimeBaseIO ;

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYn;
IMPORT ChanIO;
IMPORT DateTimeFormats, DateStrs, TimeUnits;
IMPORT Time; FROM Time IMPORT TimeT;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT CharIO, CharStrIO, CharArray, CardinalIO, FormatIO, WholeStrs;
IMPORT AOC, ISO646;

TYPE
  TimeFmtRec = (* ALIAS *) DateTimeFormats.TimeFmtRec;
  Order = (* ALIAS *) DateTimeFormats.Order;
  FormatScanRec = (* ALIAS *) FormatIO.FormatScanRec;
  
CONST
  isSeparatorChar = DateTimeFormats.isSeparatorChar;
  Separators = (* ALIAS *) DateTimeFormats.Separators;
  dmy = (* ALIAS *) DateTimeFormats.dmy;
  mdy = (* ALIAS *) DateTimeFormats.mdy;
  ymd = (* ALIAS *) DateTimeFormats.ymd;
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatTextStr = FormatIO.FormatTextStr;
  SetCustomScanSet = FormatIO.SetCustomScanSet;
  SetValidCustomFmtChars = FormatIO.SetValidCustomFmtChars;
  MonthStrToMonthNum = DateStrs.MonthStrToMonthNum;
  YearStrToYearNum = DateStrs.YearStrToYearNum;
  DateStrToDateNum = DateStrs.DateStrToDateNum;
  
VAR
  gtir : TimeFmtRec;

PROCEDURE (*[READ]*) Read( inChan : ChanIO.Channel; VAR ti : TimeT );

VAR
  inputStr : AOC.Len512;
  sCount : CARDINAL;
  res : IOStatusT;
  validStatus : BOOLEAN;

BEGIN  (* main proc *)
  CharIO.ReadStr (inChan, inputStr);
  DateTimeFormats.StringToTime ( ti, inputStr, res );
  IF NOT (res.code = IOStatus.Success) THEN
      ChanIO.SetStatus (inChan, IOStatusT { TRUE, IOStatus.WrongFormat }, validStatus )
  END (* if *)
END Read;


PROCEDURE (*[WRITE]*) Write( outChan : ChanIO.Channel; ti : TimeT );
(* Writes the textual representation of value dt to stream outChan. This
   procedure is substituted for invocations of WRITE with a DateTime argument.*)
VAR
  str : AOC.Len64;

BEGIN
  gtir := DateTimeFormats.GetStdTimeFmtRec ();
  DateTimeFormats.TimeToString (str, ti, gtir);
  CharIO.WriteStr (outChan, str)
END Write;

TYPE
  CustomType = TimeT;
  
PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items : ARRAY OF CustomType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  tfr : DateTimeFormats.TimeFmtRec;
  nc : CustomType;

BEGIN
  numItems := CAPACITYn (items, SYSTEM.TSIZE (CustomType)) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  tfr := DateTimeFormats.GetStdTimeFmtRec();
  WHILE count <= numItems DO
    nc := items[count];
    temp[0] := ISO646.NUL;
    DateTimeFormats.ParseFmtStrToTfr ( tfr, fsr.customFormatStr, res);
    DateTimeFormats.TimeToString (temp, nc, tfr);

    (* format that converted string using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

       (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END (* if *);

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END (* if *);

    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)
END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST*) ARRAY OF TimeT );
(* Writes a formatted textual representation of one or more TZ values to
   output stream outChan. The output format is determined by fmtStr. This
   procedure is substituted for invocations of WRITEF with one or more
   DateTime arguments. *)

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  SetCustomScanSet (FormatIO.DefaultCustomScanSet);
  SetValidCustomFmtChars (FormatIO.DefaultValidCustomFmtChars);
  ScanStrForFormat (fmtStr, FormatIO.GetCustomScanSet(), FormatIO.GetValidCustomFmtChars(), fsr);
  WriteFWithfsr ( outfile, fsr , items  )
END WriteF;


BEGIN  (* init *)
  gtir := DateTimeFormats.GetStdTimeFmtRec (); (* initial value; subject to change *)
END TimeBaseIO.