IMPLEMENTATION MODULE BitsetStr;
(* BITSET/string conversions *)

(* the binary string form of a BITSET is
    bit digit, {bit digit} for the number of bits in a bitset  *)

(* the literal form of a BITSET  is
     "{ "whole number,  {comma, whole number}, " }    *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYch;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

IMPORT ARRAYOFCHAR; 
IMPORT ISO646, WholeStrs;

TYPE
  OctStr = ARRAY [0..3] OF CHAR;

CONST
  Concat = ARRAYOFCHAR.Concat;
  bitsPerBitset = 8 * SYSTEM.TSIZE(BITSET);
  lastBitNum = bitsPerBitset - 1;
  maxlen = 511;

TYPE
  StrRange = [0..maxlen];
  Str = ARRAY StrRange OF CHAR;


PROCEDURE BitStrToBitset (str: ARRAY OF CHAR; VAR B: BITSET; VAR res : IOStatusT);

VAR
  bCount, sCount, last : CARDINAL;
  ch : CHAR;

BEGIN (* BitStrToBitset *)
  B := BITSET{};
  res := IOStatusT{ FALSE, IOStatus.Success};
  last := LENGTH (str);
  IF last = 0 THEN RETURN END;
  sCount := 0;
  bCount := 0;
  WHILE (sCount <= last) AND (bCount <= lastBitNum)
    DO
      ch := str [sCount];
      CASE ch OF
        "1" :
          INCL (B, bCount);
          bCount := bCount + 1;
       | "0" :
          bCount := bCount + 1;
       | " ", "." :
          (* ignore *)
        ELSE
          res := IOStatusT{ TRUE, IOStatus.WrongFormat};
          RETURN
        END; (* case *)
      sCount := sCount + 1;

    END; (* while *)
  IF sCount < last THEN res := IOStatusT{ TRUE, IOStatus.OutOfRange}; END; (* more bits than in a bitset *)
END BitStrToBitset;


(* local proc *)
PROCEDURE fetchNumFromStr (str : ARRAY OF CHAR; VAR index : StrRange) : CARDINAL;
(* pulls a string of digits from the input string and converts to a cardinal
   if no digits at location, returns a zero *)
VAR
    res : IOStatusT;
    fCount, sLen : StrRange;
    theCard : CARDINAL;
    numStr : Str;

BEGIN (* fetchNumFromFmtStr ( *)
  sLen := LENGTH (str) - 1;
  fCount := 0;
  theCard := 0;
  IF sLen = 0 THEN RETURN 0  END; (*IF*)
   WHILE  (index <=  sLen-1) AND (ISO646.isDigit (str [index]) ) DO (* copy the numeric chars *)
     numStr [fCount ] := str [index];
     fCount := fCount + 1;
     index := index + 1
   END;
   index := index - 1; (* set back to point at the last digit if we parsed anything *)
   numStr [fCount] := CHR (0);
   WholeStrs.StrToCard (numStr, theCard, res);
   RETURN theCard
END fetchNumFromStr;


PROCEDURE LitBitsetStrToBitset (str: ARRAY OF CHAR; VAR B: BITSET; VAR res : IOStatusT);

VAR
  sCount, last: StrRange;
  bNum : CARDINAL;
  ch : CHAR;
  done : BOOLEAN;

BEGIN (* LitBitsetStrToBitset *)
  B := BITSET{};
  res := IOStatusT{ FALSE, IOStatus.Success};
  last := LENGTH (str) - 1;
  IF last = 0 THEN RETURN END;
  sCount := 0;
  done := FALSE;

 WHILE (str [sCount] = " ") AND (sCount <= last)
   DO (* skip leading spaces *)
     sCount := sCount + 1
  END;

    IF sCount >= last THEN RETURN END;

    IF (str [sCount] = "{" ) THEN (* check on opening format *)
      sCount := sCount + 1;
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END;

  WHILE (sCount < last) AND NOT done
    DO
      ch := str [sCount];
      IF (ch = ",") OR (ch = " ") THEN
       (* do nothing *)
      ELSIF ISO646.isDigit (ch) THEN
        bNum := fetchNumFromStr (str, sCount);
        sCount := sCount + 1;
        IF bNum <= lastBitNum THEN
          INCL ( B, bNum );
          ch := str [sCount];
          IF ch = "}" THEN
             done := TRUE;
          END;
        ELSE
          res := IOStatusT{ TRUE, IOStatus.WrongFormat};
          RETURN
        END (* if bnum *)
    ELSIF ch = "}" THEN
     done := TRUE;
      ELSE
        res := IOStatusT{ TRUE, IOStatus.WrongFormat};
        RETURN
      END (* if ch *);
      IF NOT done THEN sCount := sCount + 1 END;
    END; (* while *)
  IF str[sCount] # "}" THEN res := IOStatusT{ TRUE, IOStatus.WrongFormat} END;
  WHILE (str [sCount] = " ") AND (sCount < last) (* skip trailing spaces *)
    DO
      sCount := sCount + 1
    END;

END LitBitsetStrToBitset;



PROCEDURE BitsetToBinStr (B: BITSET; VAR str: ARRAY OF CHAR);

VAR
  count, last, lastAvailStrPos : CARDINAL;

BEGIN (* BitsetToBinStr *)
  lastAvailStrPos  := CAPACITYch (str) - 2; (* and if a nul str is passed this will crash -- so be it *)
  IF lastAvailStrPos < lastBitNum THEN
     last := lastAvailStrPos
  ELSE
    last := lastBitNum
  END;
  count := 0;
  WHILE count <= last
    DO
      IF count IN B THEN
        str[count] := "1"
      ELSE
        str[count] := "0"
      END;
    count := count + 1
    END; (* while *)
    str [ count ] := ISO646.NUL
END BitsetToBinStr;


PROCEDURE BitsetToLitStr (B: BITSET; VAR str: ARRAY OF CHAR);

VAR
  count : CARDINAL;
  foundFirst : BOOLEAN;
  oStr : OctStr;

BEGIN (* BitsetToLitStr *)
  str [0] := "{"; str [1] := " "; str[2] := ISO646.NUL;
  count := 0;
  foundFirst := FALSE;
  WHILE (count <= lastBitNum) AND NOT foundFirst
    DO
      IF count IN B THEN
        WholeStrs.CardToDecStr (count, oStr);
        Concat (str, str, oStr);
        foundFirst := TRUE
      END; (* if *)
    count := count + 1
    END; (* while *)
  IF foundFirst  THEN
    WHILE (count <= lastBitNum)
    DO
      IF count IN B THEN
        Concat (str, str, ", ");
      WholeStrs.CardToDecStr (count, oStr);
      Concat (str, str, oStr)
      END; (* if *)
    count := count + 1
    END; (* while *)
  END; (* if *)
  Concat (str, str, " }");

END BitsetToLitStr;

END BitsetStr.
