IMPLEMENTATION MODULE  Card8Strs;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      07/20/2017  First implementation

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYo, CAPACITYch;
IMPORT CARD8; FROM CARD8 IMPORT CARD8T, CAPACITYc8;
IMPORT ISO646, ARRAYOFCHAR, CharArray;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* RS Utils *)
IMPORT FormatIO;

CONST
  zeroAsc = ORD ("0");
  maxCardDiv10  = MAX (CARD8T) / 10;
  lastMaxCardDigitChar = CHR( (MAX (CARD8T) REM 10) + zeroAsc);

TYPE
  NibbleStr = ARRAY [0..4] OF CHAR;

(* utility local procs *)

PROCEDURE HexDigitToHexChar (num : CARD8T) : CHAR;
(* Pre: the number passed must be 0..16 *)

BEGIN
  IF num < 10
    THEN RETURN CHR (ORD ("0") + num)
    ELSE RETURN CHR (ORD ("A") + num-10)
  END(*IF*)
END HexDigitToHexChar;


PROCEDURE HexCharToCard8 (ch : CHAR; VAR n : CARD8T; VAR res : IOStatusT);

VAR
  ordVal : CARD8T;

BEGIN
(* Pre: the digit passed must be a correct Hex digit *)
(* post: the function returns the decimal value of the correct hex digit, or zero if it is invalid *)
  ordVal := ORD (ch);
  IF (ordVal >= ORD ("0")) AND (ordVal <= ORD ("9"))
    THEN
      res := IOStatusT{ FALSE, IOStatus.Success};
      n := ordVal - ORD ("0")
    ELSIF (ordVal >= ORD ("A")) AND (ordVal <= ORD ("F")) THEN
      res := IOStatusT{ FALSE, IOStatus.Success};
      n := ordVal - ORD ("A") + 10
    ELSE
      res := IOStatusT{ TRUE, IOStatus.OutOfRange};
  END(*IF*)
END HexCharToCard8;

PROCEDURE HexDigitToNibbleStr (num : CARD8T) : NibbleStr;
(* Pre: the number passed must be 0..16 *)

VAR
  NStr : NibbleStr;
  
BEGIN
  CASE num OF
      0 : NStr := "0000"
    | 1 : NStr := "0001"
    | 2 : NStr := "0010"
    | 3 : NStr := "0011"
    | 4 : NStr := "0100"
    | 5 : NStr := "0101"
    | 6 : NStr := "0110"
    | 7 : NStr := "0111"
    | 8 : NStr := "1000"
    | 9 : NStr := "1001"
    | 10 : NStr := "1010"
    | 11 : NStr := "1011"
    | 12 : NStr := "1100"
    | 13 : NStr := "1101"
    | 14 : NStr := "1110"
    | 15 : NStr := "1111"
  END; (* case *)
  RETURN NStr
END HexDigitToNibbleStr;
    
    
(* the ones needing implementing *)


PROCEDURE StrToCard8 (str: ARRAY OF CHAR; VAR card8: CARD8T; VAR res : IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are in the format of an unsigned whole number, assigns a corresponding value to card. Assigns a value indicating the format of str to res.  *)

BEGIN
  FormatIO.RemoveLeadingSpaces (str);
  IF (str[0] = "0")
    THEN
      IF (str [1] = "x")
        THEN
          HexStrToCard8 (str, card8, res);
          RETURN
        ELSIF str [1] = "b" THEN
          BinStrToCard8 (str, card8, res);
          RETURN
        END;
     ELSE DecStrToCard8 (str, card8, res)
     END(*IF*)

END StrToCard8;


PROCEDURE DecStrToCard8 (str: ARRAY OF CHAR; VAR card8: CARD8T; VAR res : IOStatusT);

VAR
  accum, multiplier : CARDINAL;
  numDigits, index, base : CARD8T;

BEGIN (* DecStrToCard8 *)
  FormatIO.RemoveLeadingExtras (str);
  base := 10;
  numDigits := LENGTH (str);
  IF numDigits > 3 THEN 
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  END;
  multiplier := 1;
  index := numDigits;
  accum := 0;
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("9") ) THEN
      accum := accum + ( ORD (str [index]) - ORD ("0") ) * multiplier;
      multiplier := multiplier * base;
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END;
  UNTIL index = 0;
  IF accum <= MAX (CARD8T) THEN
    card8 := accum;
    res := IOStatusT{ FALSE, IOStatus.Success};;
    RETURN;
  ELSE
    card8 := 0;
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
    RETURN
  END (* if *)
END DecStrToCard8;


PROCEDURE HexStrToCard8 (str: ARRAY OF CHAR; VAR card8: CARD8T; VAR res : IOStatusT);
VAR
  accum, multiplier : CARDINAL;
  numDigits, index, base : CARD8T;

BEGIN (* DecStrToCard8 *)
  FormatIO.RemoveLeadingExtras (str);
  base := 16;
  numDigits := LENGTH (str);
  IF numDigits > 2 THEN
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  END;
  multiplier := 1;
  index := numDigits;
  accum := 0;
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("9") ) THEN
      accum := accum + ( ORD (str [index]) - ORD ("0") ) * multiplier;
      multiplier := multiplier * base;
    ELSIF ( ORD (str [index]) >= ORD ("A") ) AND ( ORD (str [index]) <= ORD ("F") ) THEN
      accum := accum + ( ORD (str [index]) - ORD ("A") + 10) * multiplier;
      multiplier := multiplier * base;
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END;
  UNTIL index = 0;
  (* Unlike the base 10 one we cannot overflow at this point as only two chars are used *)
  card8 := accum;
  res := IOStatusT{ FALSE, IOStatus.Success};;
  RETURN;
END HexStrToCard8;


PROCEDURE BinStrToCard8 (str: ARRAY OF CHAR; VAR card8: CARD8T; VAR res : IOStatusT);
VAR
  accum, multiplier : CARDINAL;
  numDigits, index, base : CARD8T;

BEGIN (* BinStrToCard8 *)
  FormatIO.RemoveLeadingExtras (str);
  base := 2;
  numDigits := LENGTH (str);
  IF numDigits > 8 THEN 
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  END;
  multiplier := 1;
  index := numDigits;
  accum := 0;
  REPEAT 
    index := index - 1;
    IF ( ORD (str [index]) >= ORD ("0") ) AND ( ORD (str [index]) <= ORD ("1") ) THEN
      accum := accum + ( ORD (str [index]) - ORD ("0") ) * multiplier;
      multiplier := multiplier * base;
    ELSE
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END;
  UNTIL index = 0;
  (* Unlike the base 10 one we cannot overflow at this point as only 8 chars are used *)
  card8 := accum;
  res := IOStatusT{ FALSE, IOStatus.Success};;
  RETURN;
END BinStrToCard8;


PROCEDURE Card8ToDecStr (card8: CARD8T; VAR str: ARRAY OF CHAR);

VAR
  num, mod100, div100, div10, mod10, count : CARD8T;

BEGIN (* Card8ToDecStr *)
  num := card8;
  div100 := num DIV 100; (* first digit *)
  mod100 := num MOD 100; (* whats left *)
  div10 := mod100 DIV 10; (* second digit *)
  mod10 := mod100 MOD 10; (* third digit *)
  count := 0;
  IF div100 # 0 THEN
    str[count] := CHR (ORD("0") + div100);
    count := count + 1;
    str[count] := CHR (ORD("0") + div10);
    count := count + 1;
  ELSIF div10 # 0 THEN
    str[count] := CHR (ORD("0") + div10);
    count := count + 1
  END;
  str[count] := CHR (ORD("0") + mod10);
  str[count+1] := CHR(0)
END Card8ToDecStr;


PROCEDURE Card8ToBinStr (card8: CARD8T; VAR str: ARRAY OF CHAR);

VAR
  hi, lo : CARD8T;
  nstr : NibbleStr;

BEGIN (* Card8ToBinStr *)
  hi := card8 DIV 16;
  lo := card8 MOD 16;
  ARRAYOFCHAR.Assign ( str, HexDigitToNibbleStr ( hi ) );
  ARRAYOFCHAR.Concat ( str, str, HexDigitToNibbleStr ( lo ) );
END Card8ToBinStr;


PROCEDURE Card8ToHexStr (card8: CARD8T; VAR str: ARRAY OF CHAR);

VAR
  num, hi, lo : CARD8T;
  count, strCount : CARDINAL;

BEGIN (* Card8ToHexStr *)
  count := 1;
  strCount := 0;
  REPEAT
      count := count - 1;
      hi := card8 DIV 16;
      lo := card8 MOD 16;
      str[strCount] := HexDigitToHexChar (hi);
      str[strCount+1] := HexDigitToHexChar (lo);
      strCount := strCount + 2;
    UNTIL count = 0;
    str[strCount] := CHR(0)
END Card8ToHexStr;

PROCEDURE HexStrToCard8s (hexStr: ARRAY OF CHAR; VAR card8s: ARRAY OF CARD8T; VAR res : IOStatusT);
(* removes any leading spaces, zeros, and the 0x if present and converts the rest of the inpuit string to an array
of CARD8 values *)
VAR
  lo, hi, n : CARD8T;
  sIndex, cIndex, numCard8s, maxcIndex : INTEGER;

BEGIN (* HexStrToOcts *)
  FormatIO.RemoveLeadingExtras (hexStr);
  maxcIndex := CAPACITYc8 ( card8s );
  cIndex := 0; (* start with lowest octet in the output array *)
  sIndex := LENGTH (hexStr) - 1; (* and at the last hex digit char in the string *)
  REPEAT
    IF (sIndex >= 0) AND ((hexStr[sIndex] = "'") OR (hexStr[sIndex] = " ")) (* skip any spacer chars *)
      THEN
        sIndex := sIndex - 1;
      END; (*IF*)

    IF sIndex >= 0 (* still got more chars in hex hexStr *)
      THEN
        HexCharToCard8 ( hexStr[sIndex] , lo, res );
      ELSE
        lo := 0
      END; (*IF*)
    IF res.failed THEN RETURN ELSE sIndex := sIndex -1 END;
    IF sIndex >= 0 (* still got more chars in hex hexStr *)
      THEN
        HexCharToCard8 ( hexStr[sIndex] , hi, res );
      ELSE
        hi := 0
      END; (*IF*)
    IF res.failed THEN RETURN ELSE sIndex := sIndex -1 END;
    card8s [cIndex] := 16 *  hi + lo;
    cIndex := cIndex + 1;
  UNTIL (cIndex = maxcIndex ); (* keeps going until all card8s set--zeros any extra leading ones *)
  IF sIndex >= 0 THEN (* but if ran out of car8s for the size of the hex string *)
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
  END; (*IF*)

END HexStrToCard8s;

PROCEDURE Card8ArrayToDecStr (card8s : ARRAY OF CARD8T; separator : ARRAY OF CHAR; VAR dest : ARRAY OF CHAR);

VAR
  c8dStr : ARRAY [0..3] OF CHAR;
  aCount, dCount, dRoom, sepLength, numToAppend : CARDINAL;

BEGIN
  sepLength := LENGTH (separator);
  aCount := CAPACITYc8 (card8s);
  dCount := 0;
  dRoom := CAPACITYch (dest) -2;
  dest[0] := ISO646.NUL;
  REPEAT
    aCount := aCount -1;
    Card8ToDecStr (card8s [aCount], c8dStr);
    numToAppend := (LENGTH (c8dStr) + sepLength);
    IF dRoom > numToAppend THEN
      ARRAYOFCHAR.Concat ( dest, dest, c8dStr );
      IF (LENGTH (separator) # 0) AND (aCount # 0) THEN
        ARRAYOFCHAR.Concat ( dest, dest, separator );
      END; (* IF dRoom *)
      dRoom := dRoom - numToAppend;
    ELSE
      RETURN
    END (* IF dRoom *)
  UNTIL aCount = 0;

END Card8ArrayToDecStr;


PROCEDURE Card8ArrayToHexStr (card8s : ARRAY OF CARD8T; separator : ARRAY OF CHAR; VAR dest : ARRAY OF CHAR);

VAR
  c8dStr : ARRAY [0..3] OF CHAR;
  aCount, dCount, dRoom, sepLength, numToAppend : CARDINAL;

BEGIN
  sepLength := LENGTH (separator);
  aCount := CAPACITYc8 (card8s);
  dCount := 0;
  dRoom := CAPACITYch (dest) -2;
  dest[0] := ISO646.NUL;
  REPEAT
    aCount := aCount -1;
    Card8ToHexStr (card8s [aCount], c8dStr);
    numToAppend := (LENGTH (c8dStr) + sepLength);
    IF dRoom > numToAppend THEN
      ARRAYOFCHAR.Concat ( dest, dest, c8dStr );
      IF (LENGTH (separator) # 0) AND (aCount # 0) THEN
        ARRAYOFCHAR.Concat ( dest, dest, separator );
      END; (* IF dRoom *)
      dRoom := dRoom - numToAppend;
    ELSE
      RETURN
    END (* IF dRoom *)
  UNTIL aCount = 0;

END Card8ArrayToHexStr;


PROCEDURE Card8ArrayToBinStr (card8s : ARRAY OF CARD8T; separator : ARRAY OF CHAR; VAR dest : ARRAY OF CHAR);

VAR
  c8dStr : ARRAY [0..3] OF CHAR;
  aCount, dCount, dRoom, sepLength, numToAppend : CARDINAL;

BEGIN
  sepLength := LENGTH (separator);
  aCount := CAPACITYc8 (card8s);
  dCount := 0;
  dRoom := CAPACITYch (dest) -2;
  dest[0] := ISO646.NUL;
  REPEAT
    aCount := aCount -1;
    Card8ToBinStr (card8s [aCount], c8dStr);
    numToAppend := (LENGTH (c8dStr) + sepLength);
    IF dRoom > numToAppend THEN
      ARRAYOFCHAR.Concat ( dest, dest, c8dStr );
      IF (LENGTH (separator) # 0) AND (aCount # 0) THEN
        ARRAYOFCHAR.Concat ( dest, dest, separator );
      END; (* IF dRoom *)
      dRoom := dRoom - numToAppend;
    ELSE
      RETURN
    END (* IF dRoom *)
  UNTIL aCount = 0;

END Card8ArrayToBinStr;


END Card8Strs.
