IMPLEMENTATION MODULE  OctetStrs;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      04/14/2014  First implementation

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYch, CAPACITYo, OCTET;
FROM UNSAFE IMPORT SETBIT;
IMPORT CARD8; FROM CARD8 IMPORT CARD8T;
IMPORT Card8Strs;

  (* ISO dependencies *)
IMPORT LongWholeStrs;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* RS Utils *)
IMPORT FormatIO;
IMPORT ARRAYOFCHAR, CharArray;

CONST
  zeroAsc = ORD ("0");
  maxCardDiv10  = MAX ( CARD8T ) / 10;
  lastMaxCardDigitChar = CHR( (MAX (CARD8T) REM 10) + zeroAsc);

TYPE
  NibbleStr = ARRAY [0..4] OF CHAR;

(* utility local procs *)


PROCEDURE DecHexDigitToHexChar (num : CARD8T) : CHAR;
(* Pre: the number passed must be 0..16 *)

BEGIN
  IF num < 10
    THEN RETURN CHR (ORD ("0") + num)
    ELSE RETURN CHR (ORD ("A") + num-10)
  END(*IF*)
END DecHexDigitToHexChar;


PROCEDURE HexDigitToNibbleStr (num : CARD8T) : NibbleStr;
(* Pre: the number passed must be 0..16 *)

VAR
  NStr : NibbleStr;
  
BEGIN
  CASE num OF
      0 : NStr := "0000"
    | 1 : NStr := "0001"
    | 2 : NStr := "0010"
    | 3 : NStr := "0011"
    | 4 : NStr := "0100"
    | 5 : NStr := "0101"
    | 6 : NStr := "0110"
    | 7 : NStr := "0111"
    | 8 : NStr := "1000"
    | 9 : NStr := "1001"
    | 10 : NStr := "1010"
    | 11 : NStr := "1011"
    | 12 : NStr := "1100"
    | 13 : NStr := "1101"
    | 14 : NStr := "1110"
    | 15 : NStr := "1111"
  END; (* case *)
  RETURN NStr
END HexDigitToNibbleStr;

(* the ones needing implementing *)


PROCEDURE StrToOcts (str: ARRAY OF CHAR; VAR octs: ARRAY OF OCTET; VAR res : IOStatusT);

BEGIN
  IF (str[0] = "0")
    THEN
      IF (str [1] = "x")
        THEN
          HexStrToOcts (str, octs, res);
          RETURN
        ELSIF str [1] = "b" THEN
          BinStrToOcts (str, octs, res);
          RETURN
        END;
     ELSE DecStrToOcts (str, octs, res)
     END(*IF*)
 END StrToOcts;


PROCEDURE DecStrToOcts (str: ARRAY OF CHAR; VAR octs: ARRAY OF OCTET; VAR res : IOStatusT);
TYPE
  oArray = ARRAY [0..SIZE (LONGCARD) -1] OF OCTET;

VAR
  n : LONGCARD;
  temp : oArray;
  count : CARDINAL;

BEGIN (* StrToOct *)
  FormatIO.RemoveLeadingSpaces (str);
  LongWholeStrs.StrToCard (str, n, res);
  IF SIZE (octs) < SIZE (n) THEN
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  END;
    temp := SYSTEM.CAST (oArray, n);
    count := 0;
    WHILE (count < SIZE (octs) ) AND (count < SIZE (temp) ) DO
      octs [count] := temp [count];
      count := count + 1
    END(*WHILE*);
  res := IOStatusT{ FALSE, IOStatus.Success}
END DecStrToOcts;


PROCEDURE DecStrToOct (str: ARRAY OF CHAR; VAR oct: OCTET; VAR res : IOStatusT);
VAR
  n : CARD8T;

BEGIN (* StrToOct *)
   FormatIO.RemoveLeadingSpaces (str);
  IF LENGTH (str) > 3 THEN
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
    RETURN
  ELSE
    Card8Strs.StrToCard8 (str, n, res);
    IF n > 255 THEN
      res := IOStatusT{ TRUE, IOStatus.OutOfRange};;
      RETURN
    END;
    oct := SYSTEM.CAST (OCTET, n)
  END;
END DecStrToOct;


PROCEDURE HexStrToOcts (str: ARRAY OF CHAR; VAR octs: ARRAY OF OCTET; VAR res : IOStatusT);
(* removes any leading spaces, zeros, and the 0x, 0b, or ox and converts the rest of the inpuit string to octet values *)
VAR
  lo, hi, n, numOcts : CARD8T;
  sindex, oindex : INTEGER;

BEGIN (* Hex StrToOcts *)
  FormatIO.RemoveLeadingExtras (str);
  oindex := 0; (* start with lowest octet in the input array *)
  sindex := LENGTH (str) - 1; (* and at the last hex digit in the string *)
  REPEAT
    IF (sindex >= 0) AND ((str[sindex] = "'") OR (str[sindex] = " ")) (* skip any spacer chars *)
      THEN
        sindex := sindex - 1;
      END; (*IF*)

    IF sindex >= 0 (* still got more chars in hex str *)
      THEN
        Card8Strs.HexCharToCard8 ( str[sindex] , lo, res );
      ELSE
        lo := 0
      END; (*IF*)    IF res.failed THEN RETURN ELSE sindex := sindex -1 END;
    IF sindex >= 0 (* still got more chars in hex str *)
      THEN
        Card8Strs.HexCharToCard8 ( str[sindex] , hi, res );
      ELSE
        hi := 0
      END; (*IF*)
    IF res.failed THEN RETURN ELSE sindex := sindex -1 END;
    n := 16 *  hi + lo;
    octs [oindex] := SYSTEM.CAST (OCTET, n);
    oindex := oindex + 1;
  UNTIL (oindex = VAL (INTEGER, SIZE (octs) ) ); (* keeps going until all octets set--zeros any extra leading ones *)
  IF sindex >= 0 THEN (* but if ran out of octets for the size of the hex string *)
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
  END; (*IF*)

END HexStrToOcts;


PROCEDURE BinStrToOcts (str: ARRAY OF CHAR; VAR octs: ARRAY OF OCTET; VAR res : IOStatusT);

VAR
  bitIndex, maxBitIndex, strIndex, maxStrIndex : CARDINAL;

BEGIN (* StrToOct *)
  FormatIO.RemoveLeadingExtras (str);
  maxBitIndex := VAL (INTEGER, 8 * SIZE (octs) - 1); (* start with last octet in the input array *) 
  maxStrIndex := LENGTH (str) - 1; (* and at the first bin digit *) 
  strIndex := LENGTH (str);
  bitIndex := 0;

  WHILE (strIndex > 0 ) AND (bitIndex <= maxBitIndex) DO (* still got more chars in bin str *)
        strIndex := strIndex - 1;

      (* skip any spacer chars *)
    WHILE (strIndex > 0 ) AND ((str[strIndex] = "'") OR (str[strIndex] = " ")) DO 
      strIndex := strIndex - 1
    END;
    (* exit??? if it hits 0 *)
       (* process binary chars *)
    IF str[strIndex] = "1" THEN
      SETBIT (octs, bitIndex, TRUE)
    ELSIF str[strIndex] = "0" THEN
      SETBIT (octs, bitIndex, FALSE)
    ELSE (* a non binary char *)
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      RETURN
    END (*IF*);
    bitIndex := bitIndex + 1
  END; (* while *)
  IF (strIndex < maxStrIndex) AND (bitIndex = maxBitIndex) THEN (* ran out of bit room *)
     res := IOStatusT{ TRUE, IOStatus.OutOfRange};
  ELSE
    WHILE bitIndex <= maxBitIndex DO
      SETBIT (octs, bitIndex, FALSE);
      bitIndex := bitIndex + 1
    END; 
    res := IOStatusT{ FALSE, IOStatus.Success}
  END
END BinStrToOcts;


PROCEDURE OctToDecStr (oct: OCTET; VAR str: ARRAY OF CHAR);

VAR
  num, mod100, div100, div10, mod10 : CARD8T;
  strCount, strSize : LONGCARD;

BEGIN (* OctToDecStr *)
  strSize := CAPACITYch ( str );
  strCount := 0;
  IF strSize >= 4 THEN
    num := SYSTEM.CAST (CARD8T, oct);
    div100 := num DIV 100; (* first digit *)
    mod100 := num MOD 100; (* whats left *)
    div10 := mod100 DIV 10; (* second digit *)
    mod10 := mod100 MOD 10; (* third digit *)
    IF div100 # 0 THEN
      str[strCount] := CHR (ORD("0") + div100);
      strCount := strCount + 1;
      str[strCount] := CHR (ORD("0") + div10);
      strCount := strCount + 1;
    ELSIF div10 # 0 THEN
      str[strCount] := CHR (ORD("0") + div10);
      strCount := strCount + 1
    END;
    str[strCount] := CHR (ORD("0") + mod10);
    str[strCount+1] := CHR(0)
  ELSE (* string too short return empty *)
    str[strCount] := CHR(0)
  END (* if *)

END OctToDecStr;


PROCEDURE OctsToBinStr (octs: ARRAY OF OCTET; VAR str: ARRAY OF CHAR);

VAR
  hi, lo, card8 : CARD8T;
  nstr : NibbleStr;
  octCount, strSize, strCount : LONGCARD;


BEGIN (* OctsToBinStr *)
  strSize := CAPACITYch ( str );
  octCount := CAPACITYo (octs);
  strCount := 0;
  str [0] := CHR (0);
  IF strSize >= 16 * octCount + 1 THEN (* else it returns with an empty string *)
    REPEAT
      octCount := octCount - 1;
      card8 := SYSTEM.CAST (CARD8T, octs[octCount]);
      hi := card8 DIV 16;
      lo := card8 MOD 16;
      ARRAYOFCHAR.Concat ( str, str, HexDigitToNibbleStr ( hi ) );
      ARRAYOFCHAR.Concat ( str, str, HexDigitToNibbleStr ( lo ) );
      strCount := strCount + 8;
    UNTIL octCount = 0;
    str[strCount + 1] := CHR(0)
  END (*if*);
END OctsToBinStr;


PROCEDURE OctsToHexStr (octs: ARRAY OF OCTET; VAR str: ARRAY OF CHAR);

VAR
  num, hi, lo : CARD8T;
  octCount, strSize, strCount : LONGCARD;

BEGIN (* OctsToHexStr *)
  strSize := CAPACITYch ( str );
  octCount := CAPACITYo (octs);
  strCount := 0;
  str [0] := CHR (0);
  IF strSize >= 2 * octCount + 1 THEN (* else it returns with an empty string *)
    REPEAT
      octCount := octCount - 1;
      num := SYSTEM.CAST (CARD8T, octs[octCount]);
      hi := num DIV 16;
      lo := num MOD 16;
      str [strCount] := DecHexDigitToHexChar (hi);
      str [strCount+1] := DecHexDigitToHexChar (lo);
      strCount := strCount + 2;
    UNTIL octCount = 0;
    str[strCount] := CHR(0);
  END (*if*);
END OctsToHexStr;


END OctetStrs.
