(* This library was created because the perror function
 does not work  as it should on the Mac
 It is part of the Modula-2 R10 library and is
 copyright 2017 R. Sutcliffe and B. Kowarsch
 Last revision 2017 11 19 *)

IMPLEMENTATION MODULE Perror;

IMPORT FileIO, StdIO, CharIO, errno;
IMPORT IOSIZE; FROM IOSIZE IMPORT IOSIZET;
IMPORT ISO646, ARRAYOFCHAR, WholeStrs;

TYPE
  Str255 = ARRAY [0..255] OF CHAR;
CONST
  undefinedErrorMsg = "Error number undefined or has no stored message.";

VAR
  numWritten : IOSIZET;


PROCEDURE perror(str: ARRAY OF CHAR);

VAR
  errPtr : errno.INT32Ptr;
  errMsg : errno.SysErrMsg;
  errnum : errno.SysErrNum;

BEGIN
 CharIO.WriteStr  ( StdIO.err (), str);
 CharIO.WriteStr  ( StdIO.err (), ": ");
  errPtr := errno.__error();
  errnum := errPtr^;
  fetchSysErrMessage (errnum, errMsg);
  CharIO.WriteStr (StdIO.err (), errMsg);
  CharIO.WriteStr ( StdIO.err (), "\n" );
END perror;


PROCEDURE prerr ( str: ARRAY OF CHAR );
(* Prints only the passed string to stderr. Respects the escape sequence 
 *)

BEGIN
  CharIO.WriteStr  ( StdIO.err (), str);
END prerr;


PROCEDURE prerrnum (num : INTEGER );
(* Prints only the passed integer to stderr. Puts a space before and after *)
VAR
  str : Str255;

BEGIN
  WholeStrs.IntToDecStr ( num, str );
  FileIO.WriteOctet ( StdIO.err (), ISO646.SPACE );
 CharIO.WriteStr  ( StdIO.err (), str );
  FileIO.WriteOctet ( StdIO.err (), ISO646.SPACE );
END prerrnum;


PROCEDURE fetchSysErrMessage ( errnum: INTEGER; VAR str: ARRAY OF CHAR );

VAR
  errPtr : errno.SysErrMsgPtr;
  errMsg : errno.SysErrMsg;

BEGIN
  errPtr := errno.strerror(errnum);
  errMsg := errPtr^;
  str [0] := CHR(0);
  IF (errnum <= errno.MaxSysError) & (errMsg[0] #ISO646.NUL) THEN
    ARRAYOFCHAR.Concat( str, errMsg, str );
  ELSE
    ARRAYOFCHAR.Concat( str,undefinedErrorMsg, str );
  END;
END fetchSysErrMessage;

BEGIN
END Perror.


