(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE PtrSet;

(* Generic Pointer Set *)

IMPORT UNSAFE;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Transition IMPORT CAPACITYad;

CONST
  maxPtrs = 500;
  
TYPE
  PtrSet = POINTER TO PointerStore;
  PointerArray = ARRAY [0..maxPtrs - 1] OF UNSAFE.ADDRESS;
  PointerStore =
    RECORD
      Pointers : PointerArray;
      mostRecent : UNSAFE.ADDRESS;
      count : LONGCARD
    END;

VAR
  nilPtrArray : PointerArray;


PROCEDURE NullOutPtrArray ( VAR ptrs : ARRAY OF UNSAFE.ADDRESS );
  (* replaces everything in the string with CHR (0) *)
VAR
  count : CARDINAL; (* Counting Variable *)
  hiStr : CARDINAL; (* hold high of string for comparisons *)

BEGIN
  hiStr  := CAPACITYad (ptrs) - 1;
  count := 0;
  WHILE (count <= hiStr)
    DO
      ptrs[count] := NIL;
    count := count + 1
    END
END NullOutPtrArray;



PROCEDURE Swap (VAR ptr1, ptr2 : UNSAFE.ADDRESS);

VAR
  temp : UNSAFE.ADDRESS;

BEGIN
  temp := ptr1;
  ptr1 := ptr2;
  ptr2 := temp
END Swap;


PROCEDURE New ((* NEW*) VAR set : PtrSet );

BEGIN
  NEW (set);
  IF set # NIL THEN
    set^.Pointers := nilPtrArray;
    set^.mostRecent := NIL;
    set^.count := 0
  END;
END New;


PROCEDURE isEmpty ( set : PtrSet ) : BOOLEAN;

BEGIN
  IF set = NIL THEN RETURN TRUE
  ELSE RETURN ( set^.count = 0 )
  END
END isEmpty;


PROCEDURE isPresent ( set : PtrSet;  ptr : (*CAST*) UNSAFE.ADDRESS ) : BOOLEAN;

VAR
  index : LONGCARD;

BEGIN
  IF set = NIL THEN RETURN FALSE END; (* nothing happens *)
  index := 0;
  WHILE index < set^.count DO
    IF set^.Pointers [index] = ptr THEN RETURN TRUE END;
    index := index + 1;
  END  (* while *);
  RETURN FALSE (* if we get here we didn't find it. *)
END isPresent;


PROCEDURE count ( set : PtrSet ) : LONGCARD;

BEGIN
  IF set = NIL THEN RETURN 0
  ELSE RETURN set^.count
  END
END count;


PROCEDURE Insert ( set : PtrSet; ptr : (*CAST*) UNSAFE.ADDRESS );
(* we maintain a compact list, so the count slot is the next available *)

BEGIN
  IF (set = NIL) OR isPresent (set, ptr)  THEN RETURN END; (* nothing happens *)
    set^.Pointers [set^.count] := ptr;
    set^.mostRecent := ptr;
    set^.count := set^.count + 1;
END Insert;


PROCEDURE GetMostRecent ( set : PtrSet; VAR ptr : (*CAST*) UNSAFE.ADDRESS );
  (* return most recently inserted pointer from pointer set, NIL if empty *)

BEGIN
  IF set = NIL THEN ptr := NIL
  ELSE ptr := set^.mostRecent
  END

END GetMostRecent;


PROCEDURE Remove ( set : PtrSet; ptr : (*CAST*) UNSAFE.ADDRESS );

VAR
  index : LONGCARD;

BEGIN
  IF (set = NIL) OR (ptr = NIL) OR isEmpty (set ) THEN RETURN END; (* nothing happens *)
  index := 0;
  WHILE index < set^.count DO
    IF set^.Pointers [index] = ptr THEN
      set^.Pointers [index] := NIL;
      IF set^.count # index THEN
        Swap ( set^.Pointers [index], set^.Pointers [set^.count] )  (* compact the array *)
      END;
      set^.count := set^.count - 1;
      IF ptr = set^.mostRecent THEN set^.mostRecent := set^.Pointers [set^.count] END;
        (* it may not be the second most recent, but... *)
    ELSE
      index := index + 1 (* keep going *)
    END (* if *)
  END; (* while *)
  (* if we get here without removing anything, well it wasn't there...tough *)
END Remove;


PROCEDURE Release ( VAR set : PtrSet );

BEGIN
  IF set # NIL THEN DISPOSE (set); RETURN END
END Release;

BEGIN (* main *)
  NullOutPtrArray ( nilPtrArray )
END PtrSet.
