IMPLEMENTATION MODULE RealStrs;

(* =========================================
          Implementation  1993
                by R. Sutcliffe
       (Portions coded by G. Tischer)
        Trinity Western University
7600 Glover Rd., Langley, BC Canada V3A 6H4
         e-mail: rsutc@twu.ca
    Last modification date 2017 11 02
 ===========================================*)

(* 1993 10 20 first cut
1994 02 14 RS fixed bugs in indexing of strings pointed out by Norm Black; must check high first
cleaned up Gord's code closer to our standards
1994 05 18 first cut at a p1 version
  digits used for assignment and loop control - changed
  digits < place changed to digits < ABS (place)
  in RealToEng changed loop counter to count2 to avoid threatening
1994 09 02 Tried to get some speed increases in the thing
1996 08 08 fixed an error in RealToScientific in the BuildZero part
2014 04 23 fixed the error in the places part for neg places
  But noted that this is not very efficient code
  also fixed the round off so it works properly.
  and many small bugs in the eng output
2014 06 02 no longer ISO
  separated scientific from float, fixed many small errors, improved performance
2014 06 16 fixed errors in fixed
2016 08 26 removed dependence on ISORealConv and used new algorithm with a record format for the conversions
     as they were not very accurate due to a weak algorithm.
        This cut the code size to a third of what it was. Gord's code all gone.
        The module now bears little resemblance to the ISO one we did back in the day.
        2017 11 02 Changed name to avoind conflicts with P1 debug library
*)

  (* REAL/string conversions *)

(* the string form of a signed fixed-point real number is
     ["+" | "-"], decimal digit, {decimal digit}, [".", {decimal digit}]
*)

(* the string form of a signed floating-point real number is
     signed fixed-point real number, "e", ["+" | "-"], decimal digit, {decimal digit}
*)

IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT SET32;
FROM Transition IMPORT CAPACITYch;
FROM CARDINAL32 IMPORT CARDINAL32;

IMPORT ISO646, WholeStrs, CardinalMath;
IMPORT ARRAYOFCHAR, CharArray;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;


(* ============== *)
(* the material in the next section is specific to the type of REAL *)
CONST
  rMaxExponent = 38;
  signBit = 31;
  highestExpPower = 128; (* two to the seventh *)
  sigFigs2StartBit = 22;
  exp2StartBit = 30;
  exp2EndBit = 23;
  bias = 127;
  strHigh = 255;
  expStrLen = 4; (* expChar, sign, and two digits *)
  zero = 0.0;
  expChar = "e";
 
TYPE
  CardType = CARDINAL;
  RealType = REAL;
  RealToCardCastingType = CARDINAL32;
  RealRec = RECORD;
    neg : BOOLEAN;
    exp10 : INTEGER;
    sigFigs : CardType
  END;
  strType = ARRAY [0..strHigh ] OF CHAR;
(* ======end type specific material======== *)

VAR
  sigFigHighPowerOf10 : CardType;
  maxRealRec : RealRec;


PROCEDURE realToRealRec (theReal : RealType; VAR realRec : RealRec );
(* Feed this procedure a real number and get back a record with the sign, base 10 integer unbiased
exponent, and a cardinal containing the significant figures. *)

VAR
  exp2, intAddend : INTEGER;
  bitCount, numSigFigsB4Round : CARDINAL;
  addend, tailDigits : CardType;

BEGIN (* realToRealRec *)
   (* we need to examine bits so *)
   IF (theReal # 0.0) THEN
     realRec.neg := theReal < 0.0;
     exp2 := 0;
     (* Build exponent by looking at the exponent bits (biased) starting at bit exp2StartBit *)
     intAddend := highestExpPower; (* amount to add if highest bit is set *)
     bitCount := exp2StartBit;
     WHILE bitCount >= exp2EndBit DO
       IF UNSAFE.BIT (theReal, bitCount) THEN
        exp2 := exp2 + intAddend
       END;
     (* and move along to the next exponent bit *)
     intAddend := intAddend DIV 2;
     bitCount := bitCount - 1;
    END (* while *);
     exp2 := exp2 - bias; (* unbias it *)

     (* build realRec.sigFigs as a cardinal  *)
     addend := sigFigHighPowerOf10; (* first sig fig always 1 *)
     realRec.sigFigs := addend;
     (* begin by looking at the bit position of the most significant bit of the sig figs in base 2 *)
     (* these contribute .5, .25. .125, etc, but we are building this as a cardinal with whole numbers *)
     bitCount := sigFigs2StartBit + 1; (* will reduce at start of loop *)

     (* loop through the sig fig bits of the original *)
     (* adding in a value to the sig figs for each bit set *)
     WHILE bitCount > 0 DO
      bitCount := bitCount - 1; (* so last one used is zero *)
      addend := (addend + 1) DIV 2; (* add half of the past addend for every digit with even/odd rounding *)
      IF UNSAFE.BIT (theReal, bitCount) THEN
        realRec.sigFigs := realRec.sigFigs + addend
      END; (* if *)
     END; (*while *)

     (* now look at rounding this off  *)
     (* the approach here is an attempt to be predictive--using the last two bits as though they
    influenced two more digits. This seems to give better results than other approaches.
    One must recall that accuracy is only to the 7th through 9th sig fig.*)
      IF ( UNSAFE.BIT ( theReal, 1) ) OR ( UNSAFE.BIT ( theReal, 0) ) THEN realRec.sigFigs := realRec.sigFigs + addend DIV 4 END;
      numSigFigsB4Round := CardinalMath.nearestLargerPowerOf10 (realRec.sigFigs);

     (* next we have to convert the exponent from base 2 to base 10. *)
     (* this requires the sig figs to be tranfsormed as well *)
    realRec.exp10 := VAL ( INTEGER, CardinalMath.nearestLargerPowerOf10 (realRec.sigFigs) - numSigFigsB4Round);

     (* as long as that base 2 exponent is positive, then whenever the decimal version
    occupies the first sig fig, divide by ten and increment the base 10 exponent.
    Else, multiply by 2 and decrement the base 2 exponent. *)
    IF exp2 > 0 THEN
      IF realRec.sigFigs >= sigFigHighPowerOf10 THEN (* make room to work *)
         realRec.sigFigs := realRec.sigFigs DIV 10;
      END;
      REPEAT
        realRec.sigFigs := realRec.sigFigs * 2;
        exp2 := exp2 - 1;
        IF realRec.sigFigs >= sigFigHighPowerOf10
          THEN
            realRec.sigFigs := realRec.sigFigs DIV 10;
            realRec.exp10 := realRec.exp10 + 1
          END; (* if realRec.sigFigs *)
        UNTIL exp2 = 0;

     (* OTOH as long as that base 2 exponent is negative, then whenever the decimal version
    does not occupy the first sig fig, multipy by ten and decrement the base 10 exponent.
    Else, divide by 2 and increment the base 2 exponent. *)
      ELSIF exp2 < 0 THEN
        IF realRec.sigFigs < sigFigHighPowerOf10 THEN (* make room to work *)
          realRec.sigFigs := realRec.sigFigs * 10
        END;
        REPEAT
          realRec.sigFigs := realRec.sigFigs DIV 2;
          exp2 := exp2 + 1;
          IF realRec.sigFigs < sigFigHighPowerOf10
            THEN
              realRec.sigFigs := realRec.sigFigs * 10;
              realRec.exp10 := realRec.exp10 - 1;
            END; (* if realRec.sigFigs *)
        UNTIL exp2 = 0
    END; (* IF exp2 > 0 *)
  ELSIF theReal = 0.0 THEN
    realRec.sigFigs := 0;
    realRec.exp10 := 0
  END (* if theReal # 0.0 *)
END realToRealRec;


PROCEDURE realRecToReal (realRec : RealRec; VAR theReal : RealType);
(* feed this procedure a real with its parts in a real rec (sign, integer exponent, and
a cardinal representing the sig figs, and it returns a real in IEEE format *)
(* In large part the code involves a reversing of the above code *)
VAR
  exp2, intSubtrahend : INTEGER;
  bitCount : CARDINAL;
  subtrahend : CardType;
  cCard : RealToCardCastingType;
  needExpCarry : BOOLEAN;

BEGIN (* realRecToReal *)
  theReal := 0.0;
  IF realRec.neg THEN UNSAFE.SETBIT ( theReal, signBit, TRUE ); END;

  (* now move the sig figs over to our canonical position; *)
  WHILE realRec.sigFigs < sigFigHighPowerOf10 DO
    realRec.sigFigs := realRec.sigFigs * 10
  END;

  (* so we now have a digit in that spot, though not necessarily a 1 *)
   (* the next task is to convert the exponent from base 10 to base 2. *)
      (* this requires the sig figs to be transformed as well *)
      (* as long as that base 10 exponent is positive, then whenever the decimal version
    occupies the first sig fig, divide by 2 till it does not and increment the base 2 exponent.
    Else, multiply by 10 and decrement the base 10 exponent. *)
    exp2 := 0;
  IF realRec.exp10 > 0 THEN
    REPEAT
      WHILE (realRec.sigFigs >= sigFigHighPowerOf10 DIV 5 )  DO
        realRec.sigFigs := (realRec.sigFigs) DIV 2;
        exp2 := exp2 + 1;
      END;
      IF (realRec.sigFigs < sigFigHighPowerOf10 DIV 5) THEN
        realRec.sigFigs := realRec.sigFigs * 10;
        realRec.exp10 := realRec.exp10 - 1
      END;
    UNTIL realRec.exp10 = 0;

   (* OTOH as long as that base 10 exponent is negative, then whenever the decimal version
    does not occupy the first sig fig, multiply by ten and decrement the base 10 exponent.
    Else, divide by 2 and increment the base 2 exponent. *)
    ELSIF realRec.exp10 < 0 THEN
      REPEAT
        IF realRec.sigFigs >= sigFigHighPowerOf10 THEN
          realRec.sigFigs := realRec.sigFigs DIV 10
        END;
        WHILE realRec.sigFigs < sigFigHighPowerOf10
          DO
            realRec.sigFigs := realRec.sigFigs * 2;
            exp2 := exp2 - 1
          END; (* while realRec.sigFigs *)
        IF realRec.sigFigs >= sigFigHighPowerOf10 THEN
          realRec.sigFigs := (realRec.sigFigs ) DIV 10;
          realRec.exp10 := realRec.exp10 + 1;
        END;
      UNTIL realRec.exp10 = 0;

    ELSE (* base 10 exp started at 0 *)
      WHILE (realRec.sigFigs >= sigFigHighPowerOf10 * 2 )
        DO
          realRec.sigFigs := realRec.sigFigs DIV 2;
          exp2 := exp2 + 1
        END; (* while realRec.sigFigs *)

     END; (* if realRec.exp10 *)
  (* ensure that first sig fig is in the leftmost usable place *)
  IF realRec.sigFigs < sigFigHighPowerOf10 THEN
    realRec.sigFigs := realRec.sigFigs * 10
  END;

(* now we have the local copy of realRec in base two sig figs and exponent *)
(* so we need to construct the real bit pattern, first on the sig figs *)
  subtrahend := sigFigHighPowerOf10; (* first sig fig always 1; so just remove it *)
  realRec.sigFigs := realRec.sigFigs - subtrahend;

  (* begin by looking at the bit position of the most significant bit of the sig figs in base 2 *)
   (* these contribute .5, .25. .125, etc, but we are deconstructing this as a cardinal with whole numbers *)
   bitCount := sigFigs2StartBit + 1; (* will reduce at start of loop *)

   (* loop through the sig fig bits of the original *)
   WHILE (bitCount > 0) AND (realRec.sigFigs > 0) DO 
    bitCount := bitCount - 1; (* so last one used is zero *) 
    subtrahend := subtrahend DIV 2; (* subtract half of the past subtrahend for every digit *)
    IF realRec.sigFigs >= subtrahend THEN 
      realRec.sigFigs := realRec.sigFigs - subtrahend; 
      UNSAFE.SETBIT ( theReal, bitCount, TRUE );
    END; (* if *)
  END; (*while *)
(* Now we've finished it all, so investigate a roundoff *)
(* if remaining part is more than half the last subtrahend, add one card style *)

  IF realRec.sigFigs >= subtrahend DIV 2 THEN
    cCard := SYSTEM.CAST (RealToCardCastingType, theReal);
    INC (cCard);
    theReal := SYSTEM.CAST (RealType, cCard);
    (* and in the rare event that doing so went back to the last exponent bit spot, clear this *)
    needExpCarry := UNSAFE.BIT ( theReal, exp2EndBit );
    IF needExpCarry THEN
      UNSAFE.SETBIT ( theReal, exp2EndBit, FALSE );
      INC (exp2)
    END (* IF exp2EndBit *)
  END; (* IF realRec *)

  (* finally, we need to get the exponent in shape *)
  exp2 := exp2 + bias;  (* bias it *)
  intSubtrahend := highestExpPower;
  bitCount := exp2StartBit;
  WHILE bitCount >= exp2EndBit DO
    IF exp2 >= intSubtrahend THEN
      UNSAFE.SETBIT ( theReal, bitCount, TRUE );
      exp2 := exp2 - intSubtrahend;
    END;
     (* and move along to the next exponent bit *)
    intSubtrahend := intSubtrahend DIV 2;
    bitCount := bitCount - 1
  END (* while *);
END realRecToReal;


PROCEDURE StrToReal (str: ARRAY OF CHAR; VAR real: RealType; VAR res : IOStatusT);
  (* Ignores any leading spaces in str. If the subsequent characters in str are in the format of a signed real number, assigns a corresponding value to real. Assigns a value indicating the format of str to res.  *)
VAR
  count, count2, wholePlaces, sigPlaces, maxExp, hiStr : CARDINAL;
  exp, expAdj          : INTEGER;
  neg : BOOLEAN;
  rRec : RealRec;

BEGIN
  hiStr := LENGTH (str)-1;
  count := 0;
  wholePlaces := 0;
  sigPlaces := 0;
  real := 0.0;
  rRec.neg := FALSE;
  rRec.sigFigs := 0;
  rRec.exp10 := 0;
  expAdj := 0;
  res := IOStatusT{ FALSE, IOStatus.Success};

  (* if the string is empty then exit*)
  IF str[0] = CHR (0) THEN
      res := IOStatusT{ TRUE, IOStatus.StrEmpty};
      RETURN
  END;

  (* skip over whitespace *)
  WHILE (count <= hiStr) AND ISO646.isWhitespaceOrTab (str[count])
    DO
      INC (count)
    END;

  (* get the sign, if there is one *)
  IF count < hiStr
    THEN
      IF str[count] = "-"
        THEN
          rRec.neg := TRUE;
          INC (count);
        ELSIF str[count] = "+" THEN
          INC (count)
        END
    END;

(* in this section, we convert all the sig figs to a cardinal for the realRec *)
  (* first get digits before decimal place *)

  WHILE  (count <= hiStr) AND ISO646.isDigit (str[count]) AND (res.code # IOStatus.OutOfRange) AND (sigPlaces < rMaxSigFigs + 3)
    DO
      IF (rRec.sigFigs < MAX (CardType) DIV 10 ) AND (wholePlaces <= rMaxExponent + 1)
        THEN
          rRec.sigFigs := rRec.sigFigs * 10;                              (* go to next place *)
          rRec.sigFigs := rRec.sigFigs + (ORD (str[count]) - 48);    (* add the next digit *)
          INC (count); INC (wholePlaces);  INC ( sigPlaces );     (* count the digits before decimal *)
        ELSE
          res := IOStatusT{ TRUE, IOStatus.OutOfRange};
        END
    END;
(* following is a check on nothing but zero B4 decimal *)
  IF rRec.sigFigs = 0 THEN
    wholePlaces := 1;
    sigPlaces := 1;
    DEC (expAdj)
  END;
(* following only executes if we fell out of loop at max Sig figs + 2 but still with more to process *)
  WHILE  (count <= hiStr) AND ISO646.isDigit (str[count]) AND (res.code # IOStatus.OutOfRange) DO
    IF (wholePlaces <= rMaxExponent + 1) THEN
      INC (count); INC (wholePlaces);
        ELSE
          res := IOStatusT{ TRUE, IOStatus.OutOfRange};
        END
  END;

  IF res.code = IOStatus.OutOfRange
    THEN
      real := zero;
      RETURN
    END;

  (* if we are still going and there is a fractional part *)
  IF (count <= hiStr) AND (str[count] = ".")
    THEN
      INC (count);
      (* process continuing zeros *)
      WHILE (rRec.sigFigs = 0) AND (count <= hiStr) AND (str[count] = "0") DO
         INC (count);
        DEC (expAdj) (* because these tell us to reduce the exponent *)
      END;
      (* now past all leading zeros *)
      WHILE  (count <= hiStr) AND ISO646.isDigit (str[count]) AND (sigPlaces < rMaxSigFigs + 3)
        DO
          rRec.sigFigs := rRec.sigFigs * 10;                              (* go to next place *)
          rRec.sigFigs := rRec.sigFigs + (ORD (str[count]) - 48);    (* add the next digit *)
          INC (count); INC (sigPlaces);
        END; (* WHILE *)
    ELSE (* no decimal place *)
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      real := zero; 
      RETURN
    END;
    (* round  *)
  IF sigPlaces >= rMaxSigFigs + 2 THEN
    rRec.sigFigs := rRec.sigFigs + 5;
    (* and if this bumps the number of sig figs by, say, a string of nines going to 1000... *)
    IF (CardinalMath.nearestLargerPowerOf10 (rRec.sigFigs) > sigPlaces) THEN INC (expAdj) END;
    rRec.sigFigs := rRec.sigFigs DIV 10;  (* kill off that last sig fig now *)
  END;

  (* now we may get to here and have more figures in the string. If so, ignore them *)
  WHILE  (count <= hiStr) AND ISO646.isDigit (str[count]) DO
    INC (count )
  END;

 (* done the sig figs so move on to the exponent *)
  (* if an exponent *)
  IF (count < hiStr) AND (CAP (str[count]) = "E")
    THEN
      INC (count);
      (* if the exponent is negative *)
      IF (count < hiStr) AND  (str[count] = "-")
        THEN
          neg := TRUE;
          INC (count);
        ELSIF (count < hiStr) AND (str[count] = "+") THEN
          INC (count)
        END; (* IF *)

      (* get exponent digits *)
      maxExp := rMaxExponent - wholePlaces + 1; (* reduce by number of chars before dec place less one *)
      WHILE (count <= hiStr) AND ISO646.isDigit (str[count]) AND (res.code # IOStatus.OutOfRange)
        DO
          rRec.exp10 := rRec.exp10 * 10;                                     (* go to the next place *)
          rRec.exp10 := rRec.exp10 + (VAL (INTEGER, (ORD (str[count]) - 48) ) );                     (* and add the number *)
          INC (count)
         END; (* WHILE *)
       IF NOT res.failed
        THEN
          IF neg
            THEN
              rRec.exp10 := rRec.exp10 * (-1)
            END (* if neg *);
        ELSE
          real := zero;
        END; (* IF res.success *)
  END; (* IF count < hiStr) AND (CAP (str[count]) = "E" *)

  rRec.exp10 := rRec.exp10 + VAL (INTEGER, wholePlaces - 1 ) + expAdj;
  IF (rRec.sigFigs # 0) AND ( (ABS (rRec.exp10) < rMaxExponent) OR ((ABS(rRec.exp10) = rMaxExponent) AND (rRec.sigFigs <= maxRealRec.sigFigs)) )  (* check range *)
  THEN 
    realRecToReal (rRec, real); 
  ELSIF rRec.sigFigs = 0 THEN
    real := zero
  ELSE
    res := IOStatusT{ TRUE, IOStatus.OutOfRange};
  END; (* IF rRec.exp10*) 
  IF res.failed
    THEN
      real := zero
    END;

END StrToReal;


PROCEDURE RealToFloat (real: RealType; decPlaces: CARDINAL; numIntDigits : CARDINAL; VAR str: ARRAY OF CHAR);
  (* Converts the value of real to a floating-point string form, with decPlaces after the
  decimal, and numIntDigits before the decimal, with the exponent adjusted accordingly
  and copies the possibly truncated result to str.  *)

VAR
  realRec : RealRec;

BEGIN
  realToRealRec (real, realRec);
  RealRecToFloat (realRec, decPlaces, numIntDigits, TRUE, str)
END RealToFloat;


PROCEDURE RealRecToFloat (rRec : RealRec; decPlaces: CARDINAL; numIntDigits : CARDINAL; useExp : BOOLEAN; VAR str: ARRAY OF CHAR);
(* does most of the work for RealToFloat, but splitting this out allows the RealRec to be adjusted first by RealToEng, for instance *)

VAR
  strSize, roomNeeded, strCount, loopCount, digit, digitStrCount, digitsAvailable, expStrCount, decCount : CARDINAL;
  numSigFigsB4Round, lastExpStrChar : CARDINAL;
  expAdjust : INTEGER;
  digitStr, expStr : strType;
  needExp : BOOLEAN;

BEGIN
  strSize  := CAPACITYch (str) - 1; (* need room for NUL at end *)
  strCount := 0;
  digitStrCount := 0;
  expStrCount := 0;
  expAdjust := 0;
  numSigFigsB4Round := CardinalMath.nearestLargerPowerOf10 (rRec.sigFigs);
  rRec.sigFigs := CardinalMath.round (rRec.sigFigs, CardinalMath.min2( rMaxSigFigs, decPlaces + numIntDigits ) );
  WholeStrs.CardToDecStr (rRec.sigFigs, digitStr );
  roomNeeded := decPlaces + 2; (* for the decimal and one digit before it *)
    digitsAvailable := LENGTH (digitStr);
   (* adjust the exponent if needed if rounding added a sig fig *)
    rRec.exp10 := rRec.exp10 + VAL ( INTEGER, CardinalMath.nearestLargerPowerOf10 (rRec.sigFigs) - numSigFigsB4Round);
    (* adjust the exponent if needed to have more digits than one before decimal, eg for eng *)
  expAdjust := VAL (INTEGER, numIntDigits) - 1;
  IF expAdjust # 0 THEN
    rRec.exp10 := rRec.exp10 - expAdjust
  END;

  needExp := useExp;
  IF numIntDigits > 1 THEN roomNeeded := roomNeeded + numIntDigits - 1 END; (* increase enough for digits *)

  IF needExp THEN
    roomNeeded := roomNeeded + 4
  END; (* room for exponent *)
  IF roomNeeded > strSize THEN
    str [strCount] := ISO646.NUL;
    RETURN
  END;
  IF rRec.neg THEN
      str [strCount] := "-";
      strCount := strCount + 1
  END;
  IF (numIntDigits = 0) OR (rRec.sigFigs = 0) THEN
    str [strCount] := "0";
    strCount := strCount + 1;
  ELSE
    loopCount := 0;
    WHILE loopCount < numIntDigits  DO
      IF digitStrCount < digitsAvailable THEN
        str [strCount] := digitStr [digitStrCount];
        digitStrCount := digitStrCount + 1;
      ELSE
        str [strCount] := "0"
      END;
      strCount := strCount + 1;
      loopCount := loopCount + 1
    END (* while loopCount *)
  END; (* IF numIntDigits = 0 *)

(* done the part before the decimal so... *)
  str [strCount] := ".";
  strCount := strCount + 1;

(* now the part after the decimnal  *)
  decCount := 0;
  (* we may need padding of 0s after the decimal *)
   IF decPlaces > 0 THEN
       IF NOT useExp THEN
         WHILE (rRec.exp10 < 0) AND (decCount < decPlaces) DO
           str [strCount] := "0";
           strCount := strCount + 1;
           decCount := decCount + 1;
           rRec.exp10 := rRec.exp10 + 1
         END
       END; (* IF NOT useExp *)

  (* beyond that we put in sig figs *)
    loopCount := 0;
    WHILE (decCount < decPlaces) DO
      IF (loopCount < decPlaces) AND (digitStrCount < digitsAvailable) THEN
        str [strCount] := digitStr [digitStrCount];
      ELSE
        str [strCount] := "0"
      END;
      strCount := strCount + 1;
      decCount := decCount + 1;
      loopCount := loopCount + 1;
      digitStrCount := digitStrCount + 1;

    END; (* while *)
  ELSE
    str [strCount] := "0";
    strCount := strCount + 1
  END; (* decPlaces > 0 *)

(* last, we put in the exponent *)
  IF needExp THEN
    WholeStrs.CardToDecStr (ABS(rRec.exp10), expStr );
    str [strCount] := expChar;
    strCount := strCount + 1;
    IF rRec.exp10 >= 0 THEN
      str [strCount] := "+";
      strCount := strCount + 1;
    ELSIF rRec.exp10 < 0 THEN
      str [strCount] := "-";
      strCount := strCount + 1
    END;
    IF (expStrLen = 5) AND (ABS(rRec.exp10)  < 100 ) THEN
      str [strCount] := "0";
      strCount := strCount + 1
    END;
    IF ABS (rRec.exp10) < 10 THEN
      str [strCount] := "0";
      strCount := strCount + 1
    END;
    lastExpStrChar := LENGTH (expStr) - 1;
    loopCount := 0;
    WHILE loopCount <= lastExpStrChar DO
      str [strCount] := expStr [expStrCount];
      strCount := strCount + 1;
      expStrCount := expStrCount + 1;
      loopCount := loopCount + 1
    END; (* while loopCount *)
  END; (* IF needExp *)

(* terminate string *)
  str [strCount] := ISO646.NUL;
END RealRecToFloat;


PROCEDURE RealToScientific (real: RealType; decPlaces: CARDINAL; VAR str: ARRAY OF CHAR);
  (* Converts the value of real to floating-point string form, with decPlaces figures after the decimal
  that is decPlaces + 1 significant figures, and copies the possibly truncated result to str.  *)

BEGIN
  RealToFloat (real, decPlaces, 1, str)
END RealToScientific;


PROCEDURE RealToEng (real: RealType; decPlaces: CARDINAL; VAR str: ARRAY OF CHAR);
  (* Converts the value of real to a floating-point string form, with decPlaces
  after the decimal, with the exponent adjusted accordingly  and copies the possibly
   truncated result to str. The number is scaled with one to three digits
   in the whole number part and with an exponent that is a multiple of three.  *)

VAR
  rRec : RealRec;
  adjustIntPlaces, intPlaces : INTEGER;

BEGIN
  realToRealRec (real, rRec);
  adjustIntPlaces := (ABS (rRec.exp10) ) MOD 3;
  IF rRec.exp10 >= 0 THEN
    intPlaces := 1 + adjustIntPlaces
  ELSE
    intPlaces := 4 - adjustIntPlaces
  END;
  RealRecToFloat (rRec, decPlaces, intPlaces, TRUE, str);
END RealToEng;


PROCEDURE realRecToFixed (rRec : RealRec; place: INTEGER; VAR str: ARRAY OF CHAR);
  (* Does most of the work of conversion, but the split allows for adjustments first *)
VAR
  decPlaces, sigFigs, numIntDigits : CARDINAL;

BEGIN
  IF rRec.exp10 >= 0 THEN
    numIntDigits := rRec.exp10 + 1;
  ELSE
    numIntDigits := 0
  END;

  IF place >= 0 THEN (* rounding off before decimal *)
    decPlaces := 0; (* so no sig figs after *)
    rRec.sigFigs := CardinalMath.round (rRec.sigFigs, CardinalMath.min2( rMaxSigFigs,rRec.exp10 - place + 1) );
  ELSE (* rounding off after decimal *)
      decPlaces := ABS (place);
    rRec.sigFigs := CardinalMath.round (rRec.sigFigs, rMaxSigFigs )
  END;
  RealRecToFloat (rRec, decPlaces, numIntDigits, FALSE, str);
END realRecToFixed;


PROCEDURE RealToFixed (real: RealType; place: INTEGER; VAR str: ARRAY OF CHAR);
  (* Converts the value of real to fixed-point string form, rounded to the given place relative to the decimal point, and copies the possibly truncated result to str. *)
VAR
  rRec : RealRec;

BEGIN
  realToRealRec (real, rRec);
  realRecToFixed (rRec, place, str )
END RealToFixed;


PROCEDURE RealToStr (real: RealType; VAR str: ARRAY OF CHAR);
  (* Converts the value of real as RealToFixed if the sign and magnitude can be shown within the capacity of str, or otherwise as RealToScientific, and copies the possibly truncated result to str. The number of places or significant digits are implementation-defined.
  *)
VAR
  limit, minSpaceNeeded, spaceForFixed, decPlaces : CARDINAL;
  rRec : RealRec;
  place : INTEGER;

BEGIN
  limit  := CAPACITYch (str) - 1;
  realToRealRec (real, rRec);
  IF rRec.exp10 > rMaxSigFigs - 1 THEN
    minSpaceNeeded := VAL (CARDINAL, rRec.exp10 + 1 ) + 1
  ELSE
    minSpaceNeeded := rMaxSigFigs + 1; (* the latter for a decimal *)
    END;
  spaceForFixed := minSpaceNeeded;
  IF real < 0.0 THEN
    spaceForFixed := spaceForFixed + 1; (* allow room for a sign *)
  END;

  IF (rRec.exp10 = VAL (INTEGER, (rMaxSigFigs - 1) ) ) THEN
    spaceForFixed := spaceForFixed + 1  (* one more for an extra 0 after decimal *)
  END;
  IF rRec.exp10 < 0 THEN spaceForFixed := spaceForFixed + VAL (CARDINAL, -rRec.exp10 + 1) END; (* space for zeros after decimal *)

  IF spaceForFixed > limit THEN
      decPlaces := limit - 2 - expStrLen; (* four for exponent one for first sig fig and one for a decimal *)
      IF real < 0.0 THEN
        decPlaces := decPlaces - 1
      END;
        RealRecToFloat (rRec, decPlaces, 1, TRUE, str); (* same call as RealToScientific *)
  ELSE
    place := rRec.exp10 - VAL (INTEGER, rMaxSigFigs) + 1
  END (* IF spaceForFixed *);
  realRecToFixed (rRec, place, str)
END RealToStr;

BEGIN (* module *)
  (* use one less sig fig than max to prevent overflow *)
  sigFigHighPowerOf10 := CardinalMath.exp10 (CardinalMath.maxPowerOf10() - 2);
  realToRealRec (MAX (RealType), maxRealRec )
END RealStrs.
