IMPLEMENTATION MODULE  Transition;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      04/12/2014  First implementation

  *)

IMPORT SYSTEM, UNSAFE;
IMPORT COMPLEX; FROM COMPLEX IMPORT COMPLEXT;
IMPORT LONGCOMPLEX; FROM LONGCOMPLEX IMPORT LONGCOMPLEXT;
IMPORT CARD8; FROM CARD8 IMPORT CARD8T;
IMPORT CardinalMath, LongCardinalMath, Bitwidth, RealMath;

CONST
  maxInt = 9223372036854775807.0;

PROCEDURE Cast (input : ARRAY OF SYSTEM.WORD; VAR output : ARRAY OF SYSTEM.WORD);
VAR
  count, inputCount, outputCount : CARDINAL;
BEGIN
  outputCount := CAPACITYw ( output );
  inputCount := CAPACITYw ( input );
  count := 0;
  WHILE ( count < inputCount ) AND ( count < outputCount ) DO
    output [ count ] := input [ count ];
    count := count + 1
  END
END Cast;

PROCEDURE CastOcts (input : ARRAY OF OCTET; VAR output : ARRAY OF OCTET);

VAR
  Count, InputCount, OutputCount : CARDINAL;

BEGIN
  OutputCount := CAPACITYo ( output );
  InputCount := CAPACITYo ( input );
  Count := 0;
  WHILE ( Count < InputCount ) AND ( Count < InputCount ) DO
    output [ Count ] := input [ Count ];
    Count := Count + 1
  END
END CastOcts;


PROCEDURE IncAdr ( VAR adr : SYSTEM.ADDRESS );
VAR
  card : CARDINAL;
BEGIN
  Cast ( adr, card);
  card := card + 1;
  Cast ( card, adr )
END IncAdr;


PROCEDURE CAPACITYo (ary : ARRAY OF OCTET ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYo;


PROCEDURE CAPACITYao (ary : ARRAY OF ARRAY OF OCTET ) : CARDINAL;
    (* returns the count on the array, i.e how many arrays it contains *)

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYao;


PROCEDURE CAPACITYad (ary : ARRAY OF UNSAFE.ADDRESS ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYad;


PROCEDURE CAPACITYb (ary : ARRAY OF BITSET ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYb;


PROCEDURE CAPACITYw (ary : ARRAY OF UNSAFE.WORD ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYw;


PROCEDURE CAPACITYc8 (ary : ARRAY OF CARD8T ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYc8;


PROCEDURE CAPACITYc (ary : ARRAY OF CARDINAL ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYc;


PROCEDURE CAPACITYlc (ary : ARRAY OF LONGCARD ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYlc;


PROCEDURE CAPACITYi (ary : ARRAY OF INTEGER ) : CARDINAL;
BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYi;


PROCEDURE CAPACITYli (ary : ARRAY OF LONGINT ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYli;


PROCEDURE CAPACITYr (ary : ARRAY OF REAL ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYr;


PROCEDURE CAPACITYlr (ary : ARRAY OF LONGREAL ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYlr;


PROCEDURE CAPACITYz (ary : ARRAY OF COMPLEXT ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYz;


PROCEDURE CAPACITYlz (ary : ARRAY OF LONGCOMPLEXT ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYlz;


PROCEDURE CAPACITYn (ary : ARRAY OF OCTET; size : CARDINAL ) : CARDINAL;

BEGIN
  RETURN (HIGH (ary) + 1) DIV size
END CAPACITYn;


PROCEDURE CAPACITYch (ary : ARRAY OF CHAR ) : CARDINAL;

BEGIN
  RETURN HIGH (ary) + 1
END CAPACITYch;

PROCEDURE FLOATc (card : CARDINAL) : REAL;
(* find the floating point equivalent to the cardinal in single precision real *)

CONST
  Bias = 127;
  BitsInCharacteristic = 8;
  HiBitInMantissa = 22;
  signBit = 31;
VAR
  exp, rCount, cCount, eCount, bitsKept, bitsLost : CARDINAL;
  bitsInCard, binShift : CARDINAL;
  target : REAL;
  
BEGIN
  bitsInCard := Bitwidth.CardBitwidth;
  (* ensure the target is zeroed out *)
  target := 0.0;
  (* find out how far to shift the cardinal digits to put them into
   positions HiBitInMantissa to 0 in the real *)
  binShift := UNSAFE.CLZ (card) + 1;
  IF binShift = 65 THEN (* the number is zero, so *)
    RETURN target
  END; (* if *)
  
  (* round up if needed *)
  bitsKept := HiBitInMantissa + binShift;
  bitsLost := 0;
  IF bitsKept < bitsInCard THEN
    bitsLost := bitsInCard - bitsKept;
  END; (*if *)
  IF (bitsLost > 0) AND (UNSAFE.BIT (card, bitsLost - 1) ) THEN
    card := card + CardinalMath.exp2 (bitsLost); 
    binShift := UNSAFE.CLZ (card) + 1; (* may have changed *)
  END; (* if *)

  (* compute the exponent and bias it *)
  exp := Bias + bitsInCard - binShift;
    
  (* Set up for the shift.  Don't bother moving the leading 1
  as the real format assumes it *)
  cCount := bitsInCard - binShift;
  rCount := HiBitInMantissa + 1; 
  WHILE rCount > 0 DO
    rCount := rCount - 1;
    cCount := cCount - 1;
    IF UNSAFE.BIT (card, cCount) THEN
      UNSAFE.SETBIT (target, rCount, TRUE);
    ELSE
      UNSAFE.SETBIT (target, rCount, FALSE);
    END; (* if *)
  END; (* While *)
(* now put the exponent into positions HiBitInMantissa+1 to HiBitInMantissa+8 *)
  rCount := HiBitInMantissa + 1;
  eCount := 0;
  WHILE eCount < BitsInCharacteristic DO
    IF UNSAFE.BIT (exp, eCount) THEN
      UNSAFE.SETBIT (target, rCount, TRUE);
    END; (* if *)
    rCount := rCount + 1;
    eCount := eCount + 1;
  END; (* While *)
  UNSAFE.SETBIT (target, signBit, FALSE);
  RETURN target; 
END FLOATc;


PROCEDURE FLOATi (int : INTEGER) : REAL;
VAR
  card : CARDINAL;
  target : REAL;
BEGIN
  card := ABS (int);
  target := FLOATc (card);
  IF int < 0 THEN target := -target END;
  RETURN target;
END FLOATi;

PROCEDURE TRUNC (r : REAL) : CARDINAL;

CONST
  Bias = 127;
  BitsInCharacteristic = 8;
  HiBitInMantissa = 22;
  signBit = 31;
  max = MAX (CARDINAL);
  maxSigFigs = 9;
VAR
  exp, rCount, cCount, eCount, bitsKept, bitsLost : CARDINAL;
  bitsInCard, binShift : CARDINAL;
  target : CARDINAL;

BEGIN
  IF r < 0.0 THEN RETURN max + 1 END;
(* get the exponent from positions  HiBitInMantissa+1 to HiBitInMantissa+8 *)
  rCount := HiBitInMantissa + 1;
  eCount := 0;
  exp := 0; (* set all bits to zero *)
  WHILE eCount < BitsInCharacteristic DO
    IF UNSAFE.BIT (r, rCount) THEN
      UNSAFE.SETBIT (exp, eCount, TRUE);
    END; (* if *)
    rCount := rCount + 1;
    eCount := eCount + 1;
  END; (* While *)
  IF exp >= Bias THEN 
    exp := exp - Bias;
  ELSE (* true exp is negative so *)
    RETURN 0
  END; (* if *)
  IF exp > 64 THEN
     RETURN max + 1 (* force an overflow *)
  END; (* if exp *)
  target := 0;
  (* now, starting at position exp copy set bits from the real into the cardinal          *)
  rCount := HiBitInMantissa;
  cCount := exp;
  UNSAFE.SETBIT (target, cCount, TRUE);
  WHILE cCount > 0 DO
    cCount := cCount - 1;
    IF UNSAFE.BIT (r, rCount) THEN
      UNSAFE.SETBIT (target, cCount, TRUE);
    END; (* if *)
    rCount := rCount -1
  END; (* while *)
(* Now round off to 9 sig figs as that is all we can get from converting a real *)
  
  RETURN CardinalMath.round ( target, maxSigFigs );
END TRUNC;


PROCEDURE INT (r : REAL) : INTEGER;

VAR
  card : CARDINAL;
  target : INTEGER;
  temp : REAL;
  neg : BOOLEAN;
  max : CARDINAL;

BEGIN
  max := MAX (CARDINAL) DIV 2;
  neg := (r < 0.0 );
  IF neg THEN  r := r - 0.5 END; (* to ensure neg is rounded away from 0 *)
  temp := ABS (r);
  card := TRUNC (temp);
  Cast (card, target);
  IF neg THEN
    RETURN -target;    
  ELSE 
    RETURN target
  END (* if *)
END INT;

PROCEDURE LFLOATc (card : LONGCARD) : LONGREAL;
(* find the floating point equivalent to the cardinal in single precision real *)

CONST
  Bias = 1023;
  BitsInCharacteristic = 11;
  HiBitInMantissa = 51;
  signBit = 63;

VAR
  exp, rCount, cCount, eCount, bitsKept, bitsLost : CARDINAL;
  bitsInCard, binShift : CARDINAL;
  target : LONGREAL;
  
BEGIN
  bitsInCard := Bitwidth.LongCardBitwidth;
  (* ensure the target is zeroed out *)
  target := 0.0;
  (* find out how far to shift the cardinal digits to put them into
   positions HiBitInMantissa to 0 in the real *)
  binShift := UNSAFE.CLZ (card) + 1;
  IF binShift = 65 THEN (* the number is zero, so *)
    RETURN target
  END; (* if *)
  
  (* round up if needed *)
  bitsKept := HiBitInMantissa + binShift;
  bitsLost := 0;
  IF bitsKept < bitsInCard THEN
    bitsLost := bitsInCard - bitsKept;
  END; (*if *)
  IF (bitsLost > 0) AND (UNSAFE.BIT (card, bitsLost - 1) ) THEN
    card := card + LongCardinalMath.exp2 (bitsLost); 
    binShift := UNSAFE.CLZ (card) + 1; (* may have changed *)
  END; (* if *)

  (* compute the exponent and bias it *)
  exp := Bias + bitsInCard - binShift;
    
  (* Set up for the shift.  Don't bother moving the leading 1
  as the real format assumes it *)
  cCount := bitsInCard - binShift;
  rCount := HiBitInMantissa + 1; 
  WHILE rCount > 0 DO
    rCount := rCount - 1;
    cCount := cCount - 1;
    IF UNSAFE.BIT (card, cCount) THEN
      UNSAFE.SETBIT (target, rCount, TRUE);
    ELSE
      UNSAFE.SETBIT (target, rCount, FALSE);
    END; (* if *)
  END; (* While *)
(* now put the exponent into positions HiBitInMantissa+1 to HiBitInMantissa+8 *)
  rCount := HiBitInMantissa + 1;
  eCount := 0;
  WHILE eCount < BitsInCharacteristic DO
    IF UNSAFE.BIT (exp, eCount) THEN
      UNSAFE.SETBIT (target, rCount, TRUE);
    END; (* if *)
    rCount := rCount + 1;
    eCount := eCount + 1;
  END; (* While *)
  UNSAFE.SETBIT (target, signBit, FALSE);
  RETURN target; 
END LFLOATc;

PROCEDURE LFLOATi (int : LONGINT) : LONGREAL;
VAR
  card : LONGCARD;
  target : LONGREAL;
BEGIN
  card := ABS (int);
  target := LFLOATc (card);
  IF int < 0 THEN target := -target END;
  RETURN target;
END LFLOATi;

PROCEDURE LTRUNC (r : LONGREAL) : LONGCARD;

CONST
  Bias = 1023;
  BitsInCharacteristic = 11;
  HiBitInMantissa = 51;
  signBit = 63;
  max = MAX (LONGCARD);
  maxSigFigs = 17;
VAR
  exp, rCount, cCount, eCount, bitsKept, bitsLost : CARDINAL;
  bitsInCard, binShift : CARDINAL;
  target : LONGCARD;

BEGIN
  IF r < 0.0 THEN RETURN max + 1 END;
(* get the exponent from positions  HiBitInMantissa+1 to HiBitInMantissa+8 *)
  rCount := HiBitInMantissa + 1;
  eCount := 0;
  exp := 0; (* set all bits to zero *)
  WHILE eCount < BitsInCharacteristic DO
    IF UNSAFE.BIT (r, rCount) THEN
      UNSAFE.SETBIT (exp, eCount, TRUE);
    END; (* if *)
    rCount := rCount + 1;
    eCount := eCount + 1;
  END; (* While *)
  IF exp >= Bias THEN 
    exp := exp - Bias;
  ELSE (* true exp is negative so *)
    RETURN 0
  END; (* if *)
  IF exp > 64 THEN
     RETURN max + 1 (* force an overflow *)
  END; (* if *)
  target := 0;
  (* now, starting at position exp copy set bits from the real into the cardinal          *)
  rCount := HiBitInMantissa;
  cCount := exp;
  UNSAFE.SETBIT (target, cCount, TRUE);
  WHILE cCount > 0 DO
    cCount := cCount - 1;
    IF UNSAFE.BIT (r, rCount) THEN
      UNSAFE.SETBIT (target, cCount, TRUE);
    END; (* if *)
    rCount := rCount -1
  END; (* while *)
(* Now round off to 9 sig figs as that is all we can get from converting a real *)
  
  RETURN CardinalMath.round ( target, maxSigFigs );
END LTRUNC;

PROCEDURE LINT (r : LONGREAL) : LONGINT;

VAR
  card : LONGCARD;
  target : LONGINT;
  temp : LONGREAL;
  neg : BOOLEAN;
  max : LONGCARD;

BEGIN
  max := MAX (LONGCARD) DIV 2;
  neg := (r < 0.0 );
  IF neg THEN  r := r - 0.5 END; (* to ensure neg is rounded away from 0 *)
  temp := ABS (r);
  card := LTRUNC (temp);
  Cast (card, target);
  IF neg THEN
    RETURN -target;    
  ELSE 
    RETURN target
  END (* if *)

END LINT;


  (***********************************)
  (*#####   Module Management   #####*)
  (***********************************)

  PROCEDURE InitTransition;
  BEGIN (*InitTransition*)
  END InitTransition;

BEGIN (* Transition *)
  InitTransition
END Transition.
