(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE BooleanBaseIO (*[ProtoIO] FOR CARDINAL*);

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as BooleanIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite


  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYo;
FROM Transition IMPORT CHARSET;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* RS utilities *)
IMPORT TokenIO;
IMPORT CharArray;

IMPORT FormatIO;
IMPORT OctetStrs, WholeStrs;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ISO646, Perror;
IMPORT ChanIO;


CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  BoolType = BOOLEAN;

CONST
  trueStr = "TRUE";
  falseStr = "FALSE";
  yesStr = "YES";
  noStr = "NO";
  trueChars = CHARSET{"t","T","y","Y"};
  falseChars = CHARSET{"f","F","n","N"};


PROCEDURE BoolToStr (item : BOOLEAN; yesFormat : BOOLEAN ) : FormatIO.Str;

BEGIN
  IF yesFormat THEN
    IF item THEN
      RETURN yesStr
    ELSE
      RETURN noStr
    END
   ELSE (* use true format *)
     IF item THEN
       RETURN trueStr
     ELSE
       RETURN falseStr
     END
  END
END BoolToStr;


PROCEDURE SetResult (file: ChanIO.Channel; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code OF
    IOStatus.Success    : ChanIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
    IOStatus.OutOfRange  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
    IOStatus.WrongFormat  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
    IOStatus.StrEmpty   : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
  END; (* case *)
  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
  END
END SetResult;

PROCEDURE Read ( infile : ChanIO.Channel; VAR b : BoolType );

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;

BEGIN (* Read *)
  TokenIO.ReadTextToken (infile, temp);        (* get the usigned as a string *)
  IF (CharArray.matches (temp, trueStr) )  OR (CharArray.matches (temp, yesStr) )
      OR ( ISO646.toUpper(temp[0]) IN trueChars) THEN
    b := TRUE;
    SetResult (infile, IOStatusT{ FALSE, IOStatus.Success});
  ELSIF (CharArray.matches (temp, falseStr) )  OR (CharArray.matches (temp, noStr) )
     OR ( ISO646.toUpper(temp[0]) IN falseChars) THEN
    b := FALSE;
    SetResult (infile, IOStatusT{ FALSE, IOStatus.Success});
  ELSE
    SetResult ( infile, IOStatusT{ TRUE, IOStatus.WrongFormat} );
  END
END Read;



PROCEDURE (*[WRITE]*) Write
  ( outfile : ChanIO.Channel; b : BoolType );

VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  CharStrIO.WriteStr (outfile, BoolToStr ( b, FALSE));
END Write;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF BoolType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;

BEGIN
  numItems := CAPACITYo (items) -1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO

      (* convert to the appropriate string *)
    temp := BoolToStr (items[count], fsr.yes4Bool);

     (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

     (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST >0*) ARRAY OF BoolType );

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.BooleanScanSet, FormatIO.ValidBooleanFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;


(* Kludge Section *)

TYPE
  (*TargetType = ;*) (* in definition *)
  Ary1 = ARRAY [0..0] OF TargetType;
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;


PROCEDURE Write1F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary1{item} )
END Write1F;


PROCEDURE Write2F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item1, item2 : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary2{item1, item2} )
END Write2F;


PROCEDURE Write3F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item1, item2, item3 : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary3{item1, item2, item3} )
END Write3F;


END BooleanBaseIO.
