(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE ByteArrayIO (*[ProtoIO] FOR BYTE*);

(* IO Extension Library for Values of Type BYTE *)

 (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 12  uses new FormatIO module
    RS    2015 07 13  Separated completely from Octet I/O
    RS       2016 07 12 Re-done for new library version


  *)

IMPORT SYSTEM, UNSAFE;
FROM CARDINAL8 IMPORT CARDINAL8;
FROM Transition IMPORT CAPACITYo, CAPACITYao;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* RS utilities *)
IMPORT TokenIO, FormatIO;
IMPORT Card8Strs;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ByteBaseIO, Perror, ISO646, ARRAYOFCHAR, CharArray;
IMPORT ChanIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARDINAL8;

CONST
  StrToWholeType = Card8Strs.StrToCard8;
  WholeTypeToDecStr = Card8Strs.Card8ToDecStr;
  WholeTypeToHexStr = Card8Strs.Card8ToHexStr;
  WholeTypeToBinStr = Card8Strs.Card8ToBinStr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : ChanIO.Channel; VAR array : ARRAY OF UNSAFE.BYTE );

VAR
  count : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  WHILE count < CAPACITYo (array) DO
    ChanIO.ReadOctet (infile, array [count] );
    status := ChanIO.statusOf (infile);
    count := count + 1;
    IF NOT status.failed THEN RETURN END
  END

END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : ChanIO.Channel; (*CONST*) array : ARRAY OF UNSAFE.BYTE );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYo (array) - 1;
  (* open brace for the array *)
 CharStrIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <=  last DO
    ByteBaseIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF UNSAFE.BYTE );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  numItems := CAPACITYo (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := SYSTEM.CAST (WholeType, items[count]);
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
         (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

         (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharBaseIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF UNSAFE.BYTE );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   numItems := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.ByteScanSet, FormatIO.ValidByteFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <=  numItems DO
    (* go write the array *)   
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END (* if *);
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrIO.WriteStr (outfile, "\n")  END   
    END; (* if *)
    count := count + 1
  END; (* while *)
  count := count + 1;
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)
END WriteFArray;

END ByteArrayIO.