(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE CharBaseIO (*[ProtoIO] FOR CHAR*);

(* File-Desc IO Extension Library for Values of Type CHAR *)

IMPORT SYSTEM;

(* R10 stuff *)
IMPORT ChanIO, CharStrIO, ISO646;

(* RS Stuff *)
IMPORT FormatIO;
IMPORT LongWholeStrs;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;


PROCEDURE (*[READ]*) Read
  ( infile : ChanIO.Channel; VAR ch : CHAR );

BEGIN
  ChanIO.ReadOctet ( infile, ch )
END Read;


PROCEDURE (*[WRITE]*) Write
  ( outfile : ChanIO.Channel; ch : CHAR );

BEGIN
  ChanIO.WriteOctet ( outfile, ch )
END Write;

PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items : ARRAY OF CHAR );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  lCard : LONGCARD;

BEGIN
   numItems := LENGTH ( items )-1;
   colCount := 0;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    IF fsr.numeralFormat = FormatIO.Hex THEN
      lCard := ORD ( items[count] );
      LongWholeStrs. CardToHexStr ( lCard, temp);
    ELSE
      temp[0] := items[count]; temp[1] := ISO646.NUL;
    END; (* if *)
         (* format that char using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

        (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF NOT fsr.isConcat AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrIO.WriteStr (outfile, "\n") END
    END (* if *);
    count := count + 1
  END (* while *);

  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)
END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST >0*) ARRAY OF CHAR );

VAR
  fsr : FormatScanRec;

BEGIN
   (* get the format string scanned to a format string record *)
    ScanStrForFormat (fmtStr, FormatIO.CharScanSet, FormatIO.ValidCharFmtChars, fsr);
    WriteFWithfsr ( outfile, fsr , items  )

END WriteF;


(* Kludge Section *)

TYPE
  (*TargetType = ;*) (* in definition *)
  Ary1 = ARRAY [0..0] OF TargetType;
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;


PROCEDURE Write1F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary1{item} )
END Write1F;


PROCEDURE Write2F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item1, item2 : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary2{item1, item2} )
END Write2F;


PROCEDURE Write3F (outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; item1, item2, item3 : TargetType );

BEGIN
  WriteF (outfile, fmtStr, Ary3{item1, item2, item3} )
END Write3F;

END CharBaseIO.