IMPLEMENTATION MODULE  ComplexBaseIO;

 (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 15  uses new FormatIO module
    RS    2015 05 28  Stomp bugs and refined
    RS      2016 07 26  Now IO

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYz;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenIO, FormatIO;
IMPORT OctetStrs, RealStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ISO646;
IMPORT ChanIO;
IMPORT COMPLEX; FROM COMPLEX IMPORT COMPLEXT;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
    RealType = REAL;
    ComplexType = COMPLEXT;

CONST
  StrToRealType = RealStrs.StrToReal;
  RealTypeToStr = RealStrs.RealToStr;
  RealTypeToScientific = RealStrs.RealToScientific;
  RealTypeToFloat = RealStrs.RealToFloat;
  RealTypeToEng = RealStrs.RealToEng;
  RealTypeToFixed = RealStrs.RealToFixed;
  plusStr = " + "; (* to go between the real components of a complex *)
  minusStr = " - "; (* to go between the real components of a complex *)
  imStr ="i"; (* suffix to write for im component *)


PROCEDURE SetResult (file: ChanIO.Channel; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code OF
    IOStatus.Success    : ChanIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
    IOStatus.OutOfRange  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
    IOStatus.WrongFormat  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
    IOStatus.StrEmpty   : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
  END; (* case *)

  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
      END
END SetResult;



PROCEDURE (*. [READ] .*) Read( infile : ChanIO.Channel; VAR z : COMPLEXT );VAR
  temp  : FormatIO.Str;
  res : IOStatusT;
  ch : CHAR;
  re, im : RealType;
  isRawFormat : BOOLEAN;


BEGIN (* Read *)
  z := COMPLEX.zero;
  TokenIO.LookCharToken (infile, ch);
  isRawFormat := (ch = "(" );
  IF isRawFormat THEN
      CharBaseIO.Read (infile,ch)
  END;
  TokenIO.ReadRealToken (infile, temp);  (* get the first RealType as a string *)
  StrToRealType (temp, re, res);       (* convert the string to a RealType *)
  SetResult (infile, res);
  IF res.failed THEN RETURN END;
  IF isRawFormat THEN
      CharBaseIO.Read (infile,ch);
    IF ch # "," THEN
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      SetResult (infile, res);
      RETURN
     END
  END;
  TokenIO.ReadRealToken (infile, temp);  (* get the second RealType as a string *)
  StrToRealType (temp, im, res);       (* convert the string to a RealType *)
  SetResult (infile, res);
  IF res.failed THEN RETURN END;

  TokenIO.ReadCharToken (infile, ch);
  IF (isRawFormat AND ( ch # ")" ) )
    OR (NOT isRawFormat AND (ch # "i") AND (ch # "j")  )(* engineers like the latter *)
  THEN
    res := IOStatusT{ TRUE, IOStatus.WrongFormat};
    SetResult (infile, res);
    RETURN
  END;
  z := COMPLEX.CMPLX (re, im)
END Read;



PROCEDURE (*. [WRITE] .*) Write( outfile : ChanIO.Channel; (*. CONST .*) z : COMPLEXT );
(* Writes the textual representation of value z to output stream outfile. This
   procedure is substituted for invocations of WRITE with a COMPLEX argument.*)
VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  RealTypeToScientific (z.re, RealStrs.rMaxSigFigs, temp);            (* convert the RealType to a scientific string (default) *)
  CharStrIO.WriteStr (outfile, temp);
  IF z.im < 0.0 THEN
    CharStrIO.WriteStr (outfile, minusStr);
  ELSE
    CharStrIO.WriteStr (outfile, plusStr)
  END;
  RealTypeToScientific (ABS(z.im), RealStrs.rMaxSigFigs, temp);            (* convert the RealType to a scientific string (default) *)
  CharStrIO.WriteStr (outfile, temp);
  CharStrIO.WriteStr (outfile, imStr);

END Write;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF ComplexType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  theNum : ComplexType;


PROCEDURE NumberToNumeralString (theNum: RealType); (*internal *)

BEGIN       (* convert to the appropriate string *)
  IF fsr.numeralFormat = FormatIO.Hex
  THEN
    OctetStrs.OctsToHexStr (theNum, temp);
  ELSIF fsr.numeralFormat = FormatIO.Bin THEN
    OctetStrs.OctsToBinStr (theNum, temp);
  ELSIF fsr.realFormat = FormatIO.Scientific THEN
    RealTypeToScientific (theNum, fsr.numDecPlaces, temp);
  ELSIF fsr.realFormat = FormatIO.Float THEN
    RealTypeToFloat (theNum, fsr.numDecPlaces, fsr.numIntDigitsForFloat, temp);
  ELSIF fsr.realFormat = FormatIO.Fixed THEN
    RealTypeToFixed (theNum, fsr.numDecPlaces, temp);
  ELSIF fsr.realFormat = FormatIO.Eng THEN
    RealTypeToEng (theNum, fsr.numDecPlaces, temp);
  END(*IF*)
END NumberToNumeralString;

BEGIN
  IF fsr.numDecPlaces = 0 THEN fsr.numDecPlaces  := RealStrs.rMaxSigFigs END; (* default *)
  last := CAPACITYz (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    (* open paren if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN CharStrIO.WriteStr (outfile, "( ") END;
    theNum := items[count];

     (* do real part *)
    NumberToNumeralString (theNum.re);
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);
    (* comma separator if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN
      CharStrIO.WriteStr (outfile, ", ")
    ELSE
      IF theNum.im < 0.0 THEN
        CharStrIO.WriteStr (outfile, minusStr);
      ELSE
        CharStrIO.WriteStr (outfile, plusStr)
      END;
    END;

     (* repeat for IM Part *)
    NumberToNumeralString (ABS(theNum.im));
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);
    CASE fsr.complexFormat OF
     FormatIO.Raw :
       CharStrIO.WriteStr (outfile, " )")
     | FormatIO.i :
       CharStrIO.WriteStr (outfile, "i")
     | FormatIO.j :
       CharStrIO.WriteStr (outfile, "j")
    END; (* case *)

        (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*. [WRITEF] .*) WriteF
  ( outfile      : ChanIO.Channel;
    (*. CONST .*) fmtStr : ARRAY OF CHAR;
    items        : (*. ARGLIST .*) ARRAY OF COMPLEXT );
(* Writes a formatted textual representation of one or more COMPLEX values to
   output stream outfile. The output format is determined by fmtStr. This
   procedure is substituted for invocations of WRITEF with one or more
   COMPLEX arguments. *)
VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.ComplexScanSet, FormatIO.ValidComplexFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END ComplexBaseIO.
