IMPLEMENTATION MODULE  LongComplexArrayIO;

 (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 15  uses new FormatIO module
    RS    2015 05 28  Stomp bugs and refined
    RS      2016 07 26  Now IO
    RS      2017 06 21  Now PIM

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYlz, CAPACITYao;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenIO, FormatIO;
IMPORT OctetStrs, LongRealStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ISO646;

(* R10 Stuff with mods *)
IMPORT ChanIO, LongComplexBaseIO;
IMPORT LONGCOMPLEX; FROM LONGCOMPLEX IMPORT LONGCOMPLEXT;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  RealType = LONGREAL;
  ComplexType = LONGCOMPLEXT;

CONST
  StrToRealType = LongRealStrs.StrToReal;
  RealTypeToStr = LongRealStrs.RealToStr;
  RealTypeToScientific = LongRealStrs.RealToScientific;
  RealTypeToFloat = LongRealStrs.RealToFloat;
  RealTypeToEng = LongRealStrs.RealToEng;
  RealTypeToFixed = LongRealStrs.RealToFixed;
  plusStr = " + "; (* to go between the real components of a complex *)
  minusStr = " - "; (* to go between the real components of a complex *)
  imStr ="i"; (* suffix to write for im component *)


PROCEDURE SetResult (file: ChanIO.Channel; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code OF
    IOStatus.Success    : ChanIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
    IOStatus.OutOfRange  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
    IOStatus.WrongFormat  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
    IOStatus.StrEmpty   : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
  END; (* case *)

  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
      END
END SetResult;


PROCEDURE ReadArray ( infile : ChanIO.Channel; VAR value : ARRAY OF LONGCOMPLEXT );

VAR
  count, last : CARDINAL;

BEGIN (* ReadArray *)
  count := 0;
  last := CAPACITYlz (value) - 1;
  WHILE count <= last DO
    LongComplexBaseIO.Read (infile, value [count]);
    count := count + 1
  END
END ReadArray;


PROCEDURE WriteArray ( outfile : ChanIO.Channel; value : ARRAY OF LONGCOMPLEXT );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYlz (value) - 1;
  CharStrIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <=  last DO
    LongComplexBaseIO.Write (outfile, value [count]);
    IF count < last THEN
        CharStrIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  CharStrIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items : ARRAY OF ComplexType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  theNum : ComplexType;


PROCEDURE NumberToNumeralString (theNum: RealType); (*internal *)

BEGIN       (* convert to the appropriate string *)
   IF fsr.numeralFormat = FormatIO.Hex
   THEN
     OctetStrs.OctsToHexStr (theNum, temp);
   ELSIF fsr.numeralFormat = FormatIO.Bin THEN
     OctetStrs.OctsToBinStr (theNum, temp);
   ELSIF fsr.realFormat = FormatIO.Scientific THEN
    RealTypeToScientific (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Float THEN
    RealTypeToFloat (theNum, fsr.numDecPlaces, fsr.numIntDigitsForFloat, temp);
   ELSIF fsr.realFormat = FormatIO.Fixed THEN
    RealTypeToFixed (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Eng THEN
    RealTypeToEng (theNum, fsr.numDecPlaces, temp);
   END(*IF*)
END NumberToNumeralString;

BEGIN
  IF fsr.numDecPlaces = 0 THEN fsr.numDecPlaces  := LongRealStrs.rMaxSigFigs END; (* default *)
  last := CAPACITYlz (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO

  (* open paren if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN CharStrIO.WriteStr (outfile, "( ") END;
    theNum := items[count];

    (* do real part *)
    NumberToNumeralString (theNum.re);
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);
    (* comma separator if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN
      CharStrIO.WriteStr (outfile, ", ")
    ELSE
      IF theNum.im < 0.0 THEN
        CharStrIO.WriteStr (outfile, minusStr);
      ELSE
        CharStrIO.WriteStr (outfile, plusStr)
      END;
    END;

     (* repeat for IM Part *)
    NumberToNumeralString (ABS(theNum.im));
    (* format that converted string using the fsr already parsed *)
   temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);
    CASE fsr.complexFormat OF
      FormatIO.Raw :
        CharStrIO.WriteStr (outfile, " )")
      | FormatIO.i :
        CharStrIO.WriteStr (outfile, "i")
      | FormatIO.j :
        CharStrIO.WriteStr (outfile, "j")
    END; (* case *)

        (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE WriteFArray
  ( outfile : ChanIO.Channel; fmtStr : ARRAY OF CHAR; items : (*. ARGLIST .*) ARRAY OF ARRAY OF LONGCOMPLEXT );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, last, colCount : CARDINAL;

BEGIN (* WriteFArray *)
  last := CAPACITYao (items) - 1;
  colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.ComplexScanSet, FormatIO.ValidComplexFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= last DO

    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , items [count]  );
    IF fsrForArglist.isArrayFormat AND (count # last) THEN
      CharStrIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END;
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  last ) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END
 END;

  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)

END WriteFArray;

END LongComplexArrayIO.
