IMPLEMENTATION MODULE OctetArrayIO (*[ProtoArrayIO] FOR OCTET*);

(* IO Extension Library for Array Types of OCTET *)

(* IO Extension Library for Array Types of CHAR *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYo, CAPACITYao;
FROM Transition IMPORT OCTET;

FROM CARDINAL8 IMPORT CARDINAL8;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenIO, FormatIO;
IMPORT Card8Strs;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, OctetBaseIO, Perror, ISO646, ARRAYOFCHAR, CharArray;
IMPORT ChanIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARDINAL8;

CONST
  StrToWholeType = Card8Strs.StrToCard8;
  WholeTypeToDecStr = Card8Strs.Card8ToDecStr;
  WholeTypeToHexStr = Card8Strs.Card8ToHexStr;
  WholeTypeToBinStr = Card8Strs.Card8ToBinStr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : ChanIO.Channel; VAR array : ARRAY OF OCTET );

VAR
  count : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  WHILE count <= CAPACITYo (array) - 1
    DO
      ChanIO.ReadOctet (infile, array [count] );
      status := ChanIO.statusOf (infile);
      count := count + 1;
      IF NOT status.failed THEN RETURN END
    END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : ChanIO.Channel; (*CONST*) array : ARRAY OF OCTET );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYo (array) - 1;
  (* open brace for the array *)
 CharStrIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <=  last DO
    OctetBaseIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF OCTET );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  last := CAPACITYo ( items ) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc := SYSTEM.CAST (WholeType, items[count]);
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharBaseIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF OCTET );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, last, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   last := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WholeScanSet, FormatIO.ValidWholeFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <=  last DO
    (* go write the array *)   
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # last) THEN
      CharStrIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  last ) THEN CharStrIO.WriteStr (outfile, "\n")  END   
    END; (* if *)
    count := count + 1
  END; (* while *)
   
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)
END WriteFArray;
END OctetArrayIO.