IMPLEMENTATION MODULE TokenIO;

(* RS utilities *)
FROM Transition IMPORT CHARSET;
FROM Transition IMPORT CAPACITYch;
FROM CARDINAL8 IMPORT CARDINAL8;

IMPORT Card8Strs, Transition;

(* R10 Stuff *)
IMPORT ChanIO, CharBaseIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

CONST
  TrimLeading = CHARSET{ "{", " " };
  TrimTrailing = CHARSET{ " ", "}" };
  WholeSeperators = CHARSET { "'" };
  HexSeperators = CHARSET { "'", " " };
  RealSeperators = CHARSET {};
  Signs = CHARSET {"+", "-"};


PROCEDURE SetResult (file: ChanIO.Channel; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code OF
    IOStatus.Success             : ChanIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
    IOStatus.OutOfRange          : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
    IOStatus.WrongFormat         : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
    IOStatus.UnexpectedEndOfLine : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok ) |
    IOStatus.UnexpectedEndOfInput: ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfInput} , ok )
  END; (* case *)
  IF NOT ok THEN
    Perror.prerr ( "Invalid SetResult parameters passed\n")
  END
END SetResult;


PROCEDURE ReadTextToken (file: ChanIO.Channel; VAR str : ARRAY OF CHAR);

  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes text characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)

VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN TrimLeading) *) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed  DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE (ch IN *)

   (* read printable characters *)
  WHILE NOT res.failed AND (ISO646.isPrintable (ch)  ) AND ( ch # ISO646.SP )
      AND (count <= lastToRead) DO
    str[count] := ch;
    INC (count);
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

    (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isPrintable (ch))
           OR res.failed) THEN   (* if room left in string, terminate it *)
    str[count] := CHR (0)
  ELSIF count > lastToRead THEN
   (* now, if the string for the token is not big enough, say so *)
    SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
    WHILE ISO646.isPrintable (ch) AND NOT res.failed
     DO (* and knock the rest out *)
       INC (count);
       CharBaseIO.Read (file, ch);
       res := ChanIO.statusOf (file)
     END (* while *)
  END; (* if count*)

     (* now knock out trailers *)
  WHILE (ch IN TrimTrailing) AND NOT res.failed
    DO (* skip trailing chars *)
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file)
    END; (* WHILE (ch IN *)
        (* now if there are more chars, they belong to the next read so *)
  IF NOT res.failed AND (ISO646.isPrintable (ch) ) THEN
    ChanIO.Insert (file, ch)
  END (* If NOT res.failed *);
END ReadTextToken;


PROCEDURE ReadWholeToken (file: ChanIO.Channel; VAR str: ARRAY OF CHAR; signed : BOOLEAN);
  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)

VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN TrimLeading) AND NOT(ISO646.isDigit (ch) ) AND NOT (ch IN Signs) ) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)
  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE (ch IN *)

   (* check for sign *)
  IF NOT res.failed THEN
    IF ch IN Signs THEN
      IF signed (* we do both signed and unsigned here, per parameter *)  THEN
        str[count] := ch;
        INC(count);
         CharBaseIO.Read (file, ch);
         res := ChanIO.statusOf (file);
         WHILE (ch = " ") AND NOT res.failed (* spaces after sign? *)
           DO (* skip leading spaces *)
             CharBaseIO.Read (file, ch);
             res := ChanIO.statusOf (file)
           END; (* while *)
      ELSE (* sign when not allowed *)
        res.code := IOStatus.WrongFormat;
        res.failed := TRUE
      END (* IF signed *)
    END (*  IF ch IN *)
  END; (* IF NOT res.faile *)

 (* skip any spaces after sign *)
  WHILE (ch = ISO646.SP) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE (ch = ISO646.SP) *)

  WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN WholeSeperators) )
         AND (count <= lastToRead) DO (* read numeric characters *)
    IF NOT (ch IN WholeSeperators) THEN (* ignore digit separator *)
      str[count] := ch;
      INC (count)
    END; (* if *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

      (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch))
             OR res.failed) THEN   (* if room left in string, terminate it *)
  str[count] := CHR (0)
  ELSIF count > lastToRead THEN
    (* now, if the string for the token is not big enough, say so *)
    SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
    WHILE ISO646.isDigit (ch) AND NOT res.failed  DO (* and knock the rest out *)
     INC (count);
     CharBaseIO.Read (file, ch);
     res := ChanIO.statusOf (file)
    END (* while *)
  END; (* if *)

     (* now knock out trailers *)
  WHILE (ch IN TrimTrailing) AND NOT res.failed DO (* skip trailing chars *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE *)
      (* now if there are more chars, they belong to the next read so *)
  IF NOT res.failed AND (ISO646.isPrintable (ch) ) THEN
    ChanIO.Insert (file, ch)
  END; (* If NOT res.failed *)
END ReadWholeToken;


PROCEDURE ReadHexToken (file: ChanIO.Channel; VAR str: ARRAY OF CHAR; signed : BOOLEAN);
  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes hex numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)

VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN TrimLeading) AND NOT(ch = "0") ) THEN
     SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
     RETURN
  END; (* IF res.failed *)

      (* skip leading trim chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* must hold a 0 here *)
  str[count] := ch;
  INC (count);

    (* read second char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT (ch = "x") THEN
     SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
     RETURN
  END; (* IF res.failed  *)
  str[count] := ch;
  INC (count);

  WHILE NOT res.failed AND (ISO646.isHexDigit (ch) OR (ch IN HexSeperators) )
         AND (count <= lastToRead) DO (* read numeric characters *)
    IF NOT (ch IN HexSeperators) THEN (* ignore digit separator *)
      str[count] := ch;
      INC (count)
    END; (* if *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)
  (* Now string must be of even length so the count must also be even at this point *)
  IF ODD (count) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END (* IF ODD (count) *);

      (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch))
             OR res.failed)  THEN   (* if room left in string, terminate it *)
    str[count] := CHR (0)
  ELSIF count > lastToRead THEN
    (* now, if the string for the token is not big enough, say so *)
    SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
    WHILE ISO646.isDigit (ch) AND NOT res.failed DO (* and knock the rest out *)
      INC (count);
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file)
    END (* WHILE ISO646 *)
  END; (* IF (count <= *)

     (* now knock out trailers *)
  WHILE (ch IN TrimTrailing) AND NOT res.failed  DO (* skip trailing chars *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* While *)
        (* now if there are more chars, they belong to the next read so *)
  IF NOT res.failed AND (ISO646.isPrintable (ch) ) THEN
    ChanIO.Insert (file, ch)
  END (* If NOT res.failed *);

END ReadHexToken;


PROCEDURE ReadRealToken (file: ChanIO.Channel; VAR str: ARRAY OF CHAR);
  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)

VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;
  (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN TrimLeading) AND NOT(ISO646.isDigit (ch) ) AND NOT (ch IN Signs) ) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

   (* check for sign *)
  IF NOT res.failed THEN
    IF ch IN Signs THEN
      str[count] := ch;
      INC(count);
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file);
      WHILE (ch = " ") AND NOT res.failed (* spaces after sign? *)  DO (* skip leading spaces *)
        CharBaseIO.Read (file, ch);
        res := ChanIO.statusOf (file)
      END (* while *)
    END (*IF ch IN Signs*)
  END;  (*IF NOT res.failed*)

 (* skip any spaces after sign *)
  WHILE (ch = ISO646.SP) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

    (* first digit string *)
  WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN RealSeperators) ) AND (count <= lastToRead) DO (* read numeric characters *)
    IF NOT (ch IN RealSeperators) THEN  (* ignore digit separator *)
      str[count] := ch;
      INC (count)
    END; (*  IF NOT (ch IN *)
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file)
  END; (* WHILE NOT res.failed *)

      (* then the  decimal *)
  IF NOT res.failed AND (ch = '.') THEN
    str[count] := ch;
    INC(count);
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  ELSE
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
  END; (* IF NOT res.failed *)

        (* second digit string *)
  WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN RealSeperators) ) AND (count <= lastToRead) DO (* read numeric characters *)
    IF NOT (ch IN RealSeperators) THEN  (* ignore digit separator *)
      str[count] := ch;
      INC (count)
    END; (* IF NOT (ch IN*)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE NOT res.failed *)

    (* now there may be an exponent marker *)
  IF NOT res.failed THEN
    IF (ch = 'E') OR (ch = 'e') THEN
      str[count] := ch;
      INC(count);
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file);
      IF NOT res.failed (* sign optional as first char *) THEN
        IF (ch = '-') OR (ch = '+') THEN
          str[count] := ch;
          INC(count);
          CharBaseIO.Read (file, ch);
          res := ChanIO.statusOf (file)
        END (*  IF (ch = '-') *)
      END; (* IF NOT res.failed *)
      (* third digit string for exponent *)
      WHILE NOT res.failed AND ISO646.isDigit (ch) AND (count <= lastToRead) DO (* read numeric characters *)
        IF ch # "'" THEN (* ignore digit separator *)
          str[count] := ch;
          INC (count)
        END; (* IF ch # "'" *)
        CharBaseIO.Read (file, ch);
        res := ChanIO.statusOf (file)
      END (*  WHILE NOT res.failed *)
    END (* IF (ch = 'E') *)
  END; (* IF NOT res.failed *)

  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch)) OR res.failed) THEN  (* if room left in string, terminate it *)
    str[count] := CHR (0)
    ELSIF count > lastToRead THEN
      (* now, if the string for the token is not big enough, say so *)
      SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
      WHILE ISO646.isDigit (ch) AND NOT res.failed DO (* and knock the rest out *)
        INC (count);
        CharBaseIO.Read (file, ch);
        res := ChanIO.statusOf (file)
      END (* while *)
  END; (* IF (count <= *)

     (* now knock out trailers *)
  WHILE (ch IN TrimTrailing) AND NOT res.failed DO (* skip trailing chars *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)
        (* now if there are more chars, they belong to the next read so *)
  IF NOT res.failed AND (ISO646.isPrintable (ch) ) THEN
    ChanIO.Insert (file, ch)
  END (* If NOT res.failed *);
END ReadRealToken;


PROCEDURE LookCharToken (file: ChanIO.Channel; VAR ch: CHAR);

VAR
  res : IOStatusT;

BEGIN
  (*read first char*)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN TrimLeading) *) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE (ch IN TrimLeading) *)
  ChanIO.Insert (file, ch) (* re enter it *)
    (* just go back holding the first non blank char with no skip; ChanIO will have set result  *)

END LookCharToken;


PROCEDURE ReadCharToken (file: ChanIO.Channel; VAR ch: CHAR);

VAR
  res : IOStatusT;

BEGIN
  (*read first char*)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) OR NOT(ch IN TrimLeading)  THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* WHILE (ch IN TrimLeading) *)
    (* just go back holding the first non blank char with no skip; ChanIO will have set result  *)
END ReadCharToken;


PROCEDURE ReadCharLiteralToken (file: ChanIO.Channel; VAR ch: CHAR);

VAR
  res : IOStatusT;
  resC : IOStatusT;
  code : IOStatus.Code;
  loChar, hiChar : CHAR;
  lo, hi, value : CARDINAL8;

BEGIN
  (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN TrimLeading) *) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

  IF NOT res.failed AND (ch = "0") THEN (* #1 *)
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file);
    IF NOT res.failed AND (ch = "u") THEN (* #2 *)
      CharBaseIO.Read (file, hiChar);
      res := ChanIO.statusOf (file);
      IF NOT res.failed AND (ISO646.isHexDigit (hiChar) ) THEN (* #3 *)
        CharBaseIO.Read (file, loChar);
        res := ChanIO.statusOf (file);
        IF NOT res.failed AND (ISO646.isHexDigit ( loChar) ) THEN (* #4 *)
          code := IOStatus.Success;
          Card8Strs.HexCharToCard8 (hiChar, hi, resC);
          Card8Strs.HexCharToCard8 (loChar, lo, resC);
          value := 16*hi + lo;
          ch := CHAR (value);
        ELSE
          code := IOStatus.WrongFormat;
          ch := ISO646.NUL
        END; (* IF NOT res.failed #4 *)
      ELSE
        code := IOStatus.WrongFormat;
        ch := ISO646.NUL
      END; (* IF NOT res.failed AN - #3 *)
    ELSE
      code := IOStatus.WrongFormat;
      ch := ISO646.NUL
    END; (* IF NOT res.failed #2 *) 
  ELSE
    code := IOStatus.WrongFormat;
    ch := ISO646.NUL
  END; (* IF NOT res.failed #1 *) 
  res.code := code;
  SetResult (file, res )
END ReadCharLiteralToken;


PROCEDURE ReadBitsetToken (file: ChanIO.Channel; VAR str : ARRAY OF CHAR);

CONST
  LitCharsAllowed = CHARSET{"0".."9", " ", ","};
  BitCharsAllowed = CHARSET{"0", "1", ".", " "};

VAR
  res : IOStatusT;
  ch : CHAR;
  isLiteralSetToken, done : BOOLEAN;
  sCount : CARDINAL;
  code : IOStatus.Code;

BEGIN (* ReadBitsetToken *)
  sCount := 0;
   (* read first char *)
  CharBaseIO.Read (file, ch);
  res := ChanIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) OR NOT(ch IN TrimLeading)  THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END; (* IF res.failed *)

  (* skip leading chars *)
  WHILE (ch IN TrimLeading) AND NOT res.failed DO
    CharBaseIO.Read (file, ch);
    res := ChanIO.statusOf (file)
  END; (* while *)

  IF NOT res.failed  THEN
    IF ( ch = "{" ) THEN
      isLiteralSetToken := TRUE;
      str[sCount] := ch;
      sCount := sCount + 1;
      CharBaseIO.Read (file, ch);
      res := ChanIO.statusOf (file)
    END;  (* IF ( ch = "{" ) *)

    REPEAT
      IF NOT res.failed THEN
        IF isLiteralSetToken THEN
          IF ch IN LitCharsAllowed THEN
            str[sCount] := ch;
          ELSIF ( ch = "}" )THEN
            str[sCount] := ch;
            sCount := sCount + 1;
            done := TRUE;
          ELSE (* ( isLiteralSetToken ) but  NOT (ch IN LitCharsAllowed) and also not {*)
            code := IOStatus.WrongFormat;
            str[0] := ISO646.NUL;
            done := TRUE
          END (* IF ch IN LitCharsAllowed *)
        ELSIF (* a bin string and *) ch IN BitCharsAllowed  THEN
          str[sCount] := ch;
        ELSE  (* a bin string but  ch not  IN BitCharsAllowed *)
          sCount := sCount + 1;
          done := TRUE
        END;  (* IF isLiteralSetToken *)

        IF NOT done THEN
          sCount := sCount + 1;
          CharBaseIO.Read (file, ch);
          res := ChanIO.statusOf (file)
        END (* IF NOT done *)
      ELSIF (res.code = IOStatus.UnexpectedEndOfLine) OR (res.code = IOStatus.UnexpectedEndOfInput) THEN
        done := TRUE;
      ELSE (* res no good but not end *)
        code := IOStatus.WrongFormat;
        str[0] := ISO646.NUL;
        done := TRUE
      END;  (* IF res *)
    UNTIL done;
  END; (* IF NOT res.failed *)

  res.code := code;
  SetResult (file, res );
  str[sCount] := ISO646.NUL;

END ReadBitsetToken;

END TokenIO.