(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE WordBaseIO (*[ProtoIO] FOR WORD*);
  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as CardinalIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite
    RS    2017 08 04 Removed SYSTEM.CAST in favour of my own cast

  *)
IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT CAPACITYw;


(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenIO, FormatIO;
IMPORT OctetStrs, WholeStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrIO, CharBaseIO, ISO646, ARRAYOFCHAR, CharArray;
IMPORT ChanIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARDINAL;

CONST
  StrToWholeType = WholeStrs.HexStrToCard;
  WholeTypeToDecStr = WholeStrs.CardToDecStr;
  WholeTypeToHexStr = WholeStrs.CardToHexStr;
  WholeTypeToBinStr = WholeStrs.CardToBinStr;


PROCEDURE SetResult (file: ChanIO.Channel; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code OF
    IOStatus.Success    : ChanIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
    IOStatus.OutOfRange  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
    IOStatus.WrongFormat  : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
    IOStatus.StrEmpty   : ChanIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
  END; (* case *)

  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
      END
END SetResult;


PROCEDURE (*[READ]*) Read
  ( infile : ChanIO.Channel; VAR w : UNSAFE.WORD );

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;
  nc : WholeType;

BEGIN (* Read *)
  TokenIO.ReadHexToken (infile, temp, FALSE);        (* get the hex as a string *)
  StrToWholeType (temp, nc, res);       (* convert the string to a CARDINAL *)
  w := SYSTEM.CAST (UNSAFE.WORD, nc);
  SetResult (infile, res);
END Read;



PROCEDURE (*[WRITE]*) Write
  ( outfile : ChanIO.Channel; w : UNSAFE.WORD );

VAR
  temp : FormatIO.Str;
  card : WholeType;
  
BEGIN (* Write *)
  card := SYSTEM.CAST (WholeType, w);
  WholeTypeToHexStr (card, temp);            (* convert the WholeType to a string *)
  CharStrIO.WriteStr (outfile, temp);
END Write;


PROCEDURE WriteFWithfsr ( outfile : ChanIO.Channel; fsr : FormatScanRec; items :  ARRAY OF UNSAFE.WORD );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  numItems := CAPACITYw (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := SYSTEM.CAST (WholeType, items[count]);
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrIO.WriteStr (outfile, "\n")  END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : ChanIO.Channel;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST >0*) ARRAY OF UNSAFE.WORD );

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WordScanSet, FormatIO.ValidWordFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END WordBaseIO.