(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE AddressArrayFDIO (*[ProtoArrayIO] FOR ADDRESS*);

(* IO Extension Library for Array Types of ADDRESS *)

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as AddressIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite
    RS    2017 10 19  removed BARE, updated 
 *)
IMPORT SYSTEM, UNSAFE;
FROM Transition IMPORT OCTET;
FROM Transition IMPORT CAPACITYad, CAPACITYao;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT FormatIO;
IMPORT OctetStrs, WholeStrs;

(* R10 Stuff *)
IMPORT CharStrFDIO, CharBaseFDIO, AddressBaseFDIO, Perror, ISO646, ARRAYOFCHAR, CharArray;

(* R10 Stuff with mods *)
IMPORT FileDescIO;


CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;
  NumeralFormat = FormatIO.NumeralFormat;

TYPE
  WholeType = CARDINAL;

CONST
  StrToWholeType = WholeStrs.HexStrToCard;
  WholeTypeToDecStr = WholeStrs.CardToDecStr;
  WholeTypeToHexStr = WholeStrs.CardToHexStr;
  WholeTypeToBinStr = WholeStrs.CardToBinStr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : FileDescIO.File; VAR array : ARRAY OF UNSAFE.ADDRESS );

VAR
  count, last : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  last := CAPACITYad (array) - 1;
  WHILE count <= last
    DO
      AddressBaseFDIO.Read (infile, array [count] );
      status := FileDescIO.statusOf (infile);
      count := count + 1;
      IF NOT status.failed THEN RETURN END
    END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : FileDescIO.File; (*CONST*) array : ARRAY OF UNSAFE.ADDRESS );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYad (array) - 1;
  (* open brace for the array *)
  CharStrFDIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <= last DO
    AddressBaseFDIO.Write (outfile, array [count]);
    IF count # last THEN
        CharStrFDIO.WriteStr (outfile, ", ")
    END; (*IF*)
    count := count + 1;
  END; (* while *)
  (* close brace for the array *)
  CharStrFDIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items :  ARRAY OF UNSAFE.ADDRESS );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  numItems := CAPACITYad (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := SYSTEM.CAST (WholeType, items[count]);
    temp[0] := ISO646.NUL;
       (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Address to a string *)
    END; (*IF*)
         (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);

    (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END (*WHILE count*);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)
END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : FileDescIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF UNSAFE.ADDRESS );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
  numItems := CAPACITYao (arrays);
  colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.AddressScanSet, FormatIO.ValidAddressFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= numItems DO
    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # numItems) THEN
      CharStrFDIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END; (* if *)
    count := count + 1
  END; (* while *)

  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)
END WriteFArray;

END AddressArrayFDIO.