(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE BooleanArrayFDIO (*[ProtoArrayIO] FOR BOOLEAN*);

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as BooleanIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite


  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYo, CAPACITYao;
FROM Transition IMPORT  CHARSET;
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

(* RS utilities *)
IMPORT TokenFDIO, FormatIO;
IMPORT OctetStrs, WholeStrs;

(* R10 Stuff *)
IMPORT CharStrFDIO, CharBaseFDIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT FileDescIO, BooleanBaseFDIO;


CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  BoolType = BOOLEAN;

CONST
  trueStr = "TRUE";
  falseStr = "FALSE";
  yesStr = "YES";
  noStr = "NO";
  trueChars = CHARSET{"t","T","y","Y"};
  falseChars = CHARSET{"f","F","n","N"};


PROCEDURE BoolToStr (item : BOOLEAN; yesFormat : BOOLEAN ) : FormatIO.Str;

BEGIN
  IF yesFormat THEN
    IF item THEN
      RETURN yesStr
    ELSE
      RETURN noStr
    END
   ELSE (* use true format *)
     IF item THEN
       RETURN trueStr
     ELSE
       RETURN falseStr
     END
  END
END BoolToStr;



PROCEDURE (*[READ]*) ReadArray
  ( infile : FileDescIO.File; VAR array : ARRAY OF BOOLEAN );

VAR
  count : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  WHILE count <= CAPACITYo (array) - 1
    DO
      BooleanBaseFDIO.Read (infile, array [count] );
      status := FileDescIO.statusOf (infile);
      count := count + 1;
      IF NOT status.failed THEN RETURN END
    END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : FileDescIO.File; (*CONST*) array : ARRAY OF BOOLEAN );

VAR
  count, last : CARDINAL;

BEGIN
 last := CAPACITYo (array) - 1;
 (* open brace for the array *)
 CharStrFDIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <=  last DO
    BooleanBaseFDIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrFDIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrFDIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items :  ARRAY OF BoolType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;

BEGIN
  numItems := CAPACITYo (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
      (* convert to the appropriate string *)
    temp := BoolToStr (items[count], fsr.yes4Bool);

    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : FileDescIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF BoolType );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   numItems := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.BooleanScanSet, FormatIO.ValidBooleanFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= numItems DO

    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND ((count # numItems) OR NOT lastArray) THEN
      CharStrFDIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END; (* if *)
    count := count + 1;
  END; (* while *)

  (* and if an escaped char is in the fsr, write it *)
      IF lastArray THEN CharStrFDIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *) END
END WriteFArray;

TYPE
  TargetType = BOOLEAN;
  Ary1 = ARRAY [0..0] OF TargetType;
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;

VAR
  lastArray : BOOLEAN; (* this is a kludge anyway so use a global variable to control the following *)


PROCEDURE Write1FArray
  ( outfile : FileDescIO.File; fmtStr : ARRAY OF CHAR; value1 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
    WriteFArray (outfile, fmtStr, value1 )
END Write1FArray;


PROCEDURE Write2FArray
  ( outfile : FileDescIO.File; fmtStr : ARRAY OF CHAR; value1, value2 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value2 )
END Write2FArray;


PROCEDURE Write3FArray
  ( outfile : FileDescIO.File; fmtStr : ARRAY OF CHAR; value1, value2, value3 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  WriteFArray (outfile, fmtStr, value2 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value3 )
END Write3FArray;

BEGIN (* module *)
  lastArray := TRUE;

END BooleanArrayFDIO.