(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE CharBaseFDIO (*[ProtoIO] FOR CHAR*);

(* File-Desc IO Extension Library for Values of Type CHAR *)

IMPORT SYSTEM;

(* R10 stuff *)
IMPORT FileDescIO, CharStrFDIO, ISO646;

(* RS Stuff *)
IMPORT FormatIO;
IMPORT LongWholeStrs;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;


PROCEDURE (*[READ]*) Read
  ( infile : FileDescIO.File; VAR ch : CHAR );

BEGIN
  FileDescIO.ReadOctet ( infile, ch )
END Read;


PROCEDURE (*[WRITE]*) Write
  ( outfile : FileDescIO.File; ch : CHAR );

BEGIN
  FileDescIO.WriteOctet ( outfile, ch )
END Write;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items : ARRAY OF CHAR );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  lCard : LONGCARD;

BEGIN
   numItems := LENGTH ( items )-1;
   colCount := 0;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    IF fsr.numeralFormat = FormatIO.Hex THEN
      lCard := ORD ( items[count] );
      LongWholeStrs. CardToHexStr ( lCard, temp);
    ELSE
      temp[0] := items[count]; temp[1] := ISO646.NUL;
    END; (* if *)
         (* format that char using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

        (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF NOT fsr.isConcat AND (count # numItems) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END(* while *);

  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)
END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteF
  ( outfile      : FileDescIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    items        : (*ARGLIST >0*) ARRAY OF CHAR );

VAR
  fsr : FormatScanRec;

BEGIN
   (* get the format string scanned to a format string record *)
    ScanStrForFormat (fmtStr, FormatIO.CharScanSet, FormatIO.ValidCharFmtChars, fsr);
    WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END CharBaseFDIO.