IMPLEMENTATION MODULE  LongComplexBaseFDIO;

 (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 15  uses new FormatIO module
    RS    2015 05 28  Stomp bugs and refined
    RS      2016 07 26  Now FDIO

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYlz;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFDIO, FormatIO;
IMPORT OctetStrs, LongRealStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrFDIO, CharBaseFDIO, ISO646;

(* R10 Stuff with mods *)
IMPORT FileDescIO;
IMPORT LONGCOMPLEX; FROM LONGCOMPLEX IMPORT LONGCOMPLEXT;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
    RealType = LONGREAL;
    ComplexType = LONGCOMPLEXT;

CONST
  StrToRealType = LongRealStrs.StrToReal;
  RealTypeToStr = LongRealStrs.RealToStr;
  RealTypeToScientific = LongRealStrs.RealToScientific;
  RealTypeToFloat = LongRealStrs.RealToFloat;
  RealTypeToEng = LongRealStrs.RealToEng;
  RealTypeToFixed = LongRealStrs.RealToFixed;
  plusStr = " + "; (* to go between the real components of a complex *)
  minusStr = " - "; (* to go between the real components of a complex *)
  imStr ="i"; (* suffix to write for im component *)


PROCEDURE SetResult (file: FileDescIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FileDescIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)

  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
      END
END SetResult;



PROCEDURE (*. [READ] .*) Read( infile : FileDescIO.File; VAR z : LONGCOMPLEXT );VAR
  temp  : FormatIO.Str;
  res : IOStatusT;
  ch : CHAR;
  re, im : RealType;
  isRawFormat : BOOLEAN;


BEGIN (* Read *)
  z := LONGCOMPLEX.zero;
  TokenFDIO.LookCharToken (infile, ch);
  isRawFormat := (ch = "(" );
  IF isRawFormat THEN
      CharBaseFDIO.Read (infile,ch)
  END;
  TokenFDIO.ReadRealToken (infile, temp);  (* get the first RealType as a string *)
  StrToRealType (temp, re, res);       (* convert the string to a RealType *)
  SetResult (infile, res);
  IF res.failed THEN RETURN END;
  IF isRawFormat THEN
      CharBaseFDIO.Read (infile,ch);
    IF ch # "," THEN
      res := IOStatusT{ TRUE, IOStatus.WrongFormat};
      SetResult (infile, res);
      RETURN
     END
  END;
  TokenFDIO.ReadRealToken (infile, temp);  (* get the second RealType as a string *)
  StrToRealType (temp, im, res);       (* convert the string to a RealType *)
  SetResult (infile, res);
  IF res.failed THEN RETURN END;

  TokenFDIO.ReadCharToken (infile, ch);
  IF (isRawFormat AND ( ch # ")" ) )
    OR (NOT isRawFormat AND (ch # "i") AND (ch # "j")  )(* engineers like the latter *)
  THEN
    res := IOStatusT{ TRUE, IOStatus.WrongFormat};
    SetResult (infile, res);
    RETURN
  END;
  z := LONGCOMPLEX.CMPLX (re, im)
END Read;



PROCEDURE (*. [WRITE] .*) Write( outfile : FileDescIO.File; (*. CONST .*) z : LONGCOMPLEXT );
(* Writes the textual representation of value z to output stream outfile. This
   procedure is substituted for invocations of WRITE with a LONGCOMPLEX argument.*)
VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  RealTypeToScientific (z.re, LongRealStrs.rMaxSigFigs, temp);            (* convert the RealType to a scientific string (default) *)
  CharStrFDIO.WriteStr (outfile, temp);
  IF z.im < 0.0 THEN
    CharStrFDIO.WriteStr (outfile, minusStr);
  ELSE
    CharStrFDIO.WriteStr (outfile, plusStr)
  END;
  RealTypeToScientific (ABS(z.im), LongRealStrs.rMaxSigFigs, temp);            (* convert the RealType to a scientific string (default) *)
  CharStrFDIO.WriteStr (outfile, temp);
  CharStrFDIO.WriteStr (outfile, imStr);

END Write;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items :  ARRAY OF ComplexType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  theNum : ComplexType;


PROCEDURE NumberToNumeralString (theNum: RealType); (*internal *)

BEGIN       (* convert to the appropriate string *)
  IF fsr.numeralFormat = FormatIO.Hex
  THEN
    OctetStrs.OctsToHexStr (theNum, temp);
  ELSIF fsr.numeralFormat = FormatIO.Bin THEN
    OctetStrs.OctsToBinStr (theNum, temp);
  ELSIF fsr.realFormat = FormatIO.Scientific THEN
    RealTypeToScientific (theNum, fsr.numDecPlaces, temp);
  ELSIF fsr.realFormat = FormatIO.Float THEN
    RealTypeToFloat (theNum, fsr.numDecPlaces, fsr.numIntDigitsForFloat, temp);
  ELSIF fsr.realFormat = FormatIO.Fixed THEN
    RealTypeToFixed (theNum, fsr.numDecPlaces, temp);
  ELSIF fsr.realFormat = FormatIO.Eng THEN
    RealTypeToEng (theNum, fsr.numDecPlaces, temp);
  END(*IF*)
END NumberToNumeralString;

BEGIN
  IF fsr.numDecPlaces = 0 THEN fsr.numDecPlaces  := LongRealStrs.rMaxSigFigs END; (* default *)
  last := CAPACITYlz (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    (* open paren if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN CharStrFDIO.WriteStr (outfile, "( ") END;
    theNum := items[count];

     (* do real part *)
    NumberToNumeralString (theNum.re);
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);
    (* comma separator if raw *)
    IF fsr.complexFormat = FormatIO.Raw THEN
      CharStrFDIO.WriteStr (outfile, ", ")
    ELSE
      IF theNum.im < 0.0 THEN
        CharStrFDIO.WriteStr (outfile, minusStr);
      ELSE
        CharStrFDIO.WriteStr (outfile, plusStr)
      END;
    END;

     (* repeat for IM Part *)
    NumberToNumeralString (ABS(theNum.im));
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);
    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);
    CASE fsr.complexFormat OF
     FormatIO.Raw :
       CharStrFDIO.WriteStr (outfile, " )")
     | FormatIO.i :
       CharStrFDIO.WriteStr (outfile, "i")
     | FormatIO.j :
       CharStrFDIO.WriteStr (outfile, "j")
    END; (* case *)

        (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*. [WRITEF] .*) WriteF
  ( outfile      : FileDescIO.File;
    (*. CONST .*) fmtStr : ARRAY OF CHAR;
    items        : (*. ARGLIST .*) ARRAY OF LONGCOMPLEXT );
(* Writes a formatted textual representation of one or more LONGCOMPLEX values to
   output stream outfile. The output format is determined by fmtStr. This
   procedure is substituted for invocations of WRITEF with one or more
   LONGCOMPLEX arguments. *)
VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.ComplexScanSet, FormatIO.ValidComplexFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END LongComplexBaseFDIO.
