(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE LongIntArrayFDIO (*[ProtoArrayIO] FOR LONGINT*);

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as LongIntIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYao, CAPACITYli;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFDIO, FormatIO;
IMPORT LongWholeStrs;

(* R10 Stuff *)
IMPORT CharStrFDIO, CharBaseFDIO, ISO646, CharArray, Perror;

(* R10 Stuff with mods *)
IMPORT FileDescIO, LongIntBaseFDIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = LONGINT;

CONST
  StrToWholeType = LongWholeStrs.StrToInt;
  WholeTypeToDecStr = LongWholeStrs.IntToDecStr;
  WholeTypeToHexStr = LongWholeStrs.CardToHexStr;
  WholeTypeToBinStr = LongWholeStrs.CardToBinStr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : FileDescIO.File; VAR array : ARRAY OF LONGINT );

VAR
  count, last : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  last := CAPACITYli (array) - 1;
  WHILE count <= last
    DO
      LongIntBaseFDIO.Read (infile, array [count] );
      status := FileDescIO.statusOf (infile);
      count := count + 1;
      IF NOT status.failed THEN RETURN END
    END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : FileDescIO.File; (*CONST*) array : ARRAY OF LONGINT );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYli (array) - 1;
  (* open brace for the array *)
 CharStrFDIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <=  last DO
    LongIntBaseFDIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrFDIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrFDIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items :  ARRAY OF WholeType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  last := CAPACITYli (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc := items[count];
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the LongInt to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharBaseFDIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : FileDescIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF WholeType );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, last, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   last := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WholeScanSet, FormatIO.ValidWholeFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <=  last DO
    (* go write the array *)   
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # last) THEN
      CharStrFDIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  last ) THEN CharStrFDIO.WriteStr ( outfile, "\n" ) END   
    END; (* if *)
    count := count + 1
  END; (* while *)
   
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)
END WriteFArray;
END LongIntArrayFDIO.