IMPLEMENTATION MODULE  OctetBaseFDIO;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as OctetIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite


  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYo;
FROM Transition IMPORT OCTET;
FROM CARDINAL8 IMPORT CARDINAL8;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFDIO, FormatIO;
IMPORT Card8Strs;

(* R10 Stuff *)
IMPORT CharStrFDIO, CharBaseFDIO, ISO646, CharArray, Perror;

(* R10 Stuff with mods *)
IMPORT FileDescIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARDINAL8;

CONST
  StrToWholeType = Card8Strs.StrToCard8;
  WholeTypeToDecStr = Card8Strs.Card8ToDecStr;
  WholeTypeToHexStr = Card8Strs.Card8ToHexStr;
  WholeTypeToBinStr = Card8Strs.Card8ToBinStr;


PROCEDURE SetResult (file: FileDescIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FileDescIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)

  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed" )
      END
END SetResult;


PROCEDURE Read ( infile : FileDescIO.File; VAR n : OCTET );
  (* Skips leading spaces, and removes any remaining characters from file that form part of an unsigned whole number.  The value of this number is assigned to card.  The read result is set to the value Success, OutOfRange, WrongFormat, endOfLine, or EndOfInput.  *)

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;
  nc : WholeType;

BEGIN (* Read *)
  TokenFDIO.ReadWholeToken (infile, temp, FALSE);        (* get the usigned as a string *)
  StrToWholeType (temp, nc, res);       (* convert the string to a CARDINAL *)
  IF nc > 255 THEN res := IOStatusT{ TRUE, IOStatus.OutOfRange}; END; (*IF*)
  SetResult (infile, res);
  n := SYSTEM.CAST (OCTET, nc
END Read;



PROCEDURE Write ( outfile : FileDescIO.File; n : OCTET );
  (* Writes the value of card to cid in text form, in a field of the given
minimum width. *)
VAR
  temp : FormatIO.Str;
  nc : WholeType;

BEGIN (* Write *)
  nc := SYSTEM.CAST (WholeType,n);
  WholeTypeToDecStr (nc, temp);    (* convert the OCTET to a string *)
  CharStrFDIO.WriteStr (outfile, temp)
END Write;


PROCEDURE WriteFWithfsr ( outfile : FileDescIO.File; fsr : FormatScanRec; items :  ARRAY OF OCTET );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  last := CAPACITYo ( items ) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc := SYSTEM.CAST (WholeType, items[count]);
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFDIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFDIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharBaseFDIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFDIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFDIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE WriteF ( outfile      : FileDescIO.File;
                   (*. CONST .*) fmtStr : ARRAY OF CHAR;
                   items        : (*. ARGLIST .*) ARRAY OF OCTET );

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WholeScanSet, FormatIO.ValidWholeFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END OctetBaseFDIO.

