IMPLEMENTATION MODULE TokenFDIO;

(* RS utilities *)
FROM Transition IMPORT CHARSET;
FROM Transition IMPORT CAPACITYch;
FROM CARDINAL8 IMPORT CARDINAL8;

IMPORT Card8Strs, Transition;

(* R10 Stuff *)
IMPORT FileDescIO, CharBaseFDIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;

CONST
    trimLeading = CHARSET{ "{", " " };
    trimTrailing = CHARSET{ " ", "}" };
    wholeSeperators = CHARSET { "'" };
    hexSeperators = CHARSET { "'", " " };
    realSeperators = CHARSET {};
    signs = CHARSET {"+", "-"};

PROCEDURE SetResult (file: FileDescIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success             : FileDescIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange          : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat         : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.UnexpectedEndOfLine : FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok ) |
      IOStatus.UnexpectedEndOfInput: FileDescIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfInput} , ok )
    END; (* case *)
  IF NOT ok
    THEN
      Perror.prerr ( "Invalid SetResult parameters passed\n")
  END
END SetResult;


PROCEDURE ReadTextToken (file: FileDescIO.File; VAR str : ARRAY OF CHAR);

  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes text characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)
VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN trimLeading)*)
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

   (* read printable characters *)
  WHILE NOT res.failed AND (ISO646.isPrintable (ch)  ) AND ( ch # ISO646.SP )
      AND (count <= lastToRead) DO
    str[count] := ch;
    INC (count);
    CharBaseFDIO.Read (file, ch);
    res := FileDescIO.statusOf (file)
  END; (* while *)

    (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isPrintable (ch))
           OR res.failed) THEN   (* if room left in string, terminate it *)
    str[count] := 0C
  ELSIF count > lastToRead THEN
   (* now, if the string for the token is not big enough, say so *)
    SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
    WHILE ISO646.isPrintable (ch) AND NOT res.failed
     DO (* and knock the rest out *)
       INC (count);
       CharBaseFDIO.Read (file, ch);
        res := FileDescIO.statusOf (file)
     END (* while *)
  END; (* if *)

     (* now knock out trailers *)
  WHILE (ch IN trimTrailing) AND NOT res.failed
    DO (* skip trailing chars *)
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
        (* now if there are more chars, they belong to the next read so *)
IF NOT res.failed AND (ISO646.isPrintable (ch)  )
  THEN
    FileDescIO.SetPos (file, FileDescIO.currentPos (file) - 1 )
  END;

END ReadTextToken;


PROCEDURE ReadWholeToken (file: FileDescIO.File; VAR str: ARRAY OF CHAR; signed : BOOLEAN);

  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)
VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN trimLeading) AND NOT(ISO646.isDigit (ch) ) AND NOT (ch IN signs) )
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;
  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

   (* check for sign *)
  IF NOT res.failed THEN
    IF ch IN signs
      THEN
        IF signed (* we do both signed and unsigned here, per parameter *)
          THEN
            str[count] := ch;
            INC(count);
             CharBaseFDIO.Read (file, ch);
             res := FileDescIO.statusOf (file);
             WHILE (ch = " ") AND NOT res.failed (* spaces after sign? *)
               DO (* skip leading spaces *)
                 CharBaseFDIO.Read (file, ch);
                 res := FileDescIO.statusOf (file)
               END;
          ELSE (* sign when not allowed *)
            res.code := IOStatus.WrongFormat;
            res.failed := TRUE
          END
      END
  END;

 (* skip any spaces after sign *)
  WHILE (ch = ISO646.SP) AND NOT res.failed
    DO
       CharBaseFDIO.Read (file, ch);
       res := FileDescIO.statusOf (file)
  END;

 WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN wholeSeperators) )
         AND (count <= lastToRead)
    DO (* read numeric characters *)
      IF NOT (ch IN wholeSeperators) THEN (* ignore digit separator *)
        str[count] := ch;
        INC (count)
      END;
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

      (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch))
             OR res.failed) THEN   (* if room left in string, terminate it *)
  str[count] := 0C
  ELSIF count > lastToRead THEN
   (* now, if the string for the token is not big enough, say so *)
   SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
   WHILE ISO646.isDigit (ch) AND NOT res.failed
     DO (* and knock the rest out *)
      INC (count);
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END
  END;

     (* now knock out trailers *)
  WHILE (ch IN trimTrailing) AND NOT res.failed
    DO (* skip trailing chars *)
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
        (* now if there are more chars, they belong to the next read so *)
IF NOT res.failed AND (ISO646.isPrintable (ch)  )
  THEN
    FileDescIO.SetPos (file, FileDescIO.currentPos (file) - 1 )
  END
END ReadWholeToken;


PROCEDURE ReadHexToken (file: FileDescIO.File; VAR str: ARRAY OF CHAR; signed : BOOLEAN);
  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes hex numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)
VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;

  (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN trimLeading) AND NOT(ch = "0") )
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

      (* skip leading trim chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END; (* must hold a 0 here *)
  str[count] := ch;
  INC (count);

    (* read second char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT (ch = "x")
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;
  str[count] := ch;
  INC (count);

  WHILE NOT res.failed AND (ISO646.isHexDigit (ch) OR (ch IN hexSeperators) )
         AND (count <= lastToRead) DO (* read numeric characters *)
    IF NOT (ch IN hexSeperators) THEN (* ignore digit separator *)
      str[count] := ch;
      INC (count)
    END; (* if *)
    CharBaseFDIO.Read (file, ch);
    res := FileDescIO.statusOf (file)
  END; (* while *)
  (* Now string must be of even length so the count must also be even at this point *)
  IF ODD (count) THEN
    SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
    RETURN
  END;

      (* clean up and terminate string for return *)
  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch))
             OR res.failed)
     THEN   (* if room left in string, terminate it *)
       str[count] := 0C
     ELSIF count > lastToRead THEN
       (* now, if the string for the token is not big enough, say so *)
       SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
       WHILE ISO646.isDigit (ch) AND NOT res.failed
        DO (* and knock the rest out *)
      INC (count);
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
        END
     END;

     (* now knock out trailers *)
  WHILE (ch IN trimTrailing) AND NOT res.failed
    DO (* skip trailing chars *)
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
        (* now if there are more chars, they belong to the next read so *)
IF NOT res.failed AND (ISO646.isPrintable (ch)  )
  THEN
    FileDescIO.SetPos (file, FileDescIO.currentPos (file) - 1 )
  END;

END ReadHexToken;


PROCEDURE ReadRealToken (file: FileDescIO.File; VAR str: ARRAY OF CHAR);
  (* Pre: CAPACITYch (str) - 1 > 0
  Skips leading spaces, and then removes numeric characters from the input stream file before the next space or line mark, copying to str as many as can be accommodated as a string value. The read result is set to the value Success, OutOfRange, UnexpectedEndOfLine, or UnexpectedEndOfInput.
  *)
VAR
  lastToRead, count : CARDINAL;
  ch : CHAR;
  res : IOStatusT;

BEGIN
  count := 0;
  lastToRead := CAPACITYch (str) - 1;
  (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch)) OR (NOT(ch IN trimLeading) AND NOT(ISO646.isDigit (ch) ) AND NOT (ch IN signs) )
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

   (* check for sign *)
  IF NOT res.failed THEN
    IF ch IN signs
      THEN
      str[count] := ch;
      INC(count);
       CharBaseFDIO.Read (file, ch);
       res := FileDescIO.statusOf (file);
       WHILE (ch = " ") AND NOT res.failed (* spaces after sign? *)
         DO (* skip leading spaces *)
         CharBaseFDIO.Read (file, ch);
         res := FileDescIO.statusOf (file)
         END
          END (*IF ch IN signs*)
   END;  (*IF NOT res.failed*)

 (* skip any spaces after sign *)
  WHILE (ch = ISO646.SP) AND NOT res.failed
    DO
       CharBaseFDIO.Read (file, ch);
       res := FileDescIO.statusOf (file)
  END;

    (* first digit string *)
  WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN realSeperators) ) AND (count <= lastToRead)
    DO (* read numeric characters *)
      IF NOT (ch IN realSeperators) THEN  (* ignore digit separator *)
      str[count] := ch;
      INC (count)
      END;
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

      (* then the  decimal *)
   IF NOT res.failed AND (ch = '.')
        THEN
            str[count] := ch;
            INC(count);
            CharBaseFDIO.Read (file, ch);
            res := FileDescIO.statusOf (file)
        ELSE
          SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
        END;

        (* second digit string *)
   WHILE NOT res.failed AND (ISO646.isDigit (ch) OR (ch IN realSeperators) ) AND (count <= lastToRead)
    DO (* read numeric characters *)
      IF NOT (ch IN realSeperators) THEN  (* ignore digit separator *)
      str[count] := ch;
      INC (count)
      END;
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

    (* now there may be an exponent marker *)
  IF NOT res.failed
    THEN
      IF (ch = 'E') OR (ch = 'e')
        THEN
            str[count] := ch;
            INC(count);
            CharBaseFDIO.Read (file, ch);
            res := FileDescIO.statusOf (file);
            IF NOT res.failed (* sign optional as first char *)
                THEN
                  IF (ch = '-') OR (ch = '+')
                    THEN
                        str[count] := ch;
                        INC(count);
                        CharBaseFDIO.Read (file, ch);
                        res := FileDescIO.statusOf (file)
                    END
                END;
            (* third digit string for exponent *)
            WHILE NOT res.failed AND ISO646.isDigit (ch) AND (count <= lastToRead)
                DO (* read numeric characters *)
                 IF ch # "'" THEN (* ignore digit separator *)
                    str[count] := ch;
                    INC (count)
                 END;
        CharBaseFDIO.Read (file, ch);
        res := FileDescIO.statusOf (file)
             END
        END
    END;

  IF (count <= lastToRead) AND (NOT(ISO646.isDigit (ch)) OR res.failed)
     THEN   (* if room left in string, terminate it *)
       str[count] := 0C
     ELSIF count > lastToRead THEN
       (* now, if the string for the token is not big enough, say so *)
       SetResult ( file, IOStatusT{ TRUE, IOStatus.OutOfRange} );
       WHILE ISO646.isDigit (ch) AND NOT res.failed
        DO (* and knock the rest out *)
          INC (count);
          CharBaseFDIO.Read (file, ch);
        res := FileDescIO.statusOf (file)
        END
     END;

     (* now knock out trailers *)
  WHILE (ch IN trimTrailing) AND NOT res.failed
    DO (* skip trailing chars *)
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
        (* now if there are more chars, they belong to the next read so *)
IF NOT res.failed AND (ISO646.isPrintable (ch)  )
  THEN
    FileDescIO.SetPos (file, FileDescIO.currentPos (file) - 1 )
  END;

END ReadRealToken;


PROCEDURE LookCharToken (file: FileDescIO.File; VAR ch: CHAR);

VAR
  res : IOStatusT;

BEGIN
  (*read first char*)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN trimLeading)*)
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
  FileDescIO.Insert (file, ch) (* re enter it *)
    (* just go back holding the first non blank char with no skip; FileDescIO will have set result  *)

END LookCharToken;



PROCEDURE ReadCharToken (file: FileDescIO.File; VAR ch: CHAR);

VAR
  res : IOStatusT;

BEGIN
  (*read first char*)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) (*OR NOT(ch IN trimLeading)*)
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;
    (* just go back holding the first non blank char with no skip; FileDescIO will have set result  *)
END ReadCharToken;


PROCEDURE ReadCharLiteralToken (file: FileDescIO.File; VAR ch: CHAR);

VAR
  res : IOStatusT;
  resC : IOStatusT;
  code : IOStatus.Code;
  loChar, hiChar : CHAR;
  lo, hi, value : CARDINAL8;

BEGIN
  (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) OR NOT(ch IN trimLeading)
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

  IF NOT res.failed AND (ch = "0") THEN
    CharBaseFDIO.Read (file, ch);
    res := FileDescIO.statusOf (file);
    IF NOT res.failed AND (ch = "u") THEN
      CharBaseFDIO.Read (file, hiChar);
      res := FileDescIO.statusOf (file);
      IF NOT res.failed AND (ISO646.isHexDigit (hiChar) ) THEN
        CharBaseFDIO.Read (file, loChar);
        res := FileDescIO.statusOf (file);
        IF NOT res.failed AND (ISO646.isHexDigit ( loChar) ) THEN
          code := IOStatus.Success;
          Card8Strs.HexCharToCard8 (hiChar, hi, resC);
          Card8Strs.HexCharToCard8 (loChar, lo, resC);
          value := 16*hi + lo;
          ch := CHAR (value);
        ELSE
          code := IOStatus.WrongFormat;
          ch := ISO646.NUL
        END;
        ELSE
        code := IOStatus.WrongFormat;
        ch := ISO646.NUL
        END;
      ELSE
        code := IOStatus.WrongFormat;
        ch := ISO646.NUL
      END;
    ELSE
      code := IOStatus.WrongFormat;
      ch := ISO646.NUL
    END;
    res.code := code;
  SetResult (file, res )
END ReadCharLiteralToken;


PROCEDURE ReadBitsetToken (file: FileDescIO.File; VAR str : ARRAY OF CHAR);
CONST
  litCharsAllowed = CHARSET{"0".."9", " ", ","};
  binCharsAllowed = CHARSET{"0", "1", ".", " "};

VAR
  res : IOStatusT;
  ch : CHAR;
  isLiteralSetToken, done : BOOLEAN;
  sCount : CARDINAL;
  code : IOStatus.Code;

BEGIN (* ReadBitsetToken *)
  sCount := 0;
   (* read first char *)
  CharBaseFDIO.Read (file, ch);
  res := FileDescIO.statusOf (file);

  IF res.failed OR NOT(ISO646.isPrintable (ch) ) OR NOT(ch IN trimLeading)
    THEN
       SetResult ( file, IOStatusT{ TRUE, IOStatus.WrongFormat} );
       RETURN
    END;

  (* skip leading chars *)
  WHILE (ch IN trimLeading) AND NOT res.failed
    DO
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;

  IF NOT res.failed  THEN
    IF ( ch = "{" ) THEN
      isLiteralSetToken := TRUE;
      str[sCount] := ch;
      sCount := sCount + 1;
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
    END;  (* IF ( ch = "{" ) *)

    REPEAT
    IF NOT res.failed THEN
      IF isLiteralSetToken THEN
        IF ch IN litCharsAllowed THEN
          str[sCount] := ch;
        ELSIF ( ch = "}" )THEN
          str[sCount] := ch;
          sCount := sCount + 1;
          done := TRUE;
        ELSE (* ( isLiteralSetToken ) but  NOT (ch IN litCharsAllowed) and also not {*)
          code := IOStatus.WrongFormat;
          str[0] := ISO646.NUL;
          done := TRUE
        END (* IF ch IN litCharsAllowed *)
      ELSIF (* a bin string and *) ch IN binCharsAllowed  THEN
      str[sCount] := ch;
      ELSE  (* a bin string but  ch not  IN binCharsAllowed *)
        sCount := sCount + 1;
      done := TRUE
      END;  (* IF isLiteralSetToken *)

      IF NOT done THEN
      sCount := sCount + 1;
      CharBaseFDIO.Read (file, ch);
      res := FileDescIO.statusOf (file)
      END (* IF NOT done *)
    ELSIF (res.code = IOStatus.UnexpectedEndOfLine) OR (res.code = IOStatus.UnexpectedEndOfInput) THEN
      done := TRUE;
    ELSE (* res no good but not end *)
      code := IOStatus.WrongFormat;
      str[0] := ISO646.NUL;
      done := TRUE
      END;  (* IF res *)
     UNTIL done;
  END; (* IF NOT res.failed *)

  res.code := code;
  SetResult (file, res );
  str[sCount] := ISO646.NUL;

END ReadBitsetToken;

END TokenFDIO.