(* (C) 2009-2015 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE BitsetArrayFIO (*. [ProtoIO] FOR BITSET .*);
(* IO Extension Library for Type BITSET *)
  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      2015 06 034  First implementation

  *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYb, CAPACITYao;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT OctetStrs, BitsetStr;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, BitsetBaseFIO, Perror, ISO646;

(* R10 Stuff with mods *)
IMPORT FilePtrIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;
  BitsetFormat = FormatIO.BitsetFormat;

TYPE
    BitsetType = BITSET;

CONST
  BitStrToBitsetType = BitsetStr.BitStrToBitset;
  LitStrToBitsetType = BitsetStr.LitBitsetStrToBitset;
  BitsetTypeToBinStr = BitsetStr.BitsetToBinStr;
  BitsetTypeToLitStr = BitsetStr.BitsetToLitStr;


PROCEDURE SetResult (file: FilePtrIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FilePtrIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)
  
  IF NOT ok THEN
      Perror.prerr ( "Invalid SetResult parameters passed\n" )
  END (* if *)
END SetResult;
(* Here starts the implementation of the procedures defined in the definition part. *)

PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF BitsetType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : BitsetType;

BEGIN
  last := CAPACITYb (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc := items[count];
       (* convert to the appropriate string *)
    IF fsr.bitsetFormat = FormatIO.Literal THEN
      BitsetTypeToLitStr (nc, temp);
           (* format that converted string using the fsr already parsed *)
      temp :=  FormatTextStr (temp, fsr);
    ELSE
      BitsetTypeToBinStr (nc, temp); (* convert the bitset to a string *)
           (* format that converted string using the fsr already parsed *)
      temp :=  FormatNumStr (temp, fsr); (* does not include the braces *)

    END; (*IF*)
     IF NOT (fsr.bitsetFormat = FormatIO.Literal) THEN CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr) END;
       (* write out the formatted string *)
     CharStrFIO.WriteStr (outfile, temp);
     IF NOT (fsr.bitsetFormat = FormatIO.Literal) THEN CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr) END;

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END (* if *);
    count := count + 1
  END (* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)
END WriteFWithfsr;


PROCEDURE ReadArray ( infile : FilePtrIO.File; VAR items : ARRAY OF BITSET );

VAR
  count, last : CARDINAL;

BEGIN (* ReadArray *)
  last := CAPACITYb(items) - 1;
  count := 0;
  WHILE count <= last DO
    BitsetBaseFIO.Read (infile, items [count]);
    count := count + 1
  END
END ReadArray;


PROCEDURE WriteArray ( outfile : FilePtrIO.File; items : ARRAY OF BITSET );

VAR
  count, last : CARDINAL;

BEGIN (* WriteArray *)
  last := CAPACITYb (items) -1;
  CharStrFIO.WriteStr (outfile, "{ ");
  count := 0;WHILE count <=  last DO
    BitsetBaseFIO.Write (outfile, items [count]);
    IF count < last THEN
        CharStrFIO.WriteStr (outfile, ", ")
    END; (*IF*)
    count := count + 1;
  END; (* while *)
  CharStrFIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFArray
  ( outfile : FilePtrIO.File; fmtStr : ARRAY OF CHAR; arrays : (*. ARGLIST .*) ARRAY OF ARRAY OF BITSET );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   numItems := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.BitsetScanSet, FormatIO.ValidBitsetFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= numItems DO
    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # numItems) THEN
      CharStrFIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END;
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1;
  END (* while *);

  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)

END WriteFArray;


END BitsetArrayFIO.