(* (C) 2009-2015 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE BitsetBaseFIO (*. [ProtoIO] FOR BITSET .*);
(* IO Extension Library for Type BITSET *)
  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      2015 06 034  First implementation

  *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYb;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT OctetStrs, BitsetStr;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT FilePtrIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;
  BitsetFormat = FormatIO.BitsetFormat;

TYPE
  BitsetType = BITSET;

CONST
  BitStrToBitsetType = BitsetStr.BitStrToBitset;
  LitStrToBitsetType = BitsetStr.LitBitsetStrToBitset;
  BitsetTypeToBinStr = BitsetStr.BitsetToBinStr;
  BitsetTypeToLitStr = BitsetStr.BitsetToLitStr;


PROCEDURE SetResult (file: FilePtrIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FilePtrIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)
  
  IF NOT ok THEN
      Perror.prerr ( "Invalid SetResult parameters passed\n" )
  END (* if *)
END SetResult;
(* Here starts the implementation of the procedures defined in the definition part. *)

PROCEDURE (*. [READ] .*) Read ( infile : FilePtrIO.File; VAR B : BITSET );

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;

BEGIN (* Read *)
  TokenFIO.ReadBitsetToken (infile, temp);

  IF temp[0] = "{" THEN
    LitStrToBitsetType (temp, B, res )
  ELSE
    BitStrToBitsetType  (temp, B, res )
  END;
  SetResult (infile, res)
END Read;



PROCEDURE (*. [WRITE] .*) Write ( outfile : FilePtrIO.File; B : BITSET );

VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  BitsetTypeToLitStr (B, temp);            (* convert the BitsetType to a string *)
  CharStrFIO.WriteStr (outfile, temp);
END Write;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF BitsetType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : BitsetType;

BEGIN
  last := CAPACITYb (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc := items[count];
    temp[0] := ISO646.NUL;
       (* convert to the appropriate string *)
    IF fsr.bitsetFormat = FormatIO.Literal THEN
      BitsetTypeToLitStr (nc, temp);
           (* format that converted string using the fsr already parsed *)
      temp :=  FormatTextStr (temp, fsr);
    ELSE
      BitsetTypeToBinStr (nc, temp); (* convert the bitset to a string *)
           (* format that converted string using the fsr already parsed *)
      temp :=  FormatNumStr (temp, fsr); (* does not include the braces *)

    END; (*IF*)
    IF NOT (fsr.bitsetFormat = FormatIO.Literal) THEN CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr) END;
    (* write out the formatted string *)
    CharStrFIO.WriteStr (outfile, temp);
    IF NOT (fsr.bitsetFormat = FormatIO.Literal) THEN CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr) END;

            (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

         (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END;
      count := count + 1
  END(* while *);


  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*. [WRITEF] .*) WriteF
  ( outfile      : FilePtrIO.File;
    (*. CONST .*) fmtStr : ARRAY OF CHAR;
    items        : (*. ARGLIST .*)ARRAY OF BITSET );

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.BitsetScanSet, FormatIO.ValidBitsetFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END BitsetBaseFIO.