IMPLEMENTATION MODULE  Card8BaseFIO;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as OctetIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite


  *)
IMPORT SYSTEM;
IMPORT CARD8; FROM CARD8 IMPORT CARD8T, CAPACITYc8;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT Card8Strs;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, ISO646, CharArray, Perror;

(* R10 Stuff with mods *)
IMPORT FilePtrIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARD8T;

CONST
  StrToWholeType = Card8Strs.StrToCard8;
  WholeTypeToDecStr = Card8Strs.Card8ToDecStr;
  WholeTypeToHexStr = Card8Strs.Card8ToHexStr;
  WholeTypeToBinStr = Card8Strs.Card8ToBinStr;


PROCEDURE SetResult (file: FilePtrIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FilePtrIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)
  
  IF NOT ok THEN
      Perror.prerr ( "Invalid SetResult parameters passed\n" )
  END (* if *)
END SetResult;


PROCEDURE Read ( infile : FilePtrIO.File; VAR n : CARD8T );
  (* Skips leading spaces, and removes any remaining characters from file that form part of an unsigned whole number.  The value of this number is assigned to card.  The read result is set to the value Success, OutOfRange, WrongFormat, endOfLine, or EndOfInput.  *)

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;
  nc : WholeType;

BEGIN (* Read *)
  TokenFIO.ReadWholeToken (infile, temp, FALSE);        (* get the usigned as a string *)
  StrToWholeType (temp, nc, res);       (* convert the string to a CARDINAL *)
  IF nc > 255 THEN res := IOStatusT{ TRUE, IOStatus.OutOfRange}; END; (*IF*)
  SetResult (infile, res);
  n :=  nc
END Read;



PROCEDURE Write ( outfile : FilePtrIO.File; n : CARD8T );
  (* Writes the value of card to cid in text form, in a field of the given
minimum width. *)
VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  WholeTypeToDecStr (n, temp);            (* convert the CARD8T to a string *)
  CharStrFIO.WriteStr (outfile, temp)
END Write;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF CARD8T );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  last := CAPACITYc8 ( items ) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    nc :=  items[count];
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFIO.WriteStr (outfile, temp);

          (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharBaseFIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE WriteF ( outfile      : FilePtrIO.File;
                   (*. CONST .*) fmtStr : ARRAY OF CHAR;
                   items        : (*. ARGLIST .*) ARRAY OF CARD8T );

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WholeScanSet, FormatIO.ValidWholeFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;

END Card8BaseFIO.

