(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE CardinalArrayFIO (*[ProtoArrayIO] FOR CARDINAL*);

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation as CardinalIO
    RS    2015 05 12  uses new FormatIO module
    RS    2016 07 04  uses new R10 module suite

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYao, CAPACITYc;

(* RS utilities *)
IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT WholeStrs, Perror;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, ISO646, ARRAYOFCHAR, CharArray;

(* R10 Stuff with mods *)
IMPORT FilePtrIO, CardinalBaseFIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  WholeType = CARDINAL;

CONST
  StrToWholeType = WholeStrs.StrToCard;
  WholeTypeToDecStr = WholeStrs.CardToDecStr;
  WholeTypeToHexStr = WholeStrs.CardToHexStr;
  WholeTypeToBinStr = WholeStrs.CardToBinStr;


PROCEDURE (*[READ]*) ReadArray
  ( infile : FilePtrIO.File; VAR array : ARRAY OF CARDINAL );

VAR
  count, last : CARDINAL;
  status : IOStatusT;

BEGIN
  count := 0;
  last := CAPACITYc (array) - 1;
  WHILE count <= last
    DO
      CardinalBaseFIO.Read (infile, array [count] );
      status := FilePtrIO.statusOf (infile);
      count := count + 1;
      IF NOT status.failed THEN RETURN END
    END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : FilePtrIO.File; (*CONST*) array : ARRAY OF CARDINAL );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYc (array) - 1;
  (* open brace for the array *)
  CharStrFIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <= last DO
    CardinalBaseFIO.Write (outfile, array [count]);
    IF count # last THEN
       CharStrFIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  (* close brace for the array *)
  CharStrFIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF WholeType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  res : IOStatusT;
  nc : WholeType;

BEGIN
  numItems := CAPACITYc (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr);
  END;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    nc := items[count];
    temp[0] := ISO646.NUL;
    (* convert to the appropriate string *)
    IF fsr.numeralFormat = FormatIO.Hex THEN
      IF nc # 0 THEN WholeTypeToHexStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSIF fsr.numeralFormat = FormatIO.Bin THEN
      IF nc # 0 THEN WholeTypeToBinStr (nc, temp) ELSE CharArray.CharToAOC (temp, "0") END;
    ELSE
      WholeTypeToDecStr (nc, temp) (* convert the Cardinal to a string *)
    END; (*IF*)
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFIO.WriteStr (outfile, temp);

       (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # numItems) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END;

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN CharBaseFIO.Write (outfile, ISO646.EOL) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : FilePtrIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0*) ARRAY OF ARRAY OF WholeType );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, numItems, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   numItems := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.WholeScanSet, FormatIO.ValidWholeFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <= numItems DO

    (* go write the array *)
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND ((count # numItems) OR NOT lastArray) THEN
      CharStrFIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END;
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  numItems ) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END
 END;

  (* and if an escaped char is in the fsr, write it *)
      IF lastArray THEN CharStrFIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *) END
END WriteFArray;

TYPE
  TargetType = CARDINAL;
  Ary1 = ARRAY [0..0] OF TargetType;
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;

VAR
  lastArray : BOOLEAN; (* this is a kludge anyway so use a global variable to control the following *)


PROCEDURE Write1FArray
  ( outfile : FilePtrIO.File; fmtStr : ARRAY OF CHAR; value1 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
    WriteFArray (outfile, fmtStr,value1 )
END Write1FArray;


PROCEDURE Write2FArray
  ( outfile : FilePtrIO.File; fmtStr : ARRAY OF CHAR; value1, value2 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value2 )
END Write2FArray;


PROCEDURE Write3FArray
  ( outfile : FilePtrIO.File; fmtStr : ARRAY OF CHAR; value1, value2, value3 : (*. ARGLIST .*) ARRAY OF TargetType );

BEGIN
  lastArray := FALSE;
  WriteFArray (outfile, fmtStr, value1 );
  WriteFArray (outfile, fmtStr, value2 );
  lastArray := TRUE;
  WriteFArray (outfile, fmtStr, value3 )
END Write3FArray;

BEGIN (* module *)
  lastArray := TRUE;

END CardinalArrayFIO.