(* (C) 2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE CharStrFIO (*[ProtoStrIO] FOR CHAR*);

(* IO Extension Library for String Types of CHAR *)

IMPORT SYSTEM;

(* R10 stuff *)
IMPORT FilePtrIO, TokenFIO, CharArrayFIO, CharBaseFIO, Newline, ISO646;

(* RS Stuff *)
IMPORT FormatIO;
IMPORT LongWholeStrs, CharArray;
FROM Transition IMPORT CAPACITYo;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;
  FormatTextStr = FormatIO.FormatTextStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;


PROCEDURE (*[READ]*) ReadStr
  ( infile : FilePtrIO.File; VAR str : ARRAY OF CHAR );

BEGIN
  TokenFIOIO.ReadToken ( infile, str )
END ReadStr;


PROCEDURE (*[WRITE]*) WriteStr
  ( outfile : FilePtrIO.File; (*CONST*) str : ARRAY OF CHAR );

VAR
  max, count : CARDINAL;

BEGIN
  max := LENGTH (str);
  count := 0;
  WHILE count < max DO
    IF str [count] = ISO646.BACKSLASH THEN
      count := count + 1;
      IF count > max THEN RETURN END;
      CASE str [count] OF
        ISO646.NUL : (* got to array of char terminator *)
          CharBaseFIO.Write (outfile, ISO646.BACKSLASH ); (* send out the backslash *)
          RETURN (* and do nothing else *)
        | "n" : 
             CASE Newline.mode () OF
                 Newline.LF :
                   CharBaseFIO.Write (outfile, ISO646.LF);
               | Newline.CR : 
                   CharBaseFIO.Write (outfile, ISO646.CR);
               | Newline.CRLF : 
                   CharBaseFIO.Write (outfile, ISO646.CR);
                   CharBaseFIO.Write (outfile, ISO646.LF)
              END (* case NewLine *)
        | "t" : CharBaseFIO.Write (outfile, ISO646.HT);
        ELSE (* emit all others after a slash including another backslash *)
          CharBaseFIO.Write (outfile, str [count])
        END (* case str [count] *)
     ELSE (* not a slash so emit the thing *)
       IF (str [count] # ISO646.NUL) THEN
         CharBaseFIO.Write (outfile, str [count])
       ELSE (* pack it in if nul *)
         count := max
       END  (* if *)
     END; (* if *)
     count := count + 1
  END (* while *)
END WriteStr;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items : ARRAY OF CHAR );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, numItems, colCount : CARDINAL;
  lCard : LONGCARD;

BEGIN
  numItems := LENGTH ( items )-1;
  colCount := 0;

  (* now loop through the items *)
  count := 0;
  WHILE count <= numItems DO
    IF fsr.numeralFormat = FormatIO.Hex THEN
      lCard := ORD ( items[count] );
      LongWholeStrs. CardToHexStr ( lCard, temp);
    ELSE
      temp[0] := items[count]; temp[1] := ISO646.NUL;
    END; (* if *)
       (* format that char using the fsr already parsed *)
    temp :=  FormatTextStr (temp, fsr);

         (* write out the formatted string *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # numItems) THEN WriteStr ( outfile, "\n" ) END
    END (* if *);
      count := count + 1
  END(* while *);


  (* and if an escaped char is in the fsr, write it *)
  WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFStr
  ( outfile : FilePtrIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) strings : (*ARGLIST >0*) ARRAY OF ARRAY OF CHAR );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, last, colCount : CARDINAL;

BEGIN (* WriteFArray *)
  last := CAPACITYo ( strings ) - 1;
  colCount := 0;

  (* get the format string scanned to a format string record *)
    ScanStrForFormat (fmtStr, FormatIO.CharScanSet, FormatIO.ValidCharFmtChars, fsrForArglist);

  (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.arrayItemSeperator := "";
  fsrForArraysOfItems.terminalEscStr := "";

  count := 0;
  WHILE count <= last DO

  (* go write the array *)
  WriteFWithfsr ( outfile, fsrForArraysOfItems, strings [count]  );
  IF (count # last) THEN
    WriteStr (outfile, fsrForArglist.arrayItemSeperator)
  END;

         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count # last ) THEN WriteStr ( outfile, "\n" ) END
    END
  END (* while *);
  (* and if an escaped char is in the fsr, write it *)
  WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)

END WriteFStr;

END CharStrFIO.