IMPLEMENTATION MODULE LongRealBaseFIO;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 15  uses new FormatIO module
    RS    2015 05 28  Stomp bugs and refined
    RS      2016 07 26  Now FIO

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYlr;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT OctetStrs, LongRealStrs;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT FilePtrIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  RealType = LONGREAL;

CONST
  StrToRealType = LongRealStrs.StrToReal;
  RealTypeToStr = LongRealStrs.RealToStr;
  RealTypeToScientific = LongRealStrs.RealToScientific;
  RealTypeToFloat = LongRealStrs.RealToFloat;
  RealTypeToEng = LongRealStrs.RealToEng;
  RealTypeToFixed = LongRealStrs.RealToFixed;
  defDecPlaces = LongRealStrs.rMaxSigFigs - 1;


PROCEDURE SetResult (file: FilePtrIO.File; res : IOStatusT);

VAR
  ok : BOOLEAN;

BEGIN
  CASE res.code
    OF
      IOStatus.Success    : FilePtrIO.SetStatus (file, IOStatusT { FALSE, IOStatus.Success} , ok) |
      IOStatus.OutOfRange  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.OutOfRange} , ok ) |
      IOStatus.WrongFormat  : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.WrongFormat} , ok )|
      IOStatus.StrEmpty   : FilePtrIO.SetStatus (file, IOStatusT { TRUE, IOStatus.UnexpectedEndOfLine} , ok )
    END; (* case *)
  
  IF NOT ok THEN
      Perror.prerr ( "Invalid SetResult parameters passed\n" )
  END (* if *)
END SetResult;


PROCEDURE Read ( infile : FilePtrIO.File; VAR n : RealType );
  (* Skips leading spaces, and removes any remaining characters from file that form part of an unsigned whole number.  The value of this number is assigned to card.  The read result is set to the value allRight, outOfRange, wrongFormat, endOfLine, or endOfInput.  *)

VAR
  temp  : FormatIO.Str;
  res : IOStatusT;

BEGIN (* Read *)
  TokenFIO.ReadRealToken (infile, temp);        (* get the RealType as a string *)
  StrToRealType (temp, n, res);       (* convert the string to a RealType *)
  SetResult (infile, res)
END Read;



PROCEDURE Write ( outfile : FilePtrIO.File; n : RealType );
  (* Writes the value of card to cid in text form, in a field of the given
minimum width. *)

VAR
  temp : FormatIO.Str;

BEGIN (* Write *)
  RealTypeToScientific (n, defDecPlaces, temp);            (* convert the RealType to a float string (default) *)
  CharStrFIO.WriteStr (outfile, temp);
END Write;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF RealType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  theNum : RealType;


PROCEDURE NumberToNumeralString (theNum: RealType); (*internal *)

BEGIN       (* convert to the appropriate string *)
   IF fsr.numeralFormat = FormatIO.Hex
   THEN
     OctetStrs.OctsToHexStr (theNum, temp);
   ELSIF fsr.numeralFormat = FormatIO.Bin THEN
     OctetStrs.OctsToBinStr (theNum, temp);
   ELSIF fsr.realFormat = FormatIO.Scientific THEN
    RealTypeToScientific (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Float THEN
    RealTypeToFloat (theNum, fsr.numDecPlaces, fsr.numIntDigitsForFloat, temp);
   ELSIF fsr.realFormat = FormatIO.Fixed THEN
    RealTypeToFixed (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Currency THEN
    RealTypeToFixed (theNum, -2, temp);
   ELSIF fsr.realFormat = FormatIO.Eng THEN
    RealTypeToEng (theNum, fsr.numDecPlaces, temp);
   END(*IF*)
END NumberToNumeralString;

BEGIN (*WriteFWithfsr*)
  IF (fsr.numDecPlaces = 0) AND (fsr.realFormat # FormatIO.Fixed) THEN (* set up default *)
    fsr.numDecPlaces  := LongRealStrs.rMaxSigFigs - 2
  END; (* default *)
  last := CAPACITYlr (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    theNum := items[count];
    NumberToNumeralString (theNum);
    (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
    CharStrFIO.WriteStr (outfile, temp);

        (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE WriteF ( outfile      : FilePtrIO.File;
                   (*. CONST .*) fmtStr : ARRAY OF CHAR;
                   items        : (*. ARGLIST .*) ARRAY OF LONGREAL);

VAR
  fsr : FormatScanRec;

BEGIN (* WriteF *)
   (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.RealScanSet, FormatIO.ValidRealFmtChars, fsr);
  WriteFWithfsr ( outfile, fsr , items  )

END WriteF;


BEGIN (* module *)

END LongRealBaseFIO.
