IMPLEMENTATION MODULE  RealArrayFIO;

  (*
    Implementation and Revisions:
    ============================

    Author  Date        Description of change
    ------  ----        ---------------------
    RS      03/27/2014  First implementation
    RS    2015 05 15  uses new FormatIO module
    RS    2015 05 28  Stomp bugs and refined
    RS      2016 07 26  Now FIO

  *)
IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYao, CAPACITYr;

IMPORT IOStatus; FROM IOStatus IMPORT IOStatusT;
IMPORT TokenFIO, FormatIO;
IMPORT OctetStrs, RealStrs, RealBaseFIO;

(* R10 Stuff *)
IMPORT CharStrFIO, CharBaseFIO, ISO646, Perror;

(* R10 Stuff with mods *)
IMPORT FilePtrIO;

CONST
  ScanStrForFormat = FormatIO.ScanStrForFormat;
  FormatNumStr = FormatIO.FormatNumStr;

TYPE
  FormatScanRec = FormatIO.FormatScanRec;

TYPE
  RealType = REAL;

CONST
  StrToRealType = RealStrs.StrToReal;
  RealTypeToStr = RealStrs.RealToStr;
  RealTypeToScientific = RealStrs.RealToScientific;
  RealTypeToFloat = RealStrs.RealToFloat;
  RealTypeToEng = RealStrs.RealToEng;
  RealTypeToFixed = RealStrs.RealToFixed;


PROCEDURE (*[READ]*) ReadArray
  ( infile : FilePtrIO.File; VAR array : ARRAY OF REAL );

VAR
  count, last : CARDINAL;

BEGIN
  count := 0;
  last := CAPACITYr (array) - 1;
  WHILE count <= last DO
    RealBaseFIO.Read (infile, array [count]);
    count := count + 1
  END
END ReadArray;


PROCEDURE (*[WRITE]*) WriteArray
  ( outfile : FilePtrIO.File; (*CONST*) array : ARRAY OF REAL );

VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYr (array) - 1;
  CharStrFIO.WriteStr (outfile, "{ ");
  count := 0;
  WHILE count <= last DO
    RealBaseFIO.Write (outfile, array [count]);
    IF count < last THEN
        CharStrFIO.WriteStr (outfile, ", ")
    END; (* IF *)
    count := count + 1
  END; (* while *)
  CharStrFIO.WriteStr (outfile, " }")
END WriteArray;


PROCEDURE WriteFWithfsr ( outfile : FilePtrIO.File; fsr : FormatScanRec; items :  ARRAY OF RealType );
(* come here to write out the items after the fmtStr results have been scanned to an fsr *)
VAR
  temp : FormatIO.Str;
  count, last, colCount : CARDINAL;
  res : IOStatusT;
  theNum : RealType;


PROCEDURE NumberToNumeralString (theNum: RealType); (*internal *)

BEGIN       (* convert to the appropriate string *)
   IF fsr.numeralFormat = FormatIO.Hex
   THEN
     OctetStrs.OctsToHexStr (theNum, temp);
   ELSIF fsr.numeralFormat = FormatIO.Bin THEN
     OctetStrs.OctsToBinStr (theNum, temp);
   ELSIF fsr.realFormat = FormatIO.Scientific THEN
    RealTypeToScientific (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Float THEN
    RealTypeToFloat (theNum, fsr.numDecPlaces, fsr.numIntDigitsForFloat, temp);
   ELSIF fsr.realFormat = FormatIO.Fixed THEN
    RealTypeToFixed (theNum, fsr.numDecPlaces, temp);
   ELSIF fsr.realFormat = FormatIO.Eng THEN
    RealTypeToEng (theNum, fsr.numDecPlaces, temp);
   END(*IF*)
END NumberToNumeralString;

BEGIN (*WriteFWithfsr*)
  IF (fsr.numDecPlaces = 0) AND (fsr.realFormat # FormatIO.Fixed) THEN (* set up default *)
      fsr.numDecPlaces  := RealStrs.rMaxSigFigs
  END; (* default *)
  last := CAPACITYr (items) - 1;
  colCount := 0;

    (* open an array enclosure if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayOpenStr)
  END; (* if *) 

  (* now loop through the items *)
  count := 0;
  WHILE count <= last DO
    theNum := items[count];
    NumberToNumeralString (theNum);
   (* format that converted string using the fsr already parsed *)
    temp :=  FormatNumStr (temp, fsr);

    (* write out the formatted string *)
  CharStrFIO.WriteStr (outfile, temp);

        (* and the separator *)
    IF ((fsr.isArrayFormat) OR NOT fsr.isConcat) AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsr.arrayItemSeperator)
    END; (* if *)

       (* deal with columns *)
    IF fsr.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsr.numCols;
      IF (colCount = 0) AND (count # last) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END
    END;
    count := count + 1
  END(* while *);

  (* close the array if needed *)
  IF fsr.isArrayFormat THEN
    CharStrFIO.WriteStr (outfile, fsr.arrayCloseStr)
  END; (* if *)
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsr.terminalEscStr) (* if any *)

END WriteFWithfsr;


PROCEDURE (*[WRITEF]*) WriteFArray
  ( outfile : FilePtrIO.File;
    (*CONST*) fmtStr : ARRAY OF CHAR;
    (*CONST*) arrays : (*ARGLIST >0 *) ARRAY OF ARRAY OF REAL );

VAR
  fsrForArglist, fsrForArraysOfItems : FormatScanRec;
  count, last, colCount : CARDINAL;

BEGIN (* WriteFArray *)
   last := CAPACITYao (arrays) - 1;
   colCount := 0;

  (* get the format string scanned to a format string record *)
  ScanStrForFormat (fmtStr, FormatIO.RealScanSet, FormatIO.ValidRealFmtChars, fsrForArglist);

    (* set up the fsr for the next layer in *)
  fsrForArraysOfItems:= fsrForArglist;
  fsrForArraysOfItems.isColumnFormat := FALSE;
  fsrForArraysOfItems.terminalEscStr := "";
  fsrForArraysOfItems.isArrayFormat := TRUE;

  count := 0;
  WHILE count <=  last DO
    (* go write the array *)   
    WriteFWithfsr ( outfile, fsrForArraysOfItems , arrays [count]  );
    IF fsrForArglist.isArrayFormat AND (count # last) THEN
      CharStrFIO.WriteStr (outfile, fsrForArglist.arrayItemSeperator)
    END; (* if *)
         (* deal with columns *)
    IF fsrForArglist.isColumnFormat THEN
      colCount := (colCount + 1) MOD fsrForArglist.numCols;
      IF (colCount = 0) AND (count #  last ) THEN CharStrFIO.WriteStr ( outfile, "\n" ) END   
    END; (* if *)
    count := count + 1
  END; (* while *)
   
  (* and if an escaped char is in the fsr, write it *)
  CharStrFIO.WriteStr (outfile, fsrForArglist.terminalEscStr) (* if any *)
END WriteFArray;

BEGIN (* module *)

END RealArrayFIO.
