IMPLEMENTATION MODULE PtrTable;
IMPORT ISO646;
IMPORT UNSAFE;
FROM Transition IMPORT INTEGER32;
FROM Storage IMPORT ALLOCATE, DEALLOCATE;

CONST
  errNum = MIN(INTEGER32);

TYPE
  TableRange = CARDINAL[0..tableSize - 1];
  PtrTable = POINTER TO TableData;
  TableData = RECORD
    count : TableRange;
    pairs : ARRAY TableRange OF Node;
  END;


PROCEDURE SwopNodes (table : PtrTable; VAR first, second : TableRange);

VAR
  temp : Node;

BEGIN
  temp := table^.pairs [first];
  table^.pairs [first] := table^.pairs [second];
  table^.pairs [second] := temp
END SwopNodes;


PROCEDURE New ( (*NEW *)VAR table : PtrTable);

BEGIN
  NEW (table);
  table^.count := 0
END New;


PROCEDURE Release (VAR table : PtrTable);

BEGIN
  DISPOSE (table)
END Release;


PROCEDURE EnTable (table : PtrTable; itemNum : INTEGER32; ptr : (*CAST*)UNSAFE.ADDRESS);

BEGIN
  table^.count := table^.count + 1; (* note that the zeroth entry is not used *)
  table^.pairs [table^.count].itemNum := itemNum;
  table^.pairs [table^.count].pointer := ptr;
END EnTable;


PROCEDURE Remove (table : PtrTable; nodeNum : TableRange);
(* pre: the node has been correctly found first *)

BEGIN
  SwopNodes (table, nodeNum, table^.count);
  table^.count := table^.count - 1
END Remove;


PROCEDURE DeTableByItemNum (table : PtrTable; itemNum : INTEGER32) : BOOLEAN;
(* if not there returns FALSE *)
VAR
  res : BOOLEAN;
  nodeNum : TableRange;

BEGIN
  res := FindIndexNodeNum (table, itemNum, nodeNum);
  IF res THEN
    Remove (table, nodeNum)
  END;
  RETURN res;
END DeTableByItemNum;


PROCEDURE DeTableByPointer (table : PtrTable; ptr : (*CAST*)UNSAFE.ADDRESS) : BOOLEAN;
(* if not there returns FALSE *)
VAR
  res : BOOLEAN;
  nodeNum : TableRange;

BEGIN
  res := FindPtrNodeNum (table, ptr, nodeNum);
  IF res THEN
    Remove (table, nodeNum)
  END;
  RETURN res;
END DeTableByPointer;


PROCEDURE FindPtrNodeNum (table : PtrTable;  ptr :(* CAST *)UNSAFE.ADDRESS; VAR nodeNum : TableRange ) : BOOLEAN;

VAR
  count : TableRange;
  found : BOOLEAN;

BEGIN
  count := 1;
  WHILE count <= table^.count DO
    IF table^.pairs [count].pointer = ptr THEN
      nodeNum := count;
      RETURN TRUE
    ELSE
       count := count + 1
    END;
  END; (* while *)
       (* if get here did not find so *)
  RETURN FALSE
END FindPtrNodeNum;


PROCEDURE FindIndexNodeNum (table : PtrTable; item : INTEGER32; VAR nodeNum : TableRange) : BOOLEAN;
(* if not found returns MIN (INTEGER32 *)
VAR
  count : TableRange;
  found : BOOLEAN;

BEGIN
  count := 1;
  WHILE count <= table^.count DO
    IF table^.pairs [count].itemNum = item THEN
      nodeNum := count;
      RETURN TRUE
    ELSE
       count := count + 1
    END;
  END; (* while *)
       (* if get here did not find so *)
  RETURN FALSE
END FindIndexNodeNum;


PROCEDURE GetItemNum (table : PtrTable;  ptr :(* CAST *)UNSAFE.ADDRESS) : INTEGER32;
(* if not found returns MIN (INTEGER32 *)
VAR
  nodeNum : TableRange;
  found : BOOLEAN;

BEGIN
  IF FindPtrNodeNum (table, ptr, nodeNum) THEN
    RETURN table^.pairs [nodeNum].itemNum
  ELSE
    RETURN errNum
  END;
END GetItemNum;

PROCEDURE GetHighestIndexedItemNum (table : PtrTable) : INTEGER32;
  (* some clients may process items or remove them one at a time *)
BEGIN
  RETURN table^.pairs [table^.count].itemNum 
END GetHighestIndexedItemNum;

PROCEDURE Count (table : PtrTable) : CARDINAL;
  (* Returns the current number of table entries. *)

BEGIN
  RETURN table^.count
END Count;


PROCEDURE GetPtr (table : PtrTable; itemNum : INTEGER32) : UNSAFE.ADDRESS;
(* if not found returns NIL *)
VAR
  nodeNum : TableRange;
  found : BOOLEAN;

BEGIN
  IF FindIndexNodeNum (table, itemNum, nodeNum) THEN
    RETURN table^.pairs [nodeNum].pointer
  ELSE
    RETURN NIL
  END (* while *)

END GetPtr;


PROCEDURE ProcessTable (table : PtrTable; Process : ProcessProc );
     (* process all pairs with the given procedure in reverse order. Can be more efficient *)
VAR
  count : TableRange;

BEGIN
  count := table^.count;
  WHILE count > 0 DO (* note that the zeroth item is not used *)
       Process (table, table^.pairs[count]);
       count := count - 1
    END;
END ProcessTable;


PROCEDURE isEmpty (table : PtrTable) : BOOLEAN;
  (* Returns TRUE if the table is empty, false otherwise *)

BEGIN
  RETURN (table^.count = 0)
END isEmpty;


PROCEDURE isFull (table : PtrTable) : BOOLEAN;
  (* Returns TRUE if the table is full, false otherwise *)

BEGIN
  RETURN (table^.count = tableSize - 1)
END isFull;


(* PROCEDURE DumpTable (table : PtrTable);  (* debug *)

VAR
  count : TableRange;
  found : BOOLEAN;

BEGIN
  count := 1;
  WHILE count <= table^.count DO
    sIntegerIO.WriteF (">6", table^.pairs [count].itemNum);
    AddressIO.WriteF (StdIO.out, "x>25", table^.pairs [count].pointer);
    CharIO.Write ( StdIO.out, ISO646.EOL );
    count := count + 1
  END; (* while *)
  sCharIO.WriteStr ("count = ");sCardinalIO.WriteF (">6", VAL (CARDINAL, table^.count) );
  CharIO.Write ( StdIO.out, ISO646.EOL );

END DumpTable;

*)
(* ************* *)
VAR
  testTable : PtrTable;
  a,b,c,d : POINTER TO CARDINAL;
  res : BOOLEAN;

(* PROCEDURE TestProcess (table : PtrTable; node : Node);

VAR res : BOOLEAN;

BEGIN
   sIntegerIO.WriteF (">6", node.itemNum);
   res := DeTableByIndex (table, node.itemNum);
END TestProcess;
*)
BEGIN
  (*
NEW (a); NEW (b); NEW (c); NEW (d);
  New (testTable);
  EnTable (testTable, 12, a);
  EnTable (testTable, 34, b);
  EnTable (testTable, 56, c);
  EnTable (testTable, 78, d);
  DumpTable (testTable);
  res := DeTableByIndex (testTable, 34);

  DumpTable (testTable);
  EnTable (testTable, 34, d);
  EnTable (testTable, 99, a);
   DumpTable (testTable);


  res := DeTableByPointer (testTable, c);

  sBooleanIO.Write (res);

 DumpTable (testTable);
 sIntegerIO.WriteF (">6", GetItemNum (testTable,c) );
AddressIO.WriteF (StdIO.out, "x>25", GetPtr (testTable, 102) );

  sCharIO.WriteStr ("here\n");
  ProcessTable (testTable, TestProcess);
  sCharIO.WriteStr ("here2\n");
DumpTable (testTable);
*)
END PtrTable. 