(* (C) 2009-2016 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE TextIO;

(* Driver for Line Oriented Text IO *)

(* This module interprets the contents of files as a sequence of lines of text
   delimited by newline. The module is implemented on top of module FileIO. *)

FROM Transition IMPORT CAPACITYao;

IMPORT FileIO, ISO646, CharIO;

(* Read a line of text *)


PROCEDURE ReadLine ( infile : FileIO.File; VAR str : ARRAY OF CHAR );
(* Reads a line of text from <file> and passes it back in <str>. A runtime
   error is raised if the line exceeds the capacity of <str>. *)
VAR
  count : CARDINAL;
  ch : CHAR;

BEGIN
   count := 0;
   REPEAT
    FileIO.ReadOctet (infile, ch);
    IF ch # ISO646.EOL THEN
      str [count] := ch;
      count := count + 1
    END (* if *);
  UNTIL ch = ISO646.EOL
END ReadLine;

(* Skip a line of text *)


PROCEDURE SkipLine ( infile : FileIO.File );
(* Read and ignore all characters up to and including the first newline. *)
VAR
  ch : CHAR;

BEGIN
   REPEAT
    FileIO.ReadOctet (infile, ch)
  UNTIL ch = ISO646.EOL
END SkipLine;


(* Write a line of text *)


PROCEDURE WriteLine ( outfile : FileIO.File; str : (*ARGLIST >0*) ARRAY OF ARRAY OF CHAR );
(* Writes a sequence of character strings as a line to <file>. The line is
   terminated by newline. *)
VAR
  count, last : CARDINAL;

BEGIN
  last := CAPACITYao (str) - 1;
  FOR count := 0 TO last DO
    CharIO.WriteStr ( outfile, str[count] )
  END;
  CharIO.WriteStr ( outfile, "\n" )
END WriteLine;


END TextIO.