IMPLEMENTATION MODULE ISO646;

IMPORT CARD8; FROM CARD8 IMPORT CARD8T;
TYPE
  CharSet = SET OF CHAR;
CONST
  binChars = CharSet {"0", "1"};
  decChars = CharSet {"0".."9"};
  octalChars = CharSet {"0".."7"};
  lowerAToFChars = CharSet{"a".."f"};
  upperAToFChars = CharSet{"A".."F"};
  aToFChars = lowerAToFChars + upperAToFChars;
  lowerhexChars = decChars + lowerAToFChars;
  upperhexChars = decChars + upperAToFChars;
  hexChars = decChars + lowerhexChars + upperhexChars;
  cntrlChars = CharSet {CHR(0)..CHR(31), CHR (127)};
  printableChars = CharSet {CHR(32)..CHR(126)};
  upperCaseChars = CharSet {"A".."Z"};
  lowerCaseChars = CharSet {"a".."z"};
  alphaChars = upperCaseChars + lowerCaseChars;
  DTMFChars = decChars + CharSet {"*", "#" };
  alphaNumericChars = alphaChars + decChars;
  alphaNumOrLowlineChars = alphaNumericChars + CharSet{LOWLINE};
  signChars = CharSet {"+", "-"};
  decDigitOrSignChars = decChars + signChars;
  quotingChars = CharSet {DOUBLE_QUOTE, SINGLE_QUOTE};
  whiteSpaceChar = CharSet{SPACE};
  tabChar = CharSet{HT};
  tabAndSpaceChars = whiteSpaceChar + tabChar;
  upperLowerDiff = ORD("a") - ORD("A");
  punctuation = CharSet {"!", '"', "'", "`", ",", "-", "—", ":", ';', "(", ")", ".", "{", "}", "[", "]"  };
  nonPunctuationSymbols = CharSet{"#", "$", "%", "&", "*", "+", "/", "<", "", "=", ">", "?", "@", "\", "^", "_", "~", "|"   };
  symbols = punctuation + nonPunctuationSymbols;

(* define a type to hold a collating array and a constant that holds a
  collating order for the ISO 646 characters (lower 128) in lexical order. 
     The lexical order of characters is as follows:
   space
   symbols other than numerals, letters, and punctuation marks (in their ASCII order )
   numerals (0 through 9)
   letters (A through Z where uppercase and lower case are not distinguished 
   punctuation marks (. , ; : ( ) { } [ ] ' " ! ? )     *)

TYPE
  CARD7 = [0..127];
  OrderArray = ARRAY CARD7 OF CARD7;
  
VAR
  lexOrder : OrderArray;

CONST
  LexA = OrderArray (* same ordinal order for upper and lower case *)
    { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 
16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 
33, 87, 88, 34, 35, 36, 37, 89, 90, 91, 38, 39, 92, 93, 94, 40, 
51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 95, 96, 41, 42, 43, 44, 
45, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 
76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 97, 46, 98, 47, 48, 
99, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 
76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 100, 49, 101, 50, 32 };

  LexAa = OrderArray (* upper case letter orders before lower case letter *)
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 
16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 
33, 113, 114, 34, 35, 36, 37, 115, 116, 117, 38, 39, 118, 119, 120, 40, 
51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 121, 122, 41, 42, 43, 44, 
45, 61, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 85, 87, 89, 
91, 93, 95, 97, 99, 101, 103, 105, 107, 109, 111, 123, 46, 124, 47, 48, 
125, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 86, 88, 90, 
92, 94, 96, 98, 100, 102, 104, 106, 108, 110, 112, 126, 49, 127, 50, 32 };

  LexaA = OrderArray (* lower case letter orders before upper case letter *)
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 
16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 
33, 113, 114, 34, 35, 36, 37, 115, 116, 117, 38, 39, 118, 119, 120, 40, 
51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 121, 122, 41, 42, 43, 44, 
45, 62, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 86, 88, 90, 
92, 94, 96, 98, 100, 102, 104, 106, 108, 110, 112, 123, 46, 124, 47, 48, 
125, 61, 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 85, 87, 89, 
91, 93, 95, 97, 99, 101, 103, 105, 107, 109, 111, 126, 49, 127, 50, 32 };

PROCEDURE LexORD (ch : CHAR) : CARDINAL;
(* Returns the ordinal number of the character in the lexical collating order
  rather than the ASCII order as does ORD. Note that upper and lower case
  are not distinguished, so "a" and "A" have the  same ordinal number.*)
BEGIN
  RETURN lexOrder [ORD (ch)]
END LexORD;

PROCEDURE SetLexOrder (ordArray : OrderArray );
BEGIN
  lexOrder := ordArray;
END SetLexOrder;
 
PROCEDURE toUpper( ch : CHAR ) : CHAR;

BEGIN
  IF isLowerCase( ch ) THEN
    RETURN CHR (ORD (ch) - upperLowerDiff);
  ELSE
    RETURN ch
  END
END toUpper;


PROCEDURE toLower( ch : CHAR ) : CHAR;

BEGIN
  IF isUpperCase( ch ) THEN
    RETURN CHR (ORD (ch) + upperLowerDiff);
  ELSE
    RETURN ch
  END
END toLower;


PROCEDURE isUpper (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as an upper case letter *)

BEGIN
  RETURN ch IN upperCaseChars
END isUpper;


PROCEDURE isLower (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a lower case letter *)

BEGIN
  RETURN ch IN lowerCaseChars
END isLower;


PROCEDURE isControl (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch represents a control function *)

BEGIN
  RETURN ch IN cntrlChars
END isControl;


PROCEDURE isNotControl (ch: CHAR): BOOLEAN;

BEGIN (* isNotControl *)
  RETURN NOT isControl(ch)
END isNotControl;

PROCEDURE isPrintable (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch represents a printable character *)
BEGIN 
  RETURN ch IN printableChars
END isPrintable;


PROCEDURE isNotPrintable (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch does not represent a printable character *)
BEGIN 
  RETURN NOT isPrintable(ch)
END isNotPrintable;


PROCEDURE isWhitespaceOrTab( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN tabAndSpaceChars
END isWhitespaceOrTab;


PROCEDURE isNotWhitespaceOrTab( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotWhitespaceOrTab *)
  RETURN NOT (ch IN tabAndSpaceChars)
END isNotWhitespaceOrTab;


PROCEDURE isNonPunctuationSymbol( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN nonPunctuationSymbols
END isNonPunctuationSymbol;


PROCEDURE isNOTNonPunctuationSymbol( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN nonPunctuationSymbols)
END isNOTNonPunctuationSymbol;


PROCEDURE isPunctuation( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN punctuation
END isPunctuation;

PROCEDURE isNotPunctuation( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT ( ch IN punctuation )
END isNotPunctuation;

PROCEDURE isSymbol (ch: CHAR): BOOLEAN;

BEGIN
  RETURN ch IN symbols
END isSymbol;

PROCEDURE isNonSymbol (ch: CHAR): BOOLEAN;

BEGIN
  RETURN NOT ( ch IN symbols )
END isNonSymbol;


PROCEDURE isQuote (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a quoting character *)
BEGIN
  RETURN ch IN quotingChars
END isQuote;

  
PROCEDURE isNotQuote (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a nonquoting character *)
BEGIN
  RETURN NOT (ch IN quotingChars)
END isNotQuote;


PROCEDURE isDigit (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a numeric character *)

BEGIN
  RETURN ch IN decChars
END isDigit;


PROCEDURE isNotDigit (ch: CHAR): BOOLEAN;

BEGIN (* isNotDigit *)
  RETURN NOT (ch IN decChars)
END isNotDigit;


PROCEDURE isSign (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a sign character "+" or "-" *)
BEGIN
  RETURN ch IN signChars
END isSign;


PROCEDURE isNotSign (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is not classified as a sign character "+" or "-" *)
BEGIN
  RETURN NOT (isSign (ch) )
END isNotSign;
  
  
PROCEDURE isDigitOrSign (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a numeric or sign character "+" or "-" *)
BEGIN
  RETURN ch IN decDigitOrSignChars
END isDigitOrSign;


PROCEDURE isNotDigitOrSign (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is not classified as a numeric or sign character "+" or "-" *)
BEGIN
  RETURN NOT (ch IN decDigitOrSignChars)
END isNotDigitOrSign;


PROCEDURE isOctalDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isOctalDigit *)
  RETURN ch IN octalChars
END isOctalDigit;


PROCEDURE isNotOctalDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotOctalDigit *)
  RETURN NOT (ch IN octalChars)
END isNotOctalDigit;


PROCEDURE isHexDigit (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a hexadecimal digit character *)

BEGIN
  RETURN ch IN hexChars
END isHexDigit;


PROCEDURE isNotHexDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotHexDigit *)
  RETURN NOT (ch IN hexChars)
END isNotHexDigit;


PROCEDURE isLowerHexDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isLowerHexDigit *)
  RETURN ch IN lowerhexChars
END isLowerHexDigit;


PROCEDURE isNotLowerHexDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotLowerHexDigit *)
  RETURN NOT (ch IN lowerhexChars)
END isNotLowerHexDigit;


PROCEDURE isUpperHexDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isUpperHexDigit *)
  RETURN ch IN upperhexChars
END isUpperHexDigit;


PROCEDURE isNotUpperHexDigit( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotUpperHexDigit *)
  RETURN NOT (ch IN upperhexChars)
END isNotUpperHexDigit;


PROCEDURE isAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isAtoF *)
  RETURN ch IN aToFChars
END isAtoF;


PROCEDURE isNotAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotAtoF *)
  RETURN NOT (ch IN aToFChars)
END isNotAtoF;


PROCEDURE isLowerAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isLowerAtoF *)
  RETURN ch IN lowerAToFChars
END isLowerAtoF;


PROCEDURE isNotLowerAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotLowerAtoF *)
  RETURN NOT (ch IN lowerAToFChars)
END isNotLowerAtoF;


PROCEDURE isUpperAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isUpperAtoF *)
  RETURN ch IN upperAToFChars
END isUpperAtoF;


PROCEDURE isNotUpperAtoF( ch : CHAR ) : BOOLEAN;

BEGIN (* isNotUpperAtoF *)
  RETURN NOT (ch IN upperAToFChars)
END isNotUpperAtoF;


PROCEDURE isLowerCase( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN lowerCaseChars
END isLowerCase;


PROCEDURE isNotLowerCase( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN lowerCaseChars)
END isNotLowerCase;


PROCEDURE isUpperCase( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN upperCaseChars
END isUpperCase;


PROCEDURE isNotUpperCase( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN upperCaseChars)
END isNotUpperCase;


PROCEDURE isLetter (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a letter *)

BEGIN
  RETURN ch IN alphaChars
END isLetter;


PROCEDURE isNotLetter( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN alphaChars)
END isNotLetter;


PROCEDURE isDTMF( ch : CHAR ) : BOOLEAN;
  (* returns TRUE if the character is is a dual-tone frequency modulation key
  i.e. is  on a telephone keypad *)

BEGIN
  RETURN (ch IN DTMFChars)
END isDTMF;


PROCEDURE isNotDTMF( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN DTMFChars)
END isNotDTMF;


PROCEDURE isAlphaNumeric( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN alphaNumericChars
END isAlphaNumeric;


PROCEDURE isNotAlphaNumeric( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN alphaNumericChars)
END isNotAlphaNumeric;


PROCEDURE isAlphaNumOrLowline( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN ch IN alphaNumOrLowlineChars
END isAlphaNumOrLowline;


PROCEDURE isNotAlphaNumOrLowline( ch : CHAR ) : BOOLEAN;

BEGIN
  RETURN NOT (ch IN alphaNumOrLowlineChars)
END isNotAlphaNumOrLowline;


PROCEDURE isBinaryDigit (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is classified as a binary digit character *)

BEGIN
  RETURN ch IN binChars
END isBinaryDigit;


PROCEDURE isNotBinaryDigit (ch: CHAR): BOOLEAN;
  (* Returns TRUE if and only if ch is not classified as a binary digit character *)

BEGIN
  RETURN NOT (ch IN binChars)
END isNotBinaryDigit;

BEGIN (* main *)
  lexOrder := LexA;  (* change this line if you want one of the other two orders *)
END ISO646.

(* =========================== *)
(* Here is the code usedd to generate the Lex orders *)

(* 
PROCEDURE GenerateLexOrder;
TYPE
  CharOrder = ARRAY [0..127] OF CARDINAL;
VAR 
  count, ulDiff, seqLexNum : CARDINAL;
  lex : CharOrder;
BEGIN
  seqLexNum := 0;
  ulDiff := ORD ("a") - ORD ("A");
  count := 0;
  WHILE count <= 127 DO 
   lex [count] := 0 ;   
    count := count + 1;
  END; (* While *)

(* cntrl *)
  count := 0;
  WHILE count < 128 DO 
    IF ISO646.isControl ( CHR (count) ) THEN
      lex [count] := seqLexNum;    
      seqLexNum := seqLexNum + 1
    END; (* if *)
    count := count + 1
 END; (* While *)

(* space *)
  lex [32] := seqLexNum;
  seqLexNum := seqLexNum + 1;

 (* non punct symbols *)
  count := 32;
  WHILE count < 127 DO 
    IF ISO646.isNonPunctuationSymbol ( CHR (count) ) THEN
      lex [count] := seqLexNum;    
      seqLexNum := seqLexNum + 1
    END; (* if *)
    count := count + 1
  END; (* While *)

(* numerals *) 

  count := 33;
  WHILE count < 127 DO 
    IF ISO646.isDigit ( CHR (count) ) THEN
      lex [count] := seqLexNum;    
      seqLexNum := seqLexNum + 1
    END; (* if *)
    count := count + 1
  END; (* While *)


(* letters *)
  count := 33;
  WHILE count < 127 DO 
    IF ISO646.isUpper ( CHR (count) ) THEN
      lex [count] := seqLexNum;          (* seqLexNum if a<=A,  +1 if a>A *)
      lex [count + ulDiff] := seqLexNum ; (* seqLexNum if a>=A,  +1 if a<A *)
      seqLexNum := seqLexNum + 1;        (* +1 if a=A; + 2 otherwise *)
    END; (* if *)
    count := count + 1;
  END; (* While *)


 (*  punct symbols *)
  count := 33;
  WHILE count < 127 DO 
    IF ISO646.isPunctuation ( CHR (count) ) THEN
      lex [count] := seqLexNum;    
      seqLexNum := seqLexNum + 1
    END; (* if *)
    count := count + 1
  END; (* While *)


  count := 0;
  WHILE count <= 127 DO 
    sCardinalIO.Write (lex [count]);
    IF count <> 127 THEN sCharIO.WriteStr (", ") END;
    count := count + 1;
    IF (count MOD 16) = 0 THEN sCharIO.WriteStr ("\n") END;
    END; (* While *)
    
  sCharIO.WriteStr ("\n");
  
END GenerateLexOrder;
*)