(* (C) 2009-2013 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE LONGCOMPLEX (*[ProtoComplex]*);

(* Single Precision LONGCOMPLEX Numbers *)


PROCEDURE CMPLX (real, imag : LONGREAL) : LONGCOMPLEX;
(* constructs a complex from an ordered pair *)

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := real;
  temp.im := imag;
  RETURN temp;
END CMPLX;

(* Bindings required by blueprint ProtoComplex *)

(* Monadic arithmetic operations *)


PROCEDURE (*[NEG]*) neg ( z : LONGCOMPLEX ) : LONGCOMPLEX;
(* Returns the sign reversed value of LONGCOMPLEX value z.
   This function is bound to the unary minus operator for type LONGCOMPLEX. *)

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := -z.re;
  temp.im := -z.im;
  RETURN temp;
END neg;


(* Dyadic arithmetic operations *)


PROCEDURE (*[+]*) add ( z1, z2 : LONGCOMPLEX ) : LONGCOMPLEX;
(* Adds LONGCOMPLEX values z1 and z2 and returns the result.
   This function is bound to the + operator for type LONGCOMPLEX. *)

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := z1.re + z2.re;
  temp.im := z1.im + z2.im;
  RETURN temp;
END add;


PROCEDURE (*[-]*) sub ( z1, z2 : LONGCOMPLEX ) : LONGCOMPLEX;
(* Subtracts LONGCOMPLEX value z2 from z1 and returns the result.
   This function is bound to the - operator for type LONGCOMPLEX. *)

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := z1.re - z2.re;
  temp.im := z1.im - z2.im;
  RETURN temp;
END sub;


PROCEDURE (*[*]*) multiply ( z1, z2 : LONGCOMPLEX ) : LONGCOMPLEX;
(* Multiplies LONGCOMPLEX values z1 and z2 and returns the result.
   This function it bound to the * operator for type LONGCOMPLEX. *)

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := z1.re * z2.re - z1.im * z2.im;
  temp.im := z1.re * z2.im + z1.im * z2.re;
  RETURN temp;
END multiply;


PROCEDURE conj (z : LONGCOMPLEX ) : LONGCOMPLEX;

VAR
  temp : LONGCOMPLEX;

BEGIN
  temp.re := z.re;
  temp.im := -z.im;
  RETURN temp
END conj;


PROCEDURE (*[/]*) divide ( z1, z2 : LONGCOMPLEX ) : LONGCOMPLEX;
(* Divives LONGCOMPLEX value z1 by z2 and returns the result.
   This function is bound to the / operator for type LONGCOMPLEX. *)

VAR
  temp : LONGCOMPLEX;
  normSq : LONGREAL;

BEGIN
  normSq := ( z2.re * z2.re ) + ( z2.im * z2.im );
  temp := multiply (z1, conj (z2) );
  temp.re := temp.re/normSq;
  temp.im := -temp.im/normSq;
  RETURN temp;
END divide;


(* Relational operations *)


PROCEDURE (*[=]*) isEqual ( z1, z2 : LONGCOMPLEX ) : BOOLEAN;
(* Returns TRUE if LONGCOMPLEX values z1 and z2 are equal, otherwise FALSE.
   This function is bound to operators = and # for type LONGCOMPLEX. *)

BEGIN
  RETURN ( z1.re = z2.re ) AND ( z1.im = z2.im)
END isEqual;

END LONGCOMPLEX.