(* (C) 2012-2014 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE IntegerMath (*[ProtoCardMath] FOR INTEGER*);

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYi;

  (* Mathematical functions for the type INTEGER *)

IMPORT CardinalMath;


VAR
  maxSigFigs : INTEGER;


PROCEDURE sgn ( i : INTEGER ) : INTEGER;
(* Returns the signum value of i. *)

BEGIN
  IF i < 0 THEN RETURN -1
  ELSIF i > 0 THEN RETURN 1
  ELSE RETURN 0
  END (* if *)
END sgn;

(* Floored integer division *)

PROCEDURE fdiv ( a, n : INTEGER ) : INTEGER;
(* Divides dividend a by divisor n using floored integer division and
   returns quotient q, where q = floor(a/n) and a = n * q + fmod(a, n).
   A zero divisor n will raise runtime fault DivByZero. *)

VAR
  i : INTEGER;
BEGIN
  i := a MOD n;
  IF i < 0 THEN RETURN i - 1
  ELSE RETURN i
  END (* if *)
END fdiv;


PROCEDURE fmod ( a, n : INTEGER ) : INTEGER;
(* Divides dividend a by divisor n using floored integer division and
   returns remainder r, where r = a - n * fdiv(a, n).
   A zero divisor n will raise runtime fault DivByZero. *)

BEGIN
  RETURN a - n * fdiv(a, n)
END fmod;

PROCEDURE round ( c , n : INTEGER ) : INTEGER;
(* Returns the value of r rounded to n Significant Figures. *)
VAR
  nc, log10n, power10ToAdd: CARDINAL;

BEGIN
  nc := ABS(c);
  nc:= CardinalMath.round ( nc, n );
  IF c < 0 THEN
    RETURN -VAL (INTEGER, nc)
  ELSE
    RETURN nc
  END (* IF c *)
END round;

(* Exponentiation *)


PROCEDURE power (base : INTEGER; exponent: CARDINAL): INTEGER;
  (* Returns the value of the number base raised to the power exponent *)

BEGIN
  IF exponent = 0  THEN RETURN  1;
  ELSIF exponent = 1 THEN RETURN base;
  ELSIF ODD(exponent) THEN RETURN  base * power (base * base, (exponent - 1) DIV 2)
  ELSE RETURN power (base * base, exponent DIV 2)
  END (* IF exponent = 0  *)
END power;


PROCEDURE exp10 (n: INTEGER): INTEGER;
  (* Returns 10 to the n *)
BEGIN
  RETURN power (10, n )
END exp10;


PROCEDURE exp2 (n: INTEGER): INTEGER;
  (* Returns 2 to the n *)
BEGIN
  RETURN power (2, n )
END exp2;

(* Nearest powers of ten and two *)

PROCEDURE LengthInt (n: INTEGER) : CARDINAL;

VAR
  len : CARDINAL;
BEGIN
  len := nearestLargerPowerOf10 (n);
  IF n < 0 THEN len := len + 1 END;
  RETURN len;
END LengthInt;

PROCEDURE nearestSmallerPowerOf10 ( c : INTEGER ) : INTEGER;
(* Returns the nearest power of 2 larger than c.
i.e. the number of significant figures in c less one . *)

BEGIN
  RETURN VAL (INTEGER, CardinalMath.nearestSmallerPowerOf10 (ABS (c) ) );
END nearestSmallerPowerOf10;


PROCEDURE nearestLargerPowerOf10 ( c : INTEGER ) : INTEGER;
(* Returns the nearest power of 2 larger than c.
i.e. the number of significant figures in c . *)

BEGIN
  RETURN nearestSmallerPowerOf10 ( c ) + 1
END nearestLargerPowerOf10;


PROCEDURE nearestSmallerPowerOf2 ( c : INTEGER ) : INTEGER;
(* Returns the nearest power of 2 smaller than c. *)
VAR
  nc : CARDINAL;

BEGIN
  nc := ABS (c);
  nc := CardinalMath.nearestSmallerPowerOf2 (nc);
  RETURN VAL (INTEGER, nc ) * sgn (c);
RETURN nc

END nearestSmallerPowerOf2;


PROCEDURE nearestLargerPowerOf2 ( c : INTEGER ) : INTEGER;
(* Returns the nearest power of 2 larger than c. *)

BEGIN
  RETURN nearestSmallerPowerOf2 ( c ) + 1
END nearestLargerPowerOf2;


PROCEDURE max ( items : (*ARGLIST >0*) ARRAY OF INTEGER ) : INTEGER;
(* Returns the maximum value of the arguments *)
VAR
  largest, index, last : INTEGER;

BEGIN
  last := CAPACITYi (items) -1;
    largest := items [0];
    IF last > 0 THEN
      index := 0;
    WHILE index < last  DO
      index := index + 1;
      IF items [index] > largest THEN
        largest := items [index]
      END (* IF *)
    END; (* while *)
  END; (* if last *)
  RETURN largest;
END max;


PROCEDURE min ( items : (*ARGLIST >0*) ARRAY OF INTEGER ) : INTEGER;
(* Returns the minimum value of the arguments *)
VAR
  smallest, index, last : INTEGER;

BEGIN
  last := CAPACITYi (items) -1;
    smallest := items [0];
    IF last > 0 THEN
      index := 0;
    WHILE index < last  DO
      index := index + 1;
      IF items [index] < smallest THEN
        smallest := items [index]
      END (* IF *)
    END; (* while *)
  END; (* if last *)
  RETURN smallest;
END min;

(* Kludge Section *)

TYPE
  (*TargetType = ;*) (* in definition *)
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;


PROCEDURE max2 (item1, item2 : TargetType ) : TargetType;

BEGIN
  RETURN max (Ary2 { item1, item2} )
END max2;


PROCEDURE max3 (item1, item2, item3 : TargetType ) : TargetType;

BEGIN
  RETURN max (Ary3 { item1, item2, item3} )
END max3;


PROCEDURE min2 (item1, item2 : TargetType ) : TargetType;

BEGIN
  RETURN min (Ary2 { item1, item2} )
END min2;


PROCEDURE min3 (item1, item2, item3 : TargetType ) : TargetType;

BEGIN
  RETURN min (Ary3 { item1, item2, item3} )
END min3;



BEGIN (* main *)
  maxSigFigs := nearestLargerPowerOf10 (MAX (INTEGER) )
END IntegerMath.
