(* (C) 2012-2014 by B.Kowarsch & R.Sutcliffe. All rights reserved. *)

IMPLEMENTATION MODULE LongCardinalMath (*[ProtoCardMath] FOR LONGCARD*);

  (* Mathematical functions for the type LONGCARD *)

IMPORT SYSTEM;
FROM Transition IMPORT CAPACITYlc;


PROCEDURE maxPowerOf10() : CARDINAL;
(* returns nearestLargerPowerOf10 (MAX (CARDINAL); *)
BEGIN
  RETURN nearestLargerPowerOf10 (MAX (LONGCARD) )
END maxPowerOf10;


PROCEDURE round ( c , n : LONGCARD ) : LONGCARD;
(* Returns the value of r rounded to n Significant Figures. *)
VAR
  nc, log10n, power10ToAdd: LONGCARD;

BEGIN
  nc := c;
  log10n := nearestSmallerPowerOf10 ( c );
  power10ToAdd := exp10 (log10n - n);
  IF nc < MAX (LONGCARD) - nc + 5 * power10ToAdd THEN (* otherwise must round down only *)
    nc := nc + 5 * power10ToAdd ; (* add five in the last place wanted plus 1 *)
  END; (* IF nc < MAX  *)
  IF power10ToAdd # 0 THEN
    nc := nc - (nc MOD (power10ToAdd * 10 ) ); (* chop the last sigFigs - n + 1 digits *)
  END;
  RETURN nc
END round;

(* Exponentiation *)


PROCEDURE power (base, exponent: LONGCARD): LONGCARD;
  (* Returns the value of the number base raised to the power exponent *)

BEGIN
  IF exponent = 0  THEN RETURN  1;
  ELSIF exponent = 1 THEN RETURN base;
  ELSIF ODD(exponent) THEN RETURN  base * power (base * base, (exponent - 1) DIV 2)
  ELSE RETURN power (base * base, exponent DIV 2)
  END (* IF exponent = 0 *)
END power;


PROCEDURE exp10 (n: LONGCARD): LONGCARD;
  (* Returns 10 to the n *)
BEGIN
  RETURN power (10, n )
END exp10;


PROCEDURE exp2 (n: LONGCARD): LONGCARD;
  (* Returns 2 to the n *)
BEGIN
  RETURN power (2, n )
END exp2;

(* Nearest powers of ten and two *)


PROCEDURE nearestSmallerPowerOf10 ( c : LONGCARD ) : LONGCARD;
(* Returns the nearest power of 2 larger than c.
i.e. the number of significant figures in c less one . *)
VAR
  np : LONGCARD;

BEGIN
  IF c < 10 THEN RETURN 0 END;
  np := 0;
  WHILE c >= 1000000000 DO (* biggest rocks first *)
    c := c / 1000000000;
    np := np + 9
  END; (* while *)
  WHILE c >= 10000 DO (* bigger rocks first *)
    c := c / 1000;
    np := np + 3
  END; (* while *)
  WHILE c >= 10 DO
    c := c / 10;
    np := np + 1
  END; (* while *)
RETURN np
END nearestSmallerPowerOf10;


PROCEDURE nearestLargerPowerOf10 ( c : LONGCARD ) : LONGCARD;
(* Returns the nearest power of 2 larger than c.
i.e. the number of significant figures in c . *)

BEGIN
  RETURN nearestSmallerPowerOf10 ( c ) + 1
END nearestLargerPowerOf10;


PROCEDURE nearestSmallerPowerOf2 ( c : LONGCARD ) : LONGCARD;
(* Returns the nearest power of 2 smaller than c. *)
VAR
  two30, np : LONGCARD;

BEGIN
  IF c < 2 THEN RETURN 0 END;
  two30 := exp2 (30);
  np := 0;
  WHILE c >= two30 DO (* biggest rocks first *)
    c := c / two30;
    np := np + 30
  END; (* while *)
  WHILE c >= 1024 DO (* bigger rocks first *)
    c := c / 1024;
    np := np + 10
  END; (* while *)
  WHILE c >= 32 DO (* smaller rocks next *)
    c := c / 32;
    np := np + 5;
  END; (* while *)
  WHILE c >= 2 DO
    c := c / 2;
    np := np + 1
  END; (* while *)
RETURN np

END nearestSmallerPowerOf2;


PROCEDURE nearestLargerPowerOf2 ( c : LONGCARD ) : LONGCARD;
(* Returns the nearest power of 2 larger than c. *)

BEGIN
  RETURN nearestSmallerPowerOf2 ( c ) + 1
END nearestLargerPowerOf2;


PROCEDURE max ( items : (*ARGLIST >0*) ARRAY OF LONGCARD ) : LONGCARD;
(* Returns the maximum value of the arguments *)
VAR
  largest, index, last : CARDINAL;

BEGIN
  last := CAPACITYlc (items) -1;
  largest := items [0];
  IF last > 0 THEN
    index := 0;
    WHILE index < last  DO
      index := index + 1;
      IF items [index] > largest THEN
        largest := items [index]
      END (* IF *)
    END; (* while *)
    END; (* if last *)
  RETURN largest;
END max;


PROCEDURE min ( items : (*ARGLIST >0*) ARRAY OF LONGCARD ) : LONGCARD;
(* Returns the minimum value of the arguments *)
VAR
  smallest, index, last : CARDINAL;

BEGIN
  last := CAPACITYlc (items) -1;
    smallest := items [0];
    IF last > 0 THEN
      index := 0;
    WHILE index < last  DO
      index := index + 1;
      IF items [index] < smallest THEN
        smallest := items [index]
      END (* IF *)
    END; (* while *)
  END; (* if last *)
  RETURN smallest;
END min;

(* Kludge Section *)

TYPE
  (*TargetType = ;*) (* in definition *)
  Ary2 = ARRAY [0..1] OF TargetType;
  Ary3 = ARRAY [0..2] OF TargetType;


PROCEDURE max2 (item1, item2 : TargetType ) : TargetType;

BEGIN
  RETURN max (Ary2 { item1, item2} )
END max2;


PROCEDURE max3 (item1, item2, item3 : TargetType ) : TargetType;

BEGIN
  RETURN max (Ary3 { item1, item2, item3} )
END max3;


PROCEDURE min2 (item1, item2 : TargetType ) : TargetType;

BEGIN
  RETURN min (Ary2 { item1, item2} )
END min2;


PROCEDURE min3 (item1, item2, item3 : TargetType ) : TargetType;

BEGIN
  RETURN min (Ary3 { item1, item2, item3} )
END min3;



BEGIN (* main *)


END LongCardinalMath.
